@@ -1,17 +1,31 @@
 /* Advanced array manipulation routines for S-Lang */
-/* Copyright (c) 1998, 1999, 2001, 2002 John E. Davis
- * This file is part of the S-Lang library.
- *
- * You may distribute under the terms of either the GNU General Public
- * License or the Perl Artistic License.
- */
+/*
+Copyright (C) 2004, 2005, 2006 John E. Davis
+
+This file is part of the S-Lang Library.
+
+The S-Lang Library is free software; you can redistribute it and/or
+modify it under the terms of the GNU General Public License as
+published by the Free Software Foundation; either version 2 of the
+License, or (at your option) any later version.
+
+The S-Lang Library is distributed in the hope that it will be useful,
+but WITHOUT ANY WARRANTY; without even the implied warranty of
+MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+General Public License for more details.
+
+You should have received a copy of the GNU General Public License
+along with this library; if not, write to the Free Software
+Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
+USA.  
+*/
 
 #include "slinclud.h"
 
 #include "slang.h"
 #include "_slang.h"
 
-static int next_transposed_index (int *dims, int *max_dims, unsigned int num_dims)
+static int next_transposed_index (SLindex_Type *dims, SLindex_Type *max_dims, unsigned int num_dims)
 {
    int i;
 
@@ -33,7 +47,7 @@ static int next_transposed_index (int *dims, int *max_dims, unsigned int num_dim
 
 static SLang_Array_Type *allocate_transposed_array (SLang_Array_Type *at)
 {
-   unsigned int num_elements;
+   SLuindex_Type num_elements;
    SLang_Array_Type *bt;
    VOID_STAR b_data;
 
@@ -78,8 +92,13 @@ static int check_for_empty_array (char *fun, unsigned int num)
 #endif
 #define SUM_FUNCTION sum_floats
 #define SUM_RESULT_TYPE float
+#define CUMSUM_FUNCTION cumsum_floats
+#define CUMSUM_RESULT_TYPE float
 #define MIN_FUNCTION min_floats
 #define MAX_FUNCTION max_floats
+#define ANY_FUNCTION any_floats
+#define ALL_FUNCTION all_floats
+#define IS_NAN_FUNCTION _pSLmath_isnan
 #include "slarrfun.inc"
 
 /* -------------- DOUBLE --------------------- */
@@ -95,8 +114,13 @@ static int check_for_empty_array (char *fun, unsigned int num)
 #endif
 #define SUM_FUNCTION sum_doubles
 #define SUM_RESULT_TYPE double
+#define CUMSUM_FUNCTION cumsum_doubles
+#define CUMSUM_RESULT_TYPE double
 #define MIN_FUNCTION min_doubles
 #define MAX_FUNCTION max_doubles
+#define ANY_FUNCTION any_doubles
+#define ALL_FUNCTION all_doubles
+#define IS_NAN_FUNCTION _pSLmath_isnan
 #include "slarrfun.inc"
 
 #define GENERIC_TYPE_A double
@@ -124,8 +148,12 @@ static int check_for_empty_array (char *fun, unsigned int num)
 #define TRANSPOSE_2D_ARRAY transpose_ints
 #define SUM_FUNCTION sum_ints
 #define SUM_RESULT_TYPE double
+#define CUMSUM_FUNCTION cumsum_ints
+#define CUMSUM_RESULT_TYPE double
 #define MIN_FUNCTION min_ints
 #define MAX_FUNCTION max_ints
+#define ANY_FUNCTION any_ints
+#define ALL_FUNCTION all_ints
 #include "slarrfun.inc"
 
 /* -------------- UNSIGNED INT --------------------- */
@@ -134,6 +162,8 @@ static int check_for_empty_array (char *fun, unsigned int num)
 #define SUM_RESULT_TYPE double
 #define MIN_FUNCTION min_uints
 #define MAX_FUNCTION max_uints
+#define ANY_FUNCTION any_uints
+#define ALL_FUNCTION all_uints
 #include "slarrfun.inc"
 
 #if SIZEOF_LONG != SIZEOF_INT
@@ -144,6 +174,8 @@ static int check_for_empty_array (char *fun, unsigned int num)
 # define SUM_RESULT_TYPE double
 # define MIN_FUNCTION min_longs
 # define MAX_FUNCTION max_longs
+#define ANY_FUNCTION any_longs
+#define ALL_FUNCTION all_longs
 # include "slarrfun.inc"
 /* -------------- UNSIGNED LONG --------------------- */
 # define GENERIC_TYPE unsigned long
@@ -151,6 +183,8 @@ static int check_for_empty_array (char *fun, unsigned int num)
 # define SUM_RESULT_TYPE double
 # define MIN_FUNCTION min_ulongs
 # define MAX_FUNCTION max_ulongs
+#define ANY_FUNCTION any_ulongs
+#define ALL_FUNCTION all_ulongs
 # include "slarrfun.inc"
 #else
 # define transpose_longs transpose_ints
@@ -160,6 +194,10 @@ static int check_for_empty_array (char *fun, unsigned int num)
 # define min_ulongs min_uints
 # define max_longs max_ints
 # define max_ulongs max_uints
+# define any_longs any_ints
+# define any_ulongs any_uints
+# define all_longs all_ints
+# define all_ulongs all_uints
 #endif
 
 #if SIZEOF_SHORT != SIZEOF_INT
@@ -170,6 +208,8 @@ static int check_for_empty_array (char *fun, unsigned int num)
 # define SUM_RESULT_TYPE double
 # define MIN_FUNCTION min_shorts
 # define MAX_FUNCTION max_shorts
+# define ANY_FUNCTION any_shorts
+# define ALL_FUNCTION all_shorts
 # include "slarrfun.inc"
 /* -------------- UNSIGNED SHORT --------------------- */
 # define GENERIC_TYPE unsigned short
@@ -177,6 +217,8 @@ static int check_for_empty_array (char *fun, unsigned int num)
 # define SUM_RESULT_TYPE double
 # define MIN_FUNCTION min_ushorts
 # define MAX_FUNCTION max_ushorts
+# define ANY_FUNCTION any_ushorts
+# define ALL_FUNCTION all_ushorts
 # include "slarrfun.inc"
 #else
 # define transpose_shorts transpose_ints
@@ -186,6 +228,10 @@ static int check_for_empty_array (char *fun, unsigned int num)
 # define min_ushorts min_uints
 # define max_shorts max_ints
 # define max_ushorts max_uints
+# define any_shorts any_ints
+# define any_ushorts any_uints
+# define all_shorts all_ints
+# define all_ushorts all_uints
 #endif
 
 /* -------------- CHAR --------------------- */
@@ -195,6 +241,8 @@ static int check_for_empty_array (char *fun, unsigned int num)
 #define SUM_RESULT_TYPE double
 #define MIN_FUNCTION min_chars
 #define MAX_FUNCTION max_chars
+#define ANY_FUNCTION any_chars
+#define ALL_FUNCTION all_chars
 #include "slarrfun.inc"
 /* -------------- UNSIGNED CHAR --------------------- */
 #define GENERIC_TYPE unsigned char
@@ -202,17 +250,19 @@ static int check_for_empty_array (char *fun, unsigned int num)
 #define SUM_RESULT_TYPE double
 #define MIN_FUNCTION min_uchars
 #define MAX_FUNCTION max_uchars
+#define ANY_FUNCTION any_uchars
+#define ALL_FUNCTION all_uchars
 #include "slarrfun.inc"
 
 /* This routine works only with linear arrays */
 static SLang_Array_Type *transpose (SLang_Array_Type *at)
 {
-   int dims [SLARRAY_MAX_DIMS];
-   int *max_dims;
+   SLindex_Type dims [SLARRAY_MAX_DIMS];
+   SLindex_Type *max_dims;
    unsigned int num_dims;
    SLang_Array_Type *bt;
    int i;
-   unsigned int sizeof_type;
+   size_t sizeof_type;
    int is_ptr;
    char *b_data;
 
@@ -270,7 +320,7 @@ static SLang_Array_Type *transpose (SLang_Array_Type *at)
 
    do
      {
-	if (-1 == _SLarray_aget_transfer_elem (at, dims, (VOID_STAR) b_data,
+	if (-1 == _pSLarray_aget_transfer_elem (at, dims, (VOID_STAR) b_data,
 					       sizeof_type, is_ptr))
 	  {
 	     SLang_free_array (bt);
@@ -346,8 +396,8 @@ static void do_inner_product (void)
    void (*fun)(SLang_Array_Type *, SLang_Array_Type *, SLang_Array_Type *,
 	       unsigned int, unsigned int, unsigned int, unsigned int, 
 	       unsigned int);
-   unsigned char c_type;
-   int dims[SLARRAY_MAX_DIMS];
+   SLtype c_type;
+   SLindex_Type dims[SLARRAY_MAX_DIMS];
    int status;
    unsigned int a_loops, b_loops, b_inc, a_stride;
    int ai_dims, i, j;
@@ -530,76 +580,131 @@ static void do_inner_product (void)
    SLang_free_array (b);
 }
 #endif
-int SLarray_contract_array (SLarray_Contract_Type *c)
-{
-   int k = 0;
-   int use_all_dims;
-   SLang_Array_Type *at, *new_at;
-   int dims [SLARRAY_MAX_DIMS];
-   int *old_dims;
-   int new_dims[SLARRAY_MAX_DIMS];
-   unsigned int i, j, num_dims;
-   int w[SLARRAY_MAX_DIMS], wk;
-   char *data, *new_data;
-   unsigned int sizeof_type, new_sizeof_type;
-   unsigned int dims_k;
-   int from_type;
 
-   if (c == NULL)
-     return -1;
+static int map_or_contract_array (SLCONST SLarray_Map_Type *c, int use_contraction,
+				  int dim_specified, int *use_this_dim, 
+				  VOID_STAR clientdata)
+{
+   int k, use_all_dims;
+   SLang_Array_Type *at, *new_at;
+   SLindex_Type *old_dims;
+   SLindex_Type old_dims_buf[SLARRAY_MAX_DIMS];
+   SLindex_Type sub_dims[SLARRAY_MAX_DIMS];
+   SLindex_Type tmp_dims[SLARRAY_MAX_DIMS];
+   unsigned int i, j, old_num_dims, sub_num_dims;
+   SLtype new_data_type, old_data_type;
+   char *old_data, *new_data;
+   SLindex_Type w[SLARRAY_MAX_DIMS], wk;
+   size_t old_sizeof_type, new_sizeof_type;
+   SLuindex_Type dims_k;
+   int from_type;
+   SLCONST SLarray_Map_Type *csave;
+   SLarray_Map_Fun_Type *fmap;
+   SLarray_Contract_Fun_Type *fcon;
 
    use_all_dims = 1;
-   if (SLang_Num_Function_Args == 2)
+   k = 0;
+   if (dim_specified)
+     {
+	if (use_this_dim != NULL)
+	  {
+	     k = *use_this_dim;
+	     use_all_dims = 0;
+	  }
+     }
+   else if (SLang_Num_Function_Args == 2)
      {
 	if (-1 == SLang_pop_integer (&k))
 	  return -1;
+
 	use_all_dims = 0;
      }
-   
+
    if (-1 == (from_type = SLang_peek_at_stack1 ()))
      return -1;
 
+   csave = c;
    while (c->f != NULL)
      {
 	if (c->from_type == (SLtype) from_type)
 	  break;
 	c++;
      }
-   
-   if (c->f == NULL)
+
+   /* Look for a more generic version */
+   if (c->f != NULL)
      {
-	SLang_verror (SL_TYPE_MISMATCH, "%s is not supported by this function", SLclass_get_datatype_name (at->data_type));
-	return -1;
+	if (-1 == SLang_pop_array_of_type (&at, c->typecast_to_type))
+	  return -1;
+     }
+   else
+     {
+	/* Look for a wildcard match */
+	c = csave;
+	while (c->f != NULL)
+	  {
+	     if (c->from_type == SLANG_VOID_TYPE)
+	       break;
+	     c++;
+	  }
+	if (c->f == NULL)
+	  {
+	     SLang_verror (SL_TYPE_MISMATCH, "%s is not supported by this function", SLclass_get_datatype_name (from_type));
+	     return -1;
+	  }
+	
+	/* Found it. So, typecast it to appropriate type */
+	if (c->typecast_to_type == SLANG_VOID_TYPE)
+	  {
+	     if (-1 == SLang_pop_array (&at, 1))
+	       return -1;
+	  }
+	else if (-1 == SLang_pop_array_of_type (&at, c->typecast_to_type))
+	  return -1;
      }
 
-   if (-1 == SLang_pop_array_of_type (&at, c->typecast_to_type))
-     return -1;
+   old_data_type = at->data_type;
+   if (SLANG_VOID_TYPE == (new_data_type = c->result_type))
+     new_data_type = old_data_type;
+
+   old_num_dims = at->num_dims;
 
    if (use_all_dims == 0)
      {
 	if (k < 0)
-	  k += at->num_dims;
-	
-	if ((k < 0) || (k >= (int)at->num_dims))
+	  k += old_num_dims;
+
+	if ((k < 0) || (k >= (int)old_num_dims))
 	  {
-	     SLang_verror (SL_INVALID_PARM, "dimension is out of range for contraction");
+	     SLang_verror (SL_INVALID_PARM, "Dimension %d is invalid for a %d-d array",
+			   k, old_num_dims);
 	     SLang_free_array (at);
 	     return -1;
 	  }
+	old_dims = at->dims;
      }
+   else
+     {
+	old_dims = old_dims_buf;
+	old_dims[0] = (SLindex_Type)at->num_elements;
+	old_num_dims = 1;
+     }
+   
+   fcon = (SLarray_Contract_Fun_Type *) c->f;
+   fmap = c->f;
 
-   num_dims = at->num_dims;
-   if (use_all_dims || (num_dims == 1))
+   if (use_contraction 
+       && (use_all_dims || (old_num_dims == 1)))
      {
 	SLang_Class_Type *cl;
 	VOID_STAR buf;
 	int status = 0;
 	
-	cl = _SLclass_get_class (c->result_type);
+	cl = _pSLclass_get_class (new_data_type);
 	buf = cl->cl_transfer_buf;
 
-	if ((-1 == (*c->f) (at->data, 1, at->num_elements, buf))
-	    || (-1 == SLang_push_value (c->result_type, buf)))
+	if ((-1 == (*fcon) (at->data, 1, at->num_elements, buf))
+	    || (-1 == SLang_push_value (new_data_type, buf)))
 	  status = -1;
 	
 	SLang_free_array (at);
@@ -610,16 +715,19 @@ int SLarray_contract_array (SLarray_Contract_Type *c)
     * i_0*W_0 + i_1*W_1 + ... i_{N-1}*W{N-1}
     * where W_j = d_{j+1}d_{j+2}...d_{N-1}
     * and d_k is the number of elements of the kth dimension.
+    * 
+    * For a specified value of k, we
     * So, summing over all elements in the kth dimension of the array
     * means using the set of offsets given by 
     *  
     *   i_k*W_k + sum(j!=k) i_j*W_j.
+    *
+    * So, we want to loop of all dimensions except for the kth using an 
+    * offset given by sum(j!=k)i_jW_j, and an increment W_k between elements.
     */
 
-   old_dims = at->dims;
    wk = 1;
-
-   i = num_dims;
+   i = old_num_dims;
    while (i != 0)
      {
 	i--;
@@ -627,55 +735,93 @@ int SLarray_contract_array (SLarray_Contract_Type *c)
 	wk *= old_dims[i];
      }
    wk = w[k];
-
+   
+   /* Now set up the sub array */
    j = 0;
-   for (i = 0; i < num_dims; i++)
+   for (i = 0; i < old_num_dims; i++)
      {
 	if (i == (unsigned int) k)
 	  continue;
 	
-	new_dims[j] = old_dims[i];
+	sub_dims[j] = old_dims[i];
 	w[j] = w[i];
-	dims[j] = 0;
+	tmp_dims[j] = 0;
 	j++;
      }
+   sub_num_dims = old_num_dims - 1;
 
-   num_dims--;
-   
-   if (NULL == (new_at = SLang_create_array1 (c->result_type, 0, NULL, new_dims, num_dims, 1)))
+   if (use_contraction)
+     new_at = SLang_create_array1 (new_data_type, 0, NULL, sub_dims, sub_num_dims, 1);
+   else
+     new_at = SLang_create_array1 (new_data_type, 0, NULL, old_dims, old_num_dims, 1);
+
+   if (new_at == NULL)
      {
 	SLang_free_array (at);
 	return -1;
      }
 
    new_data = (char *)new_at->data;
-   data = (char *)at->data;
-   sizeof_type = at->sizeof_type;
+   old_data = (char *)at->data;
+   old_sizeof_type = at->sizeof_type;
    new_sizeof_type = new_at->sizeof_type;
    dims_k = old_dims[k] * wk;
 
-   do
+   /* Skip this for cases such as sum(Double_Type[0,0], 1).  Otherwise, 
+    * (*fcon) will write to new_data, which has no length
+    */
+   if (new_at->num_elements) do
      {
-	unsigned int offset = 0;
-	for (i = 0; i < num_dims; i++)
-	  offset += w[i] * dims[i];
+	size_t offset = 0;
+	int status;
+
+	for (i = 0; i < sub_num_dims; i++)
+	  offset += w[i] * tmp_dims[i];
 	
-	if (-1 == (*c->f) ((VOID_STAR) (data + offset*sizeof_type), wk, 
-			   dims_k, (VOID_STAR) new_data))
+	if (use_contraction)
+	  {
+	     status = (*fcon) ((VOID_STAR)(old_data + offset*old_sizeof_type), wk,
+			       dims_k, (VOID_STAR) new_data);
+	     new_data += new_sizeof_type;
+	  }
+	else 
+	  {
+	     status = (*fmap) (old_data_type, (VOID_STAR) (old_data + offset*old_sizeof_type),
+			       wk, dims_k,
+			       new_data_type, (VOID_STAR) (new_data + offset*new_sizeof_type),
+			       clientdata);
+	  }
+	
+	if (status == -1)
 	  {
 	     SLang_free_array (new_at);
 	     SLang_free_array (at);
 	     return -1;
 	  }
-
-	new_data += new_sizeof_type;
      }
-   while (-1 != _SLarray_next_index (dims, new_dims, num_dims));
+   while (-1 != _pSLarray_next_index (tmp_dims, sub_dims, sub_num_dims));
 
    SLang_free_array (at);
    return SLang_push_array (new_at, 1);
 }
 
+   
+int SLarray_map_array (SLCONST SLarray_Map_Type *m)
+{
+   return map_or_contract_array (m, 0, 0, NULL, NULL);
+}
+
+int SLarray_map_array_1 (SLCONST SLarray_Map_Type *m, int *use_this_dim, 
+			 VOID_STAR clientdata)
+{
+   return map_or_contract_array (m, 0, 1, use_this_dim, clientdata);
+}
+
+int SLarray_contract_array (SLCONST SLarray_Contract_Type *c)
+{
+   return map_or_contract_array ((SLarray_Map_Type *)c, 1, 0, NULL, NULL);
+}
+
 #if SLANG_HAS_COMPLEX
 static int sum_complex (VOID_STAR zp, unsigned int inc, unsigned int num, VOID_STAR sp)
 {
@@ -698,20 +844,46 @@ static int sum_complex (VOID_STAR zp, unsigned int inc, unsigned int num, VOID_S
    s[1] = si;
    return 0;
 }
+
+static int cumsum_complex (SLtype xtype, VOID_STAR xp, unsigned int inc, 
+			   unsigned int num,
+			   SLtype ytype, VOID_STAR yp, VOID_STAR clientdata)
+{
+   double *z, *zmax;
+   double cr, ci;
+   double *s;
+
+   (void) xtype; (void) ytype; (void) clientdata;
+   z = (double *)xp;
+   zmax = z + 2*num;
+   s = (double *)yp;
+   inc *= 2;
+   cr = ci = 0.0;
+   while (z < zmax)
+     {
+	cr += z[0];
+	ci += z[1];
+	s[0] = cr;
+	s[1] = ci;
+	z += inc;
+	s += inc;
+     }
+   return 0;
+}
 #endif
 #if SLANG_HAS_FLOAT
-static SLarray_Contract_Type Sum_Functions [] =
+static SLCONST SLarray_Contract_Type Sum_Functions [] =
 {
+     {SLANG_DOUBLE_TYPE, SLANG_DOUBLE_TYPE, SLANG_DOUBLE_TYPE, (SLarray_Contract_Fun_Type *) sum_doubles},
+     {SLANG_FLOAT_TYPE, SLANG_FLOAT_TYPE, SLANG_FLOAT_TYPE, (SLarray_Contract_Fun_Type *) sum_floats},
+     {SLANG_INT_TYPE, SLANG_INT_TYPE, SLANG_DOUBLE_TYPE, (SLarray_Contract_Fun_Type *) sum_ints},
      {SLANG_CHAR_TYPE, SLANG_CHAR_TYPE, SLANG_DOUBLE_TYPE, (SLarray_Contract_Fun_Type *) sum_chars},
      {SLANG_UCHAR_TYPE, SLANG_UCHAR_TYPE, SLANG_DOUBLE_TYPE, (SLarray_Contract_Fun_Type *) sum_uchars},
      {SLANG_SHORT_TYPE, SLANG_SHORT_TYPE, SLANG_DOUBLE_TYPE, (SLarray_Contract_Fun_Type *) sum_shorts},
      {SLANG_USHORT_TYPE, SLANG_USHORT_TYPE, SLANG_DOUBLE_TYPE, (SLarray_Contract_Fun_Type *) sum_ushorts},
      {SLANG_UINT_TYPE, SLANG_UINT_TYPE, SLANG_DOUBLE_TYPE, (SLarray_Contract_Fun_Type *) sum_uints},
-     {SLANG_INT_TYPE, SLANG_INT_TYPE, SLANG_DOUBLE_TYPE, (SLarray_Contract_Fun_Type *) sum_ints},
      {SLANG_LONG_TYPE, SLANG_LONG_TYPE, SLANG_DOUBLE_TYPE, (SLarray_Contract_Fun_Type *) sum_longs},
      {SLANG_ULONG_TYPE, SLANG_ULONG_TYPE, SLANG_DOUBLE_TYPE, (SLarray_Contract_Fun_Type *) sum_ulongs},
-     {SLANG_FLOAT_TYPE, SLANG_FLOAT_TYPE, SLANG_FLOAT_TYPE, (SLarray_Contract_Fun_Type *) sum_floats},
-     {SLANG_DOUBLE_TYPE, SLANG_DOUBLE_TYPE, SLANG_DOUBLE_TYPE, (SLarray_Contract_Fun_Type *) sum_doubles},
 #if SLANG_HAS_COMPLEX
      {SLANG_COMPLEX_TYPE, SLANG_COMPLEX_TYPE, SLANG_COMPLEX_TYPE, (SLarray_Contract_Fun_Type *) sum_complex},
 #endif
@@ -724,7 +896,7 @@ static void array_sum (void)
 }
 #endif
 
-static SLarray_Contract_Type Array_Min_Funs [] = 
+static SLCONST SLarray_Contract_Type Array_Min_Funs [] = 
 {
      {SLANG_CHAR_TYPE, SLANG_CHAR_TYPE, SLANG_CHAR_TYPE, (SLarray_Contract_Fun_Type *) min_chars},
      {SLANG_UCHAR_TYPE, SLANG_UCHAR_TYPE, SLANG_UCHAR_TYPE, (SLarray_Contract_Fun_Type *) min_uchars},
@@ -747,7 +919,7 @@ array_min (void)
    (void) SLarray_contract_array (Array_Min_Funs);
 }
 
-static SLarray_Contract_Type Array_Max_Funs [] =
+static SLCONST SLarray_Contract_Type Array_Max_Funs [] =
 {
      {SLANG_CHAR_TYPE, SLANG_CHAR_TYPE, SLANG_CHAR_TYPE, (SLarray_Contract_Fun_Type *) max_chars},
      {SLANG_UCHAR_TYPE, SLANG_UCHAR_TYPE, SLANG_UCHAR_TYPE, (SLarray_Contract_Fun_Type *) max_uchars},
@@ -771,19 +943,349 @@ array_max (void)
 }
 
 
+static SLCONST SLarray_Map_Type CumSum_Functions [] =
+{
+     {SLANG_DOUBLE_TYPE, SLANG_DOUBLE_TYPE, SLANG_DOUBLE_TYPE, (SLarray_Map_Fun_Type *) cumsum_doubles},
+     {SLANG_INT_TYPE, SLANG_INT_TYPE, SLANG_DOUBLE_TYPE, (SLarray_Map_Fun_Type *) cumsum_ints},
+     {SLANG_LONG_TYPE, SLANG_DOUBLE_TYPE, SLANG_DOUBLE_TYPE, (SLarray_Map_Fun_Type *) cumsum_doubles},
+     {SLANG_FLOAT_TYPE, SLANG_FLOAT_TYPE, SLANG_FLOAT_TYPE, (SLarray_Map_Fun_Type *) cumsum_floats},
+     {SLANG_UINT_TYPE, SLANG_DOUBLE_TYPE, SLANG_DOUBLE_TYPE, (SLarray_Map_Fun_Type *) cumsum_doubles},
+     {SLANG_ULONG_TYPE, SLANG_DOUBLE_TYPE, SLANG_DOUBLE_TYPE, (SLarray_Map_Fun_Type *) cumsum_doubles},
+     {SLANG_CHAR_TYPE, SLANG_FLOAT_TYPE, SLANG_FLOAT_TYPE, (SLarray_Map_Fun_Type *) cumsum_floats},
+     {SLANG_UCHAR_TYPE, SLANG_FLOAT_TYPE, SLANG_FLOAT_TYPE, (SLarray_Map_Fun_Type *) cumsum_floats},
+     {SLANG_SHORT_TYPE, SLANG_FLOAT_TYPE, SLANG_FLOAT_TYPE, (SLarray_Map_Fun_Type *) cumsum_floats},
+     {SLANG_USHORT_TYPE, SLANG_FLOAT_TYPE, SLANG_FLOAT_TYPE, (SLarray_Map_Fun_Type *) cumsum_floats},
+     {SLANG_VOID_TYPE, SLANG_DOUBLE_TYPE, SLANG_DOUBLE_TYPE, (SLarray_Map_Fun_Type *) cumsum_doubles},
+#if SLANG_HAS_COMPLEX
+     {SLANG_COMPLEX_TYPE, SLANG_COMPLEX_TYPE, SLANG_COMPLEX_TYPE, (SLarray_Map_Fun_Type *) cumsum_complex},
+#endif
+     {0, 0, 0, NULL}
+};
+
+static void array_cumsum (void)
+{
+   (void) SLarray_map_array (CumSum_Functions);
+}
+
+static int pop_writable_array (SLang_Array_Type **atp)
+{
+   SLang_Array_Type *at;
+
+   if (-1 == SLang_pop_array (&at, 0))
+     return -1;
+   
+   if (at->flags & SLARR_DATA_VALUE_IS_READ_ONLY)
+     {
+	SLang_set_error (SL_ReadOnly_Error);
+	SLang_free_array (at);
+	return -1;
+     }
+   
+   *atp = at;
+   return 0;
+}
+
+static int check_range_index (int len, int *ip)
+{
+   int i = *ip;
+   if (i < 0)
+     i += len;
+
+   if ((i < 0) || (i >= len))
+     {
+	SLang_set_error (SL_Index_Error);
+	return -1;
+     }
+   *ip = i;
+   return 0;
+}
+
+static int check_range_indices (int len, int *ip, int *jp)
+{
+   int i = *ip, j = *jp;
+
+   if ((-1 == check_range_index (len, &i))
+       || (-1 == check_range_index (len, &j)))
+     return -1;
+
+   if (i > j)
+     {
+	int t = i; i = j; j = t;
+     }
+   *ip = i;
+   *jp = j;
+   return 0;
+}
+
+   
+/* Usage: array_swap (a, i, j [,dim]);  (dim not yet supported) */
+static void array_swap (void)
+{
+   int i, j;
+   int len;
+   unsigned char *src, *dst;
+   size_t sizeof_type;
+   unsigned int k;
+   int dim, have_dim;
+   SLang_Array_Type *at;
+#if 0
+   SLindex_Type dims[SLARRAY_MAX_DIMS];
+#endif
+   have_dim = 0;
+   if (SLang_Num_Function_Args == 4)
+     {
+	if (-1 == SLang_pop_integer (&dim))
+	  return;
+	have_dim = 1;
+     }
+   
+   if ((-1 == SLang_pop_integer (&j))
+       || (-1 == SLang_pop_integer (&i)))
+     return;
+   
+   if (i == j)
+     return;			       /* leave array on stack */
+
+   if (-1 == pop_writable_array (&at))
+     return;
+
+   if (have_dim)
+     {
+	if (-1 == check_range_index (at->num_dims, &dim))
+	  {
+	     SLang_free_array (at);
+	     return;
+	  }
+	len = at->dims [dim];
+     }
+   else len = (int) at->num_elements;
+
+   if (-1 == check_range_indices (len, &i, &j))
+     {
+	SLang_free_array (at);
+	return;
+     }
+
+   sizeof_type = at->cl->cl_sizeof_type;
+   if (have_dim == 0)
+     {
+	src = (unsigned char *)at->data + j*sizeof_type;
+	dst = (unsigned char *)at->data + i*sizeof_type;
+   
+	for (k = 0; k < sizeof_type; k++)
+	  {
+	     unsigned char tmp = src[k];
+	     src[k] = dst[k];
+	     dst[k] = tmp;
+	  }
+	SLang_free_array (at);
+	return;
+     }
+
+   SLang_verror (SL_NOT_IMPLEMENTED, "dim not implemented");
+#if 0
+   /* Otherwise we have perform this swap:
+    * 
+    *    A[*,..,i,*,...] <--> A[*,...,j,*...]
+    * 
+    * Consider 2x2:
+    *        a00 a01 a02 ...
+    *   A =  a10 a11 a12 ...
+    *        a20 a21 a22 ...
+    *         .
+    *
+    * Suppose we swap A[1,*] <--> A[2,*].  We want:
+    * 
+    *        a00 a01 a02 ...
+    *  A' =  a20 a21 a22 ...
+    *        a10 a11 a12 ...
+    *         .
+    * 
+    * Similarly, swapping A[*,1] <--> A[*,2]:
+    * 
+    *        a00 a02 a01 ...
+    *   A =  a10 a12 a11 ...
+    *        a20 a22 a21 ...
+    *         .
+    */
+
+   memset ((char *) dims, 0, sizeof (dims));
+   max_dims = at->dims;
+   dims[dim] = i;
+   src_ptr = (unsigned char *)at->data;
+   ofs = 1;
+   for (d = swap_dim + 1; d < max_dims; d++)
+     {
+	ofs = ofs * max_dims[d];
+     }
+   src_ptr = (unsigned char *)at->data + i * ofs;
+   dst_ptr = (unsigned char *)at->data + j * ofs;
+   
+   for (d = swap_dim; d < max_dims; d++)
+     {
+	stride = 
+	  while (1)
+	    {
+	       int d;
+	       unsigned char *src_ptr;
+	       for (d = num_dims-1; d >= 0; d--)
+		 {
+		    SLindex_Type dims_d;
+		    if (d == swap_dim)
+		      {
+			 src_ptr += sizeof_slice;
+			 continue;
+		      }
+		    XXXXXXXXXXXX   all wrong.
+		      dims_d = dims[d] + 1;
+		    if (dims_d != (int) max_dims[d])
+		      break;
+		    dims[d] = 0;
+		    src_ptr += sizeof_type;
+		 }
+	       if (d == -1)
+		 break;
+	       stride = 1;
+	       k = 0;
+	       while (k < dim)
+		 stride *= at->dims[k];
+	       
+	       k = dim + 1;
+	       
+	       src = (unsigned char *)at->data + j*sizeof_type;
+	       dst = (unsigned char *)at->data + i*sizeof_type;
+	    }
+     }
+#endif	
+}
+
+/* Usage: array_reverse (a, [,from, to] [,dim]) */
+static void array_reverse (void)
+{
+   int len;
+   unsigned char *src, *dst;
+   size_t sizeof_type;
+   int dim = 0;
+   /* int has_dim = 0; */
+   int from = 0;
+   int to = -1;
+   int nargs;
+
+   SLang_Array_Type *at;
+   
+   nargs = SLang_Num_Function_Args;
+   if ((nargs == 2) || (nargs == 4))
+     {
+	/* FIXME!!! */
+	/* has_dim = 1; */
+	if (-1 == SLang_pop_integer (&dim))
+	  return;
+	SLang_verror (SL_NotImplemented_Error, "dim argument not yet implemented");
+	return;
+     }
+
+   if (nargs >= 3)
+     {
+	if ((-1 == SLang_pop_integer (&to))
+	    || (-1 == SLang_pop_integer (&from)))
+	  return;
+     }
+   if (from == to)
+     return;			       /* leave array on stack */
+   
+
+   if (-1 == pop_writable_array (&at))
+     return;
+
+   len = (int) at->num_elements;
+   if (-1 == check_range_indices (len, &from, &to))
+     {
+	SLang_free_array (at);
+	return;
+     }
+
+   sizeof_type = at->cl->cl_sizeof_type;
+
+   src = (unsigned char *)at->data + from*sizeof_type;
+   dst = (unsigned char *)at->data + to*sizeof_type;
+   while (src < dst)
+     {
+	unsigned int k;
+
+	for (k = 0; k < sizeof_type; k++)
+	  {
+	     unsigned char tmp = src[k];
+	     src[k] = dst[k];
+	     dst[k] = tmp;
+	  }
+	
+	src += sizeof_type;
+	dst -= sizeof_type;
+     }
+   SLang_free_array (at);
+}
+     
+static SLCONST SLarray_Contract_Type Array_Any_Funs [] =
+{
+     {SLANG_CHAR_TYPE, SLANG_CHAR_TYPE, SLANG_CHAR_TYPE, (SLarray_Contract_Fun_Type *) any_chars},
+     {SLANG_UCHAR_TYPE, SLANG_UCHAR_TYPE, SLANG_CHAR_TYPE, (SLarray_Contract_Fun_Type *) any_uchars},
+     {SLANG_SHORT_TYPE, SLANG_SHORT_TYPE, SLANG_CHAR_TYPE, (SLarray_Contract_Fun_Type *) any_shorts},
+     {SLANG_USHORT_TYPE, SLANG_USHORT_TYPE, SLANG_CHAR_TYPE, (SLarray_Contract_Fun_Type *) any_ushorts},
+     {SLANG_INT_TYPE, SLANG_INT_TYPE, SLANG_CHAR_TYPE, (SLarray_Contract_Fun_Type *) any_ints},
+     {SLANG_UINT_TYPE, SLANG_UINT_TYPE, SLANG_CHAR_TYPE, (SLarray_Contract_Fun_Type *) any_uints},
+     {SLANG_LONG_TYPE, SLANG_LONG_TYPE, SLANG_CHAR_TYPE, (SLarray_Contract_Fun_Type *) any_longs},
+     {SLANG_ULONG_TYPE, SLANG_ULONG_TYPE, SLANG_CHAR_TYPE, (SLarray_Contract_Fun_Type *) any_ulongs},
+#if SLANG_HAS_FLOAT
+     {SLANG_FLOAT_TYPE, SLANG_FLOAT_TYPE, SLANG_CHAR_TYPE, (SLarray_Contract_Fun_Type *) any_floats},
+     {SLANG_DOUBLE_TYPE, SLANG_DOUBLE_TYPE, SLANG_CHAR_TYPE, (SLarray_Contract_Fun_Type *) any_doubles},
+#endif
+     {0, 0, 0, NULL}
+};
+
+static void 
+array_any (void)
+{
+   (void) SLarray_contract_array (Array_Any_Funs);
+}
+
+
+static SLCONST SLarray_Contract_Type Array_All_Funs [] =
+{
+     {SLANG_CHAR_TYPE, SLANG_CHAR_TYPE, SLANG_CHAR_TYPE, (SLarray_Contract_Fun_Type *) all_chars},
+     {SLANG_UCHAR_TYPE, SLANG_UCHAR_TYPE, SLANG_CHAR_TYPE, (SLarray_Contract_Fun_Type *) all_uchars},
+     {SLANG_SHORT_TYPE, SLANG_SHORT_TYPE, SLANG_CHAR_TYPE, (SLarray_Contract_Fun_Type *) all_shorts},
+     {SLANG_USHORT_TYPE, SLANG_USHORT_TYPE, SLANG_CHAR_TYPE, (SLarray_Contract_Fun_Type *) all_ushorts},
+     {SLANG_INT_TYPE, SLANG_INT_TYPE, SLANG_CHAR_TYPE, (SLarray_Contract_Fun_Type *) all_ints},
+     {SLANG_UINT_TYPE, SLANG_UINT_TYPE, SLANG_CHAR_TYPE, (SLarray_Contract_Fun_Type *) all_uints},
+     {SLANG_LONG_TYPE, SLANG_LONG_TYPE, SLANG_CHAR_TYPE, (SLarray_Contract_Fun_Type *) all_longs},
+     {SLANG_ULONG_TYPE, SLANG_ULONG_TYPE, SLANG_CHAR_TYPE, (SLarray_Contract_Fun_Type *) all_ulongs},
+#if SLANG_HAS_FLOAT
+     {SLANG_FLOAT_TYPE, SLANG_FLOAT_TYPE, SLANG_CHAR_TYPE, (SLarray_Contract_Fun_Type *) all_floats},
+     {SLANG_DOUBLE_TYPE, SLANG_DOUBLE_TYPE, SLANG_CHAR_TYPE, (SLarray_Contract_Fun_Type *) all_doubles},
+#endif
+     {0, 0, 0, NULL}
+};
+
+static void 
+array_all (void)
+{
+   (void) SLarray_contract_array (Array_All_Funs);
+}
+
+
 static SLang_Intrin_Fun_Type Array_Fun_Table [] =
 {
    MAKE_INTRINSIC_1("transpose", array_transpose, SLANG_VOID_TYPE, SLANG_ARRAY_TYPE),
-   SLANG_END_INTRIN_FUN_TABLE
-};
-
-static SLang_Intrin_Fun_Type Array_Math_Fun_Table [] =
-{
 #if SLANG_HAS_FLOAT
    MAKE_INTRINSIC_0("sum", array_sum, SLANG_VOID_TYPE),
+   MAKE_INTRINSIC_0("cumsum", array_cumsum, SLANG_VOID_TYPE),
 #endif
+   MAKE_INTRINSIC_0("array_swap", array_swap, SLANG_VOID_TYPE),
+   MAKE_INTRINSIC_0("array_reverse", array_reverse, SLANG_VOID_TYPE),
    MAKE_INTRINSIC_0("min", array_min, SLANG_VOID_TYPE),
    MAKE_INTRINSIC_0("max", array_max, SLANG_VOID_TYPE),
+   MAKE_INTRINSIC_0("any", array_any, SLANG_VOID_TYPE),
+   MAKE_INTRINSIC_0("all", array_all, SLANG_VOID_TYPE),
    SLANG_END_INTRIN_FUN_TABLE
 };
 
@@ -792,15 +1294,13 @@ int SLang_init_array (void)
    if (-1 == SLadd_intrin_fun_table (Array_Fun_Table, "__SLARRAY__"))
      return -1;
 #if SLANG_HAS_FLOAT
-   _SLang_Matrix_Multiply = do_inner_product;
+   _pSLang_Matrix_Multiply = do_inner_product;
 #endif
    return 0;
 }
 
 int SLang_init_array_extra (void)
 {
-   if (-1 == SLadd_intrin_fun_table (Array_Math_Fun_Table, "__SLARRAY__"))
-     return -1;
    return 0;
 }
 
