@@ -1,9 +1,24 @@
-/* Copyright (c) 1998, 1999, 2001, 2002, 2003 John E. Davis
- * This file is part of the S-Lang library.
- *
- * You may distribute under the terms of either the GNU General Public
- * License or the Perl Artistic License.
- */
+/*
+Copyright (C) 2004, 2005, 2006 John E. Davis
+
+This file is part of the S-Lang Library.
+
+The S-Lang Library is free software; you can redistribute it and/or
+modify it under the terms of the GNU General Public License as
+published by the Free Software Foundation; either version 2 of the
+License, or (at your option) any later version.
+
+The S-Lang Library is distributed in the hope that it will be useful,
+but WITHOUT ANY WARRANTY; without even the implied warranty of
+MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+General Public License for more details.
+
+You should have received a copy of the GNU General Public License
+along with this library; if not, write to the Free Software
+Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
+USA.  
+*/
+
 #include "slinclud.h"
 
 /* #define SL_APP_WANTS_FOREACH */
@@ -12,37 +27,200 @@
 
 #define USE_NEW_ANYTYPE_CODE 1
 
-typedef struct _SLAssoc_Array_Element_Type
+#define MAP_HASH_TO_INDEX(hash, table_len) \
+   (unsigned int)((hash)&(table_len-1))
+
+/* Must be a power of 2 */
+#define MIN_TABLE_SIZE 512
+
+static char *Deleted_Key = "*deleted*";
+
+typedef struct _pSLAssoc_Array_Element_Type
 {
    char *key;		       /* slstring */
-   struct _SLAssoc_Array_Element_Type *next;
+   unsigned long hash;
    SLang_Object_Type value;
 }
-_SLAssoc_Array_Element_Type;
+_pSLAssoc_Array_Element_Type;
 
 typedef struct
 {
-   _SLAssoc_Array_Element_Type *elements[SLASSOC_HASH_TABLE_SIZE];
+   _pSLAssoc_Array_Element_Type *elements;
+   unsigned int table_len;
+   unsigned int num_occupied;	       /* includes deletions */
+   unsigned int num_deleted;
+   unsigned int resize_num;	       /* resize when num_occupied hits this number */
    SLang_Object_Type default_value;
-   unsigned int num_elements;
 #define HAS_DEFAULT_VALUE	1
    unsigned int flags;
-   unsigned char type;
-#if _SLANG_OPTIMIZE_FOR_SPEED
+   SLtype type;
+#if SLANG_OPTIMIZE_FOR_SPEED
    int is_scalar_type;
 #endif
 }
 SLang_Assoc_Array_Type;
 
-#define USE_CACHED_STRING	1
+static int HASH_AGAIN (char *str, unsigned long hash, unsigned int table_len)
+{
+   int h;
+   (void) table_len; (void) str;
+   h = (int)(hash % 311);	       /* 311 should be smaller than MIN_TABLE_SIZE */
+   if (0 == (h & 1))
+     h++;
+   return h;
+}
 
-#if USE_CACHED_STRING 
-static char *Cached_String;
-static SLang_Object_Type *Cached_Obj;
-static SLang_Assoc_Array_Type *Cached_Array;
-#endif
 
-static SLang_Assoc_Array_Type *alloc_assoc_array (unsigned char type, int has_default_value)
+static _pSLAssoc_Array_Element_Type *
+find_element (SLang_Assoc_Array_Type *a, char *str, unsigned long hash)
+{
+   int i, c;
+   _pSLAssoc_Array_Element_Type *e, *elements;
+   int table_len;
+
+   table_len = a->table_len;
+   i = MAP_HASH_TO_INDEX(hash, table_len);
+   e = a->elements + i;
+   if (e->key == str)
+     return e;
+   
+   if (e->key == NULL)
+     return NULL;
+
+   c = HASH_AGAIN(str, hash, table_len);
+   elements = a->elements;
+
+   while (1)
+     {
+	i -= c;
+	if (i < 0)
+	  i += table_len;
+	
+	e = elements + i;
+	if (e->key == str)
+	  return e;
+	if (e->key == NULL)
+	  return NULL;
+     }
+}
+
+static _pSLAssoc_Array_Element_Type *
+find_empty_element (_pSLAssoc_Array_Element_Type *elements, unsigned int table_len,
+		    char *str, unsigned long hash)
+{
+   int i, c;
+   _pSLAssoc_Array_Element_Type *e;
+
+   i = MAP_HASH_TO_INDEX(hash, table_len);
+   e = elements + i;
+   if ((e->key == NULL) || (e->key == Deleted_Key))
+     return e;
+
+   c = HASH_AGAIN(str, hash, table_len);
+   while (1)
+     {
+	i -= c;
+	if (i < 0)
+	  i += table_len;
+	
+	e = elements + i;
+	if ((e->key == NULL) || (e->key == Deleted_Key))
+	  return e;
+     }
+}
+
+static int resize_table (SLang_Assoc_Array_Type *a)
+{
+   int num_occupied, new_table_len;
+   _pSLAssoc_Array_Element_Type *old_es;
+   _pSLAssoc_Array_Element_Type *new_es;
+
+   num_occupied = a->num_occupied - a->num_deleted;
+   new_table_len = a->table_len;
+   
+   if (num_occupied == 0)
+     num_occupied = (MIN_TABLE_SIZE >> 1);
+
+   new_table_len = a->table_len;
+   if (new_table_len < MIN_TABLE_SIZE)
+     new_table_len = MIN_TABLE_SIZE;
+
+   /* In practice, num_occupied*2 will not overflow because we would be
+    * out of memory if would be num_occupied objects stored.
+    */
+   num_occupied *= 2;
+   while (num_occupied > new_table_len)
+     {
+	new_table_len *= 2;
+	if (new_table_len < 0)
+	  {
+	     SLang_set_error (SL_Malloc_Error);
+	     return -1;
+	  }
+     }
+   
+   new_es = (_pSLAssoc_Array_Element_Type *)SLcalloc (new_table_len, sizeof (_pSLAssoc_Array_Element_Type));
+   if (new_es == NULL)
+     return -1;
+   if (NULL != (old_es = a->elements))
+     {
+	_pSLAssoc_Array_Element_Type *new_e, *old_e, *old_emax;
+	
+	old_e = old_es;
+	old_emax = old_e + a->table_len;
+	while (old_e < old_emax)
+	  {
+	     char *key = old_e->key;
+
+	     if ((key == NULL) || (key == Deleted_Key))
+	       {
+		  old_e++;
+		  continue;
+	       }
+	     
+	     /* Cannot fail */
+	     new_e = find_empty_element (new_es, new_table_len, key, old_e->hash);
+	     *new_e = *old_e;
+	     old_e++;
+	  }
+	SLfree ((char *)old_es);
+     }
+   a->elements = new_es;
+   a->table_len = new_table_len;
+   a->num_deleted = 0;
+   a->resize_num = 13*(new_table_len>>4);
+   
+   return 0;
+}
+
+static void delete_assoc_array (SLang_Assoc_Array_Type *a)
+{
+   _pSLAssoc_Array_Element_Type *e, *emax;
+
+   if (a == NULL) return;
+   
+   e = a->elements;
+   if (e != NULL)
+     {
+	emax = e + a->table_len;
+	while (e < emax)
+	  {
+	     if ((e->key != NULL) && (e->key != Deleted_Key))
+	       {
+		  _pSLfree_hashed_string (e->key, strlen (e->key), e->hash);
+		  SLang_free_object (&e->value);
+	       }
+	     e++;
+	  }
+	SLfree ((char *) a->elements);
+     }
+   if (a->flags & HAS_DEFAULT_VALUE)
+     SLang_free_object (&a->default_value);
+
+   SLfree ((char *) a);
+}
+
+static SLang_Assoc_Array_Type *alloc_assoc_array (SLtype type, int has_default_value)
 {
    SLang_Assoc_Array_Type *a;
 
@@ -56,17 +234,17 @@ static SLang_Assoc_Array_Type *alloc_assoc_array (unsigned char type, int has_de
 
    memset ((char *) a, 0, sizeof (SLang_Assoc_Array_Type));
    a->type = type;
-#if _SLANG_OPTIMIZE_FOR_SPEED
-   a->is_scalar_type = (SLANG_CLASS_TYPE_SCALAR == _SLang_get_class_type (type));
+#if SLANG_OPTIMIZE_FOR_SPEED
+   a->is_scalar_type = (SLANG_CLASS_TYPE_SCALAR == _pSLang_get_class_type (type));
 #endif
 
    if (has_default_value)
      {
 	if (
 #if USE_NEW_ANYTYPE_CODE
-	    ((type != SLANG_ANY_TYPE) && (-1 == SLclass_typecast (type, 1, 1)))
+	    ((type != SLANG_ANY_TYPE) && (-1 == SLclass_typecast (type, 1, 0)))
 #else
-	    (-1 == SLclass_typecast (type, 1, 1))
+	    (-1 == SLclass_typecast (type, 1, 0))
 #endif
 	    || (-1 == SLang_pop (&a->default_value)))
 	  {
@@ -76,146 +254,46 @@ static SLang_Assoc_Array_Type *alloc_assoc_array (unsigned char type, int has_de
 
 	a->flags |= HAS_DEFAULT_VALUE;
      }
+   if (-1 == resize_table (a))
+     {
+	delete_assoc_array (a);
+	return NULL;
+     }
    return a;
 }
 
-static void free_element (_SLAssoc_Array_Element_Type *e)
-{
-   if (e == NULL)
-     return;
 
-   SLang_free_object (&e->value);
-   SLang_free_slstring (e->key);
-#if USE_CACHED_STRING
-   if (e->key == Cached_String)
-     Cached_String = NULL;
-#endif
-   SLfree ((char *)e);
-}
-
-static void delete_assoc_array (SLang_Assoc_Array_Type *a)
-{
-   unsigned int i;
-
-   if (a == NULL) return;
-
-   for (i = 0; i < SLASSOC_HASH_TABLE_SIZE; i++)
-     {
-	_SLAssoc_Array_Element_Type *e;
-
-	e = a->elements[i];
-	while (e != NULL)
-	  {
-	     _SLAssoc_Array_Element_Type *next_e;
-
-	     next_e = e->next;
-	     free_element (e);
-	     e = next_e;
-	  }
-     }
-   if (a->flags & HAS_DEFAULT_VALUE)
-     SLang_free_object (&a->default_value);
-
-   SLfree ((char *) a);
-}
-
-_INLINE_
-static SLang_Object_Type *
-find_element (SLang_Assoc_Array_Type *a, char *str, unsigned long hash)
-{
-   unsigned int h;
-   _SLAssoc_Array_Element_Type *e;
-
-   h = (unsigned int) (hash % SLASSOC_HASH_TABLE_SIZE);
-   e = a->elements[h];
-
-   while (e != NULL)
-     {
-	if (str == e->key)       /* slstrings can be compared this way */
-	  {
-#if USE_CACHED_STRING
-	     Cached_String = str;
-	     Cached_Obj = &e->value;
-	     Cached_Array = a;
-#endif
-	     return &e->value;
-	  }
-
-	e = e->next;
-     }
-
-   return NULL;
-}
-
-static _SLAssoc_Array_Element_Type *
-create_element (SLang_Assoc_Array_Type *a, char *str, unsigned long hash)
-{
-   unsigned int h;
-   _SLAssoc_Array_Element_Type *e;
-
-   e = (_SLAssoc_Array_Element_Type *) SLmalloc (sizeof (_SLAssoc_Array_Element_Type));
-   if (e == NULL)
-     return NULL;
-
-   memset ((char *) e, 0, sizeof (_SLAssoc_Array_Element_Type));
-   h = (unsigned int) (hash % SLASSOC_HASH_TABLE_SIZE);
-
-   if (NULL == (str = _SLstring_dup_hashed_string (str, hash)))
-     {
-	SLfree ((char *) e);
-	return NULL;
-     }
-
-   e->key = str;
-   e->next = a->elements[h];
-   a->elements[h] = e;
-
-   a->num_elements += 1;
-#if USE_CACHED_STRING
-   Cached_String = str;
-   Cached_Obj = &e->value;
-   Cached_Array = a;
-#endif
-   return e;
-}
-
-static int store_object (SLang_Assoc_Array_Type *a, char *s, SLang_Object_Type *obj)
+static int store_object (SLang_Assoc_Array_Type *a, SLstr_Type *s, SLang_Object_Type *obj)
 {
    unsigned long hash;
-   SLang_Object_Type *v;
+   _pSLAssoc_Array_Element_Type *e;
 
-#if USE_CACHED_STRING
-   if ((s == Cached_String) && (a == Cached_Array))
-     {
-	v = Cached_Obj;
-	SLang_free_object (v);
-     }
+   hash = _pSLstring_get_hash (s);
+   if (NULL != (e = find_element (a, s, hash)))
+     SLang_free_object (&e->value);
    else
      {
-#endif
-	hash = _SLcompute_string_hash (s);
-	if (NULL != (v = find_element (a, s, hash)))
-	  SLang_free_object (v);
+	if ((a->num_occupied == a->resize_num)
+	    && (-1 == resize_table (a)))
+	  return -1;
+	
+	if (NULL == (e = find_empty_element (a->elements, a->table_len, s, hash)))
+	  return -1;
+	if (e->key == Deleted_Key)
+	  a->num_deleted--;
 	else
-	  {
-	     _SLAssoc_Array_Element_Type *e;
-	     
-	     e = create_element (a, s, hash);
-	     if (e == NULL)
-	       return -1;
-	     
-	     v = &e->value;
-	  }
-#if USE_CACHED_STRING
+	  a->num_occupied++;
+
+	if (NULL == (e->key = _pSLstring_dup_hashed_string (s, hash)))
+	  return -1;
+	
+	e->hash = hash;
      }
-#endif
-
-   *v = *obj;
-
+   e->value = *obj;
    return 0;
 }
 
-static void assoc_destroy (unsigned char type, VOID_STAR ptr)
+static void assoc_destroy (SLtype type, VOID_STAR ptr)
 {
    (void) type;
    delete_assoc_array ((SLang_Assoc_Array_Type *) ptr);
@@ -224,7 +302,7 @@ static void assoc_destroy (unsigned char type, VOID_STAR ptr)
 static int pop_index (unsigned int num_indices,
 		      SLang_MMT_Type **mmt,
 		      SLang_Assoc_Array_Type **a,
-		      char **str)
+		      SLstr_Type **str)
 {
    if (NULL == (*mmt = SLang_pop_mmt (SLANG_ASSOC_TYPE)))
      {
@@ -249,10 +327,12 @@ static int pop_index (unsigned int num_indices,
    return 0;
 }
 
-int _SLassoc_aget (SLtype type, unsigned int num_indices)
+int _pSLassoc_aget (SLtype type, unsigned int num_indices)
 {
+   unsigned long hash;
    SLang_MMT_Type *mmt;
-   char *str;
+   SLstr_Type *str;
+   _pSLAssoc_Array_Element_Type *e;
    SLang_Assoc_Array_Type *a;
    SLang_Object_Type *obj;
    int ret;
@@ -262,42 +342,42 @@ int _SLassoc_aget (SLtype type, unsigned int num_indices)
    if (-1 == pop_index (num_indices, &mmt, &a, &str))
      return -1;
 
-#if USE_CACHED_STRING
-   if ((str == Cached_String) && (a == Cached_Array))
-     obj = Cached_Obj;
-   else
-#endif
-     obj = find_element (a, str, _SLcompute_string_hash (str));
+   hash = _pSLstring_get_hash (str);
+   e = find_element (a, str, hash);
 
-   if ((obj == NULL)
-       && (a->flags & HAS_DEFAULT_VALUE))
-     obj = &a->default_value;
-
-   if (obj == NULL)
+   if (e == NULL)
      {
-	SLang_verror (SL_INTRINSIC_ERROR,
-		      "No such element in Assoc Array: %s", str);
-	ret = -1;
-     }
-   else
-     {
-#if _SLANG_OPTIMIZE_FOR_SPEED
-	if (a->is_scalar_type)
-	  ret = SLang_push (obj);
+	if (a->flags & HAS_DEFAULT_VALUE)
+	  obj = &a->default_value;
 	else
-#endif
-	  ret = _SLpush_slang_obj (obj);
+	  {
+	     ret = -1;
+	     SLang_verror (SL_INTRINSIC_ERROR,
+			   "No such element in Assoc Array: %s", str);
+	     goto free_and_return;
+	  }
+	     
      }
+   else obj = &e->value;
 
-   SLang_free_slstring (str);
+#if SLANG_OPTIMIZE_FOR_SPEED
+   if (a->is_scalar_type)
+     ret = SLang_push (obj);
+   else
+#endif
+     ret = _pSLpush_slang_obj (obj);
+   
+   free_and_return:
+
+   _pSLang_free_slstring (str);
    SLang_free_mmt (mmt);
    return ret;
 }
 
-int _SLassoc_aput (SLtype type, unsigned int num_indices)
+int _pSLassoc_aput (SLtype type, unsigned int num_indices)
 {
    SLang_MMT_Type *mmt;
-   char *str;
+   SLstr_Type *str;
    SLang_Assoc_Array_Type *a;
    SLang_Object_Type obj;
    int ret;
@@ -318,7 +398,7 @@ int _SLassoc_aput (SLtype type, unsigned int num_indices)
 	    )
 	  {
 	     (void) SLang_push (&obj);
-	     if ((-1 == SLclass_typecast (a->type, 1, 1))
+	     if ((-1 == SLclass_typecast (a->type, 1, 0))
 		 || (-1 == SLang_pop (&obj)))
 	       goto the_return;
 	  }
@@ -330,12 +410,12 @@ int _SLassoc_aput (SLtype type, unsigned int num_indices)
      }
 
    the_return:
-   SLang_free_slstring (str);
+   _pSLang_free_slstring (str);
    SLang_free_mmt (mmt);
    return ret;
 }
 
-static int assoc_anew (unsigned char type, unsigned int num_dims)
+static int assoc_anew (SLtype type, unsigned int num_dims)
 {
    SLang_MMT_Type *mmt;
    SLang_Assoc_Array_Type *a;
@@ -384,34 +464,34 @@ static int assoc_anew (unsigned char type, unsigned int num_dims)
 static void assoc_get_keys (SLang_Assoc_Array_Type *a)
 {
    SLang_Array_Type *at;
-   int num;
-   unsigned int i, j;
+   SLindex_Type i, num;
    char **data;
+   _pSLAssoc_Array_Element_Type *e, *emax;
 
    /* Note: If support for threads is added, then we need to modify this
     * algorithm to prevent another thread from modifying the array.
     * However, that should be handled in inner_interp.
     */
-   num = a->num_elements;
+   num = a->num_occupied - a->num_deleted;
 
    if (NULL == (at = SLang_create_array (SLANG_STRING_TYPE, 0, NULL, &num, 1)))
      return;
 
    data = (char **)at->data;
-
+   
+   e = a->elements;
+   emax = e + a->table_len;
+   
    i = 0;
-   for (j = 0; j < SLASSOC_HASH_TABLE_SIZE; j++)
+   while (e < emax)
      {
-	_SLAssoc_Array_Element_Type *e;
-
-	e = a->elements[j];
-	while (e != NULL)
+	if ((e->key != NULL) && (e->key != Deleted_Key))
 	  {
 	     /* Next cannot fail because it is an slstring */
-	     data [i] = SLang_create_slstring (e->key);
-	     e = e->next;
+	     data [i] = _pSLstring_dup_hashed_string (e->key, e->hash);
 	     i++;
 	  }
+	e++;
      }
    (void) SLang_push_array (at, 1);
 }
@@ -428,7 +508,7 @@ transfer_element (SLang_Class_Type *cl, VOID_STAR dest_data,
      {
 	SLang_Any_Type *any;
 
-	if ((-1 == _SLpush_slang_obj (obj))
+	if ((-1 == _pSLpush_slang_obj (obj))
 	    || (-1 == SLang_pop_anytype (&any)))
 	  return -1;
 	
@@ -444,7 +524,7 @@ transfer_element (SLang_Class_Type *cl, VOID_STAR dest_data,
 	return 0;
      }
 
-   src_data = _SLclass_get_ptr_to_value (cl, obj);
+   src_data = _pSLclass_get_ptr_to_value (cl, obj);
 
    if (-1 == (*cl->cl_acopy) (cl->cl_data_type, src_data, dest_data))
      return -1;
@@ -455,21 +535,21 @@ transfer_element (SLang_Class_Type *cl, VOID_STAR dest_data,
 static void assoc_get_values (SLang_Assoc_Array_Type *a)
 {
    SLang_Array_Type *at;
-   int num;
-   unsigned int j;
+   SLindex_Type num;
    char *dest_data;
-   unsigned char type;
+   SLtype type;
    SLang_Class_Type *cl;
    unsigned int sizeof_type;
+   _pSLAssoc_Array_Element_Type *e, *emax;
 
    /* Note: If support for threads is added, then we need to modify this
     * algorithm to prevent another thread from modifying the array.
     * However, that should be handled in inner_interp.
     */
-   num = a->num_elements;
+   num = a->num_occupied - a->num_deleted;
    type = a->type;
 
-   cl = _SLclass_get_class (type);
+   cl = _pSLclass_get_class (type);
    sizeof_type = cl->cl_sizeof_type;
 
    if (NULL == (at = SLang_create_array (type, 0, NULL, &num, 1)))
@@ -477,58 +557,42 @@ static void assoc_get_values (SLang_Assoc_Array_Type *a)
 
    dest_data = (char *)at->data;
 
-   for (j = 0; j < SLASSOC_HASH_TABLE_SIZE; j++)
+   e = a->elements;
+   emax = e + a->table_len;
+   
+   while (e < emax)
      {
-	_SLAssoc_Array_Element_Type *e;
-
-	e = a->elements[j];
-	while (e != NULL)
+	if ((e->key != NULL) && (e->key != Deleted_Key))
 	  {
 	     if (-1 == transfer_element (cl, (VOID_STAR) dest_data, &e->value))
 	       {
 		  SLang_free_array (at);
 		  return;
 	       }
-
 	     dest_data += sizeof_type;
-	     e = e->next;
 	  }
+	e++;
      }
    (void) SLang_push_array (at, 1);
 }
 
 static int assoc_key_exists (SLang_Assoc_Array_Type *a, char *key)
 {
-   return (NULL != find_element (a, key, _SLcompute_string_hash (key)));
+   return (NULL != find_element (a, key, _pSLcompute_string_hash (key)));
 }
 
 static void assoc_delete_key (SLang_Assoc_Array_Type *a, char *key)
 {
-   unsigned int h;
-   _SLAssoc_Array_Element_Type *v, *v0;
+   _pSLAssoc_Array_Element_Type *e;
 
-   h = (unsigned int) (_SLcompute_string_hash (key) % SLASSOC_HASH_TABLE_SIZE);
-
-   v0 = NULL;
-   v = a->elements[h];
-   while (v != NULL)
-     {
-	if (v->key == key)
-	  {
-	     if (v0 != NULL)
-	       v0->next = v->next;
-	     else
-	       a->elements[h] = v->next;
-
-	     free_element (v);
-	     a->num_elements -= 1;
-	     return;
-	  }
-	v0 = v;
-	v = v->next;
-     }
-
-   /* No such element.  Let it pass with no error. */
+   e = find_element (a, key, _pSLstring_get_hash (key));
+   if (e == NULL)
+     return;
+   
+   _pSLang_free_slstring (e->key);
+   SLang_free_object (&e->value);
+   e->key = Deleted_Key;
+   a->num_deleted++;
 }
 
 #define A SLANG_ASSOC_TYPE
@@ -545,32 +609,31 @@ static SLang_Intrin_Fun_Type Assoc_Table [] =
 #undef A
 #undef S
 
-static int assoc_length (unsigned char type, VOID_STAR v, unsigned int *len)
+static int assoc_length (SLtype type, VOID_STAR v, unsigned int *len)
 {
    SLang_Assoc_Array_Type *a;
 
    (void) type;
    a = (SLang_Assoc_Array_Type *) SLang_object_from_mmt (*(SLang_MMT_Type **)v);
-   *len = a->num_elements;
+   *len = a->num_occupied - a->num_deleted;
    return 0;
 }
 
-struct _SLang_Foreach_Context_Type
+struct _pSLang_Foreach_Context_Type
 {
    SLang_MMT_Type *mmt;
    SLang_Assoc_Array_Type *a;
-   unsigned int this_hash_index;
-   unsigned int next_same_hash_index;
+   unsigned int next_hash_index;
 #define CTX_WRITE_KEYS		1
 #define CTX_WRITE_VALUES	2
    unsigned char flags;
-#if _SLANG_OPTIMIZE_FOR_SPEED
+#if SLANG_OPTIMIZE_FOR_SPEED
    int is_scalar;
 #endif
 };
 
 static SLang_Foreach_Context_Type *
-cl_foreach_open (unsigned char type, unsigned int num)
+cl_foreach_open (SLtype type, unsigned int num)
 {
    SLang_Foreach_Context_Type *c;
    unsigned char flags;
@@ -602,12 +665,12 @@ cl_foreach_open (unsigned char type, unsigned int num)
 	     SLang_verror (SL_NOT_IMPLEMENTED,
 			   "using '%s' not supported by SLassoc_Type",
 			   s);
-	     SLang_free_slstring (s);
+	     _pSLang_free_slstring (s);
 	     SLang_free_mmt (mmt);
 	     return NULL;
 	  }
 
-	SLang_free_slstring (s);
+	_pSLang_free_slstring (s);
      }
 
    if (NULL == (c = (SLang_Foreach_Context_Type *) SLmalloc (sizeof (SLang_Foreach_Context_Type))))
@@ -623,13 +686,13 @@ cl_foreach_open (unsigned char type, unsigned int num)
    c->flags = flags;
    c->mmt = mmt;
    c->a = (SLang_Assoc_Array_Type *) SLang_object_from_mmt (mmt);
-#if _SLANG_OPTIMIZE_FOR_SPEED
-   c->is_scalar = (SLANG_CLASS_TYPE_SCALAR == _SLang_get_class_type (c->a->type));
+#if SLANG_OPTIMIZE_FOR_SPEED
+   c->is_scalar = (SLANG_CLASS_TYPE_SCALAR == _pSLang_get_class_type (c->a->type));
 #endif
    return c;
 }
 
-static void cl_foreach_close (unsigned char type, SLang_Foreach_Context_Type *c)
+static void cl_foreach_close (SLtype type, SLang_Foreach_Context_Type *c)
 {
    (void) type;
    if (c == NULL) return;
@@ -637,11 +700,10 @@ static void cl_foreach_close (unsigned char type, SLang_Foreach_Context_Type *c)
    SLfree ((char *) c);
 }
 
-static int cl_foreach (unsigned char type, SLang_Foreach_Context_Type *c)
+static int cl_foreach (SLtype type, SLang_Foreach_Context_Type *c)
 {
    SLang_Assoc_Array_Type *a;
-   _SLAssoc_Array_Element_Type *e;
-   unsigned int i, j;
+   _pSLAssoc_Array_Element_Type *e, *emax;
 
    (void) type;
 
@@ -650,35 +712,20 @@ static int cl_foreach (unsigned char type, SLang_Foreach_Context_Type *c)
 
    a = c->a;
 
-   i = c->this_hash_index;
-   if (i >= SLASSOC_HASH_TABLE_SIZE)
-     return 0;
-
-   e = a->elements[i];
-
-   j = c->next_same_hash_index;
-   c->next_same_hash_index = j + 1;
-
-   while ((j > 0) && (e != NULL))
+   e = a->elements + c->next_hash_index;
+   emax = a->elements + a->table_len;
+   while (1)
      {
-	j--;
-	e = e->next;
+	if (e == emax)
+	  return 0;
+	
+	if ((e->key != NULL) && (e->key != Deleted_Key))
+	  break;
+	
+	e++;
      }
 
-   if (e == NULL)
-     {
-	do
-	  {
-	     i++;
-	     if (i >= SLASSOC_HASH_TABLE_SIZE)
-	       return 0;		       /* no more */
-	  }
-	while (a->elements [i] == NULL);
-
-	e = a->elements[i];
-	c->this_hash_index = i;
-	c->next_same_hash_index = 1;
-     }
+   c->next_hash_index = (e - a->elements) + 1;
 
    if ((c->flags & CTX_WRITE_KEYS)
        && (-1 == SLang_push_string (e->key)))
@@ -686,7 +733,7 @@ static int cl_foreach (unsigned char type, SLang_Foreach_Context_Type *c)
 
    if (c->flags & CTX_WRITE_VALUES)
      {
-#if _SLANG_OPTIMIZE_FOR_SPEED
+#if SLANG_OPTIMIZE_FOR_SPEED
 	if (c->is_scalar)
 	  {
 	     if (-1 == SLang_push (&e->value))
@@ -694,7 +741,7 @@ static int cl_foreach (unsigned char type, SLang_Foreach_Context_Type *c)
 	  }
 	else
 #endif
-	  if (-1 == _SLpush_slang_obj (&e->value))
+	  if (-1 == _pSLpush_slang_obj (&e->value))
 	    return -1;
      }
    
@@ -713,8 +760,8 @@ int SLang_init_slassoc (void)
      return -1;
 
    (void) SLclass_set_destroy_function (cl, assoc_destroy);
-   (void) SLclass_set_aput_function (cl, _SLassoc_aput);
-   (void) SLclass_set_aget_function (cl, _SLassoc_aget);
+   (void) SLclass_set_aput_function (cl, _pSLassoc_aput);
+   (void) SLclass_set_aget_function (cl, _pSLassoc_aget);
    (void) SLclass_set_anew_function (cl, assoc_anew);
    cl->cl_length = assoc_length;
    cl->cl_foreach_open = cl_foreach_open;
