@@ -1,15 +1,30 @@
-/* Copyright (c) 1998, 1999, 2001, 2002 John E. Davis
- * This file is part of the S-Lang library.
- *
- * You may distribute under the terms of either the GNU General Public
- * License or the Perl Artistic License.
- */
+/*
+Copyright (C) 2004, 2005, 2006 John E. Davis
+
+This file is part of the S-Lang Library.
+
+The S-Lang Library is free software; you can redistribute it and/or
+modify it under the terms of the GNU General Public License as
+published by the Free Software Foundation; either version 2 of the
+License, or (at your option) any later version.
+
+The S-Lang Library is distributed in the hope that it will be useful,
+but WITHOUT ANY WARRANTY; without even the implied warranty of
+MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+General Public License for more details.
+
+You should have received a copy of the GNU General Public License
+along with this library; if not, write to the Free Software
+Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
+USA.  
+*/
+
 #include "slinclud.h"
 
 #include "slang.h"
 #include "_slang.h"
 
-struct _SLang_BString_Type
+struct _pSLang_BString_Type
 {
    unsigned int num_refs;
    unsigned int len;
@@ -142,7 +157,7 @@ void SLbstring_free (SLang_BString_Type *b)
 	break;
 
       case IS_SLSTRING:
-	SLang_free_slstring ((char *)b->v.ptr);
+	_pSLang_free_slstring ((char *)b->v.ptr);
 	break;
 
       case IS_MALLOCED:
@@ -173,8 +188,8 @@ int SLang_push_bstring (SLang_BString_Type *b)
 }
 
 static int
-bstring_bstring_bin_op_result (int op, unsigned char a, unsigned char b,
-			       unsigned char *c)
+bstring_bstring_bin_op_result (int op, SLtype a, SLtype b,
+			       SLtype *c)
 {
    (void) a;
    (void) b;
@@ -254,8 +269,8 @@ static void free_n_bstrings (SLang_BString_Type **a, unsigned int n)
 
 static int
 bstring_bstring_bin_op (int op,
-			unsigned char a_type, VOID_STAR ap, unsigned int na,
-			unsigned char b_type, VOID_STAR bp, unsigned int nb,
+			SLtype a_type, VOID_STAR ap, unsigned int na,
+			SLtype b_type, VOID_STAR bp, unsigned int nb,
 			VOID_STAR cp)
 {
    char *ic;
@@ -408,8 +423,8 @@ make_n_bstrings (SLang_BString_Type **b, char **a, unsigned int n, int ptr_type)
 
 static int
 bstring_string_bin_op (int op,
-		       unsigned char a_type, VOID_STAR ap, unsigned int na,
-		       unsigned char b_type, VOID_STAR bp, unsigned int nb,
+		       SLtype a_type, VOID_STAR ap, unsigned int na,
+		       SLtype b_type, VOID_STAR bp, unsigned int nb,
 		       VOID_STAR cp)
 {
    SLang_BString_Type **b;
@@ -430,8 +445,8 @@ bstring_string_bin_op (int op,
 
 static int
 string_bstring_bin_op (int op,
-		       unsigned char a_type, VOID_STAR ap, unsigned int na,
-		       unsigned char b_type, VOID_STAR bp, unsigned int nb,
+		       SLtype a_type, VOID_STAR ap, unsigned int na,
+		       SLtype b_type, VOID_STAR bp, unsigned int nb,
 		       VOID_STAR cp)
 {
    SLang_BString_Type **a;
@@ -451,21 +466,21 @@ string_bstring_bin_op (int op,
    return ret;
 }
 
-static void bstring_destroy (unsigned char unused, VOID_STAR s)
+static void bstring_destroy (SLtype unused, VOID_STAR s)
 {
    (void) unused;
    SLbstring_free (*(SLang_BString_Type **) s);
 }
 
-static int bstring_push (unsigned char unused, VOID_STAR sptr)
+static int bstring_push (SLtype unused, VOID_STAR sptr)
 {
    (void) unused;
 
    return SLang_push_bstring (*(SLang_BString_Type **) sptr);
 }
 
-static int string_to_bstring (unsigned char a_type, VOID_STAR ap, unsigned int na,
-			      unsigned char b_type, VOID_STAR bp)
+static int string_to_bstring (SLtype a_type, VOID_STAR ap, unsigned int na,
+			      SLtype b_type, VOID_STAR bp)
 {
    char **s;
    SLang_BString_Type **b;
@@ -482,8 +497,8 @@ static int string_to_bstring (unsigned char a_type, VOID_STAR ap, unsigned int n
    return 1;
 }
 
-static int bstring_to_string (unsigned char a_type, VOID_STAR ap, unsigned int na,
-			      unsigned char b_type, VOID_STAR bp)
+static int bstring_to_string (SLtype a_type, VOID_STAR ap, unsigned int na,
+			      SLtype b_type, VOID_STAR bp)
 {
    char **s;
    unsigned int i;
@@ -510,7 +525,7 @@ static int bstring_to_string (unsigned char a_type, VOID_STAR ap, unsigned int n
 	     while (i != 0)
 	       {
 		  i--;
-		  SLang_free_slstring (s[i]);
+		  _pSLang_free_slstring (s[i]);
 		  s[i] = NULL;
 	       }
 	     return -1;
@@ -520,7 +535,7 @@ static int bstring_to_string (unsigned char a_type, VOID_STAR ap, unsigned int n
    return 1;
 }
 
-static char *bstring_string (unsigned char type, VOID_STAR v)
+static char *bstring_string (SLtype type, VOID_STAR v)
 {
    SLang_BString_Type *s;
    unsigned char buf[128];
@@ -578,14 +593,14 @@ static unsigned int bstrlen_cmd (SLang_BString_Type *b)
 static SLang_Intrin_Fun_Type BString_Table [] = /*{{{*/
 {
    MAKE_INTRINSIC_1("bstrlen",  bstrlen_cmd, SLANG_UINT_TYPE, SLANG_BSTRING_TYPE),
-   MAKE_INTRINSIC_0("pack", _SLpack, SLANG_VOID_TYPE),
-   MAKE_INTRINSIC_2("unpack", _SLunpack, SLANG_VOID_TYPE, SLANG_STRING_TYPE, SLANG_BSTRING_TYPE),
-   MAKE_INTRINSIC_1("pad_pack_format", _SLpack_pad_format, SLANG_VOID_TYPE, SLANG_STRING_TYPE),
-   MAKE_INTRINSIC_1("sizeof_pack", _SLpack_compute_size, SLANG_UINT_TYPE, SLANG_STRING_TYPE),
+   MAKE_INTRINSIC_0("pack", _pSLpack, SLANG_VOID_TYPE),
+   MAKE_INTRINSIC_2("unpack", _pSLunpack, SLANG_VOID_TYPE, SLANG_STRING_TYPE, SLANG_BSTRING_TYPE),
+   MAKE_INTRINSIC_1("pad_pack_format", _pSLpack_pad_format, SLANG_VOID_TYPE, SLANG_STRING_TYPE),
+   MAKE_INTRINSIC_1("sizeof_pack", _pSLpack_compute_size, SLANG_UINT_TYPE, SLANG_STRING_TYPE),
    SLANG_END_INTRIN_FUN_TABLE
 };
 
-int _SLang_init_bstring (void)
+int _pSLang_init_bstring (void)
 {
    SLang_Class_Type *cl;
 
