@@ -1,9 +1,24 @@
-/* Copyright (c) 1998, 1999, 2001, 2002 John E. Davis
- * This file is part of the S-Lang library.
- *
- * You may distribute under the terms of either the GNU General Public
- * License or the Perl Artistic License.
- */
+/*
+Copyright (C) 2004, 2005 John E. Davis
+
+This file is part of the S-Lang Library.
+
+The S-Lang Library is free software; you can redistribute it and/or
+modify it under the terms of the GNU General Public License as
+published by the Free Software Foundation; either version 2 of the
+License, or (at your option) any later version.
+
+The S-Lang Library is distributed in the hope that it will be useful,
+but WITHOUT ANY WARRANTY; without even the implied warranty of
+MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+General Public License for more details.
+
+You should have received a copy of the GNU General Public License
+along with this library; if not, write to the Free Software
+Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
+USA.  
+*/
+
 #include <stdio.h>
 
 #ifndef SLANG_VERSION
@@ -13,14 +28,22 @@
 /* This is a temporary hack until lynx is fixed to not include this file. */
 #ifndef LYCURSES_H
 
+typedef unsigned long SLcurses_Char_Type;
+typedef struct SLcurses_Cell_Type
+{
+    SLcurses_Char_Type main;
+    SLwchar_Type combining[SLSMG_MAX_CHARS_PER_CELL-1];
+}
+SLcurses_Cell_Type;
+
 typedef struct
 {
    unsigned int _begy, _begx, _maxy, _maxx;
    unsigned int _curx, _cury;
    unsigned int nrows, ncols;
    unsigned int scroll_min, scroll_max;
-   SLsmg_Char_Type **lines;
-   SLsmg_Char_Type color;
+   SLcurses_Cell_Type **lines;
+   int color;
    int is_subwin;
    SLtt_Char_Type attr;
    int delay_off;
@@ -102,7 +125,7 @@ extern SLcurses_Window_Type *SLcurses_Stdscr;
 
 #define winch(w) \
     ((((w)->_cury < (w)->nrows) && ((w)->_curx < (w)->ncols)) \
-       ? ((w)->lines[(w)->_cury][(w)->_curx]) : 0)
+       ? ((w)->lines[(w)->_cury][(w)->_curx].main) : 0)
 
 #define inch() winch(stdscr)
 #define mvwinch(w,x,y) \
@@ -130,7 +153,7 @@ extern int SLcurses_wattroff (SLcurses_Window_Type *, SLtt_Char_Type);
 #define wattr_get(w) ((w)->color << 8)
 #define attr_get() wattr_get(stdscr)
 
-#define COLOR_PAIR(x) ((x) << 8)
+#define COLOR_PAIR(x) ((SLcurses_Char_Type)(x) << 24)
 
 extern int SLcurses_start_color (void);
 #define start_color SLcurses_start_color
@@ -252,18 +275,19 @@ extern SLtt_Char_Type SLcurses_Acs_Map [128];
 #define vline(x,y) SLcurses_nil ()
 #endif
 
-#define A_CHARTEXT	0x00FF
-#define A_NORMAL 0
-#define A_BOLD		0x1000
-#define A_REVERSE	0x2000
+#define A_CHARTEXT	0x001FFFFFUL	/* was 0x00FF */
+#define A_NORMAL	0x00000000UL
+#define A_UNUSED	0x00e00000UL	/* maybe more colors? */
+#define A_COLOR		0x0f000000UL	/* was 0x0F00 */
+#define A_BOLD		0x10000000UL	/* was 0x1000 */
+#define A_REVERSE	0x20000000UL	/* was 0x2000 */
 #define A_STANDOUT	A_REVERSE
-#define A_UNDERLINE	0x4000
-#define A_BLINK		0
-#define A_COLOR		0x0700
-#define A_ALTCHARSET	0x8000
-#define A_DIM		0
-#define A_PROTECT	0
-#define A_INVIS		0
+#define A_UNDERLINE	0x40000000UL	/* was 0x4000 */
+#define A_ALTCHARSET	0x80000000UL	/* was 0x8000 */
+#define A_BLINK		0x00000000UL
+#define A_DIM		0x00000000UL
+#define A_PROTECT	0x00000000UL
+#define A_INVIS		0x00000000UL
 
 #define COLOR_BLACK	SLSMG_COLOR_BLACK
 #define COLOR_RED	SLSMG_COLOR_RED
