@@ -1,10 +1,25 @@
 /* Intrinsic Structure type implementation */
-/* Copyright (c) 1998, 1999, 2001, 2002, 2003 John E. Davis
- * This file is part of the S-Lang library.
- *
- * You may distribute under the terms of either the GNU General Public
- * License or the Perl Artistic License.
- */
+/*
+Copyright (C) 2004, 2005, 2006 John E. Davis
+
+This file is part of the S-Lang Library.
+
+The S-Lang Library is free software; you can redistribute it and/or
+modify it under the terms of the GNU General Public License as
+published by the Free Software Foundation; either version 2 of the
+License, or (at your option) any later version.
+
+The S-Lang Library is distributed in the hope that it will be useful,
+but WITHOUT ANY WARRANTY; without even the implied warranty of
+MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+General Public License for more details.
+
+You should have received a copy of the GNU General Public License
+along with this library; if not, write to the Free Software
+Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
+USA.  
+*/
+
 #include "slinclud.h"
 
 #include "slang.h"
@@ -18,16 +33,16 @@ typedef struct
    VOID_STAR addr;
    SLang_IStruct_Field_Type *fields;
 }
-_SLang_IStruct_Type;
+_pSLang_IStruct_Type;
 
 static SLang_IStruct_Field_Type *istruct_pop_field (char *name, int no_readonly, VOID_STAR *addr)
 {
-   _SLang_IStruct_Type *s;
+   _pSLang_IStruct_Type *s;
    SLang_IStruct_Field_Type *f;
    char *struct_addr;
 
    /* Note: There is no need to free this object */
-   if (-1 == SLclass_pop_ptr_obj (SLANG_ISTRUCT_TYPE, (VOID_STAR *) &s))
+   if (-1 == SLclass_pop_ptr_obj (SLANG_ISTRUCT_TYPE, VOID_STAR_STAR(&s)))
      return NULL;
 
    if (NULL == (struct_addr = *(char **)s->addr))
@@ -63,7 +78,7 @@ static SLang_IStruct_Field_Type *istruct_pop_field (char *name, int no_readonly,
    return NULL;
 }
 
-static int istruct_sget (unsigned char type, char *name)
+static int istruct_sget (SLtype type, char *name)
 {
    SLang_IStruct_Field_Type *f;
    VOID_STAR addr;
@@ -73,12 +88,12 @@ static int istruct_sget (unsigned char type, char *name)
      return -1;
 
    type = f->type;
-   cl = _SLclass_get_class (type);
+   cl = _pSLclass_get_class (type);
 
    return (cl->cl_push_intrinsic)(f->type, addr);
 }
 
-static int istruct_sput (unsigned char type, char *name)
+static int istruct_sput (SLtype type, char *name)
 {
    SLang_IStruct_Field_Type *f;
    VOID_STAR addr;
@@ -88,16 +103,16 @@ static int istruct_sput (unsigned char type, char *name)
      return -1;
 
    type = f->type;
-   cl = _SLclass_get_class (type);
+   cl = _pSLclass_get_class (type);
 
    return (*cl->cl_pop) (type, addr);
 }
 
-static int istruct_push (unsigned char type, VOID_STAR ptr)
+static int istruct_push (SLtype type, VOID_STAR ptr)
 {
-   _SLang_IStruct_Type *s;
+   _pSLang_IStruct_Type *s;
 
-   s = *(_SLang_IStruct_Type **) ptr;
+   s = *(_pSLang_IStruct_Type **) ptr;
    if ((s == NULL) 
        || (s->addr == NULL)
        || (*(char **) s->addr == NULL))
@@ -106,12 +121,12 @@ static int istruct_push (unsigned char type, VOID_STAR ptr)
    return SLclass_push_ptr_obj (type, (VOID_STAR) s);
 }
 
-static int istruct_pop (unsigned char type, VOID_STAR ptr)
+static int istruct_pop (SLtype type, VOID_STAR ptr)
 {
    return SLclass_pop_ptr_obj (type, (VOID_STAR *)ptr);
 }
 
-static void istruct_destroy (unsigned char type, VOID_STAR ptr)
+static void istruct_destroy (SLtype type, VOID_STAR ptr)
 {
    (void) type;
    (void) ptr;
@@ -121,7 +136,7 @@ static void istruct_destroy (unsigned char type, VOID_STAR ptr)
  * is passed here is actually a pointer to the struct.  So, pass its address
  * to istruct_push since v is a variable.  Confusing, n'est pas?
  */
-static int istruct_push_intrinsic (unsigned char type, VOID_STAR v)
+static int istruct_push_intrinsic (SLtype type, VOID_STAR v)
 {
    return istruct_push (type, (VOID_STAR) &v);
 }
@@ -144,7 +159,7 @@ static int init_intrin_struct (void)
    cl->cl_destroy = istruct_destroy;
    cl->cl_push_intrinsic = istruct_push_intrinsic;
 
-   if (-1 == SLclass_register_class (cl, SLANG_ISTRUCT_TYPE, sizeof (_SLang_IStruct_Type *),
+   if (-1 == SLclass_register_class (cl, SLANG_ISTRUCT_TYPE, sizeof (_pSLang_IStruct_Type *),
 				     SLANG_CLASS_TYPE_PTR))
      return -1;
 
@@ -154,7 +169,7 @@ static int init_intrin_struct (void)
 
 int SLns_add_istruct_table (SLang_NameSpace_Type *ns, SLang_IStruct_Field_Type *fields, VOID_STAR addr, char *name)
 {
-   _SLang_IStruct_Type *s;
+   _pSLang_IStruct_Type *s;
    SLang_IStruct_Field_Type *f;
 
    if (-1 == init_intrin_struct ())
@@ -193,11 +208,11 @@ int SLns_add_istruct_table (SLang_NameSpace_Type *ns, SLang_IStruct_Field_Type *
 	f++;
      }
 
-   s = (_SLang_IStruct_Type *)SLmalloc (sizeof (_SLang_IStruct_Type));
+   s = (_pSLang_IStruct_Type *)SLmalloc (sizeof (_pSLang_IStruct_Type));
    if (s == NULL)
      return -1;
 
-   memset ((char *)s, 0, sizeof (_SLang_IStruct_Type));
+   memset ((char *)s, 0, sizeof (_pSLang_IStruct_Type));
    if (NULL == (s->name = SLang_create_slstring (name)))
      {
 	SLfree ((char *) s);
