@@ -1,9 +1,24 @@
-/* Copyright (c) 1992, 1999, 2001, 2002, 2003 John E. Davis
- * This file is part of the S-Lang library.
- *
- * You may distribute under the terms of either the GNU General Public
- * License or the Perl Artistic License.
- */
+/*
+Copyright (C) 2004, 2005, 2006 John E. Davis
+
+This file is part of the S-Lang Library.
+
+The S-Lang Library is free software; you can redistribute it and/or
+modify it under the terms of the GNU General Public License as
+published by the Free Software Foundation; either version 2 of the
+License, or (at your option) any later version.
+
+The S-Lang Library is distributed in the hope that it will be useful,
+but WITHOUT ANY WARRANTY; without even the implied warranty of
+MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+General Public License for more details.
+
+You should have received a copy of the GNU General Public License
+along with this library; if not, write to the Free Software
+Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
+USA.  
+*/
+
 #include "slinclud.h"
 
 #ifdef SL_MALLOC_DEBUG
@@ -75,14 +90,8 @@ static void fixup (unsigned char *p, unsigned long n, char *what)
 #endif
 }
 
-static void SLmalloc_doerror (char *buf)
-{
-   SLang_doerror (buf);
-}
-
 static int check_memory (unsigned char *p, char *what)
 {
-   char buf[128];
    unsigned long n;
 
    register_at_exit_fun ();
@@ -94,8 +103,7 @@ static int check_memory (unsigned char *p, char *what)
 
    if (n == 0xFFFFFFFFUL)
      {
-	sprintf (buf, "%s: %p: Already FREE! Abort NOW.", what, p - Chunk);
-	SLmalloc_doerror (buf);
+	SLang_verror (SL_INVALID_DATA_ERROR, "%s: %p: Already FREE! Abort NOW.", what, p - Chunk);
 	return -1;
      }
 
@@ -104,8 +112,7 @@ static int check_memory (unsigned char *p, char *what)
        || (*(p + (int) (n + 2)) != 81)
        || (*(p + (int) (n + 3)) != 86))
      {
-	sprintf (buf, "\007%s: %p: Memory corrupt! Abort NOW.", what, p);
-	SLmalloc_doerror (buf);
+	SLang_verror (SL_INVALID_DATA_ERROR, "\007%s: %p: Memory corrupt! Abort NOW.", what, p);
 	return -1;
      }
 
@@ -114,9 +121,8 @@ static int check_memory (unsigned char *p, char *what)
    Total_Allocated -= (long) n;
    if (Total_Allocated < 0)
      {
-	sprintf (buf, "\007%s: %p\nFreed %ld, Allocated is: %ld!\n",
+	SLang_verror (SL_INVALID_DATA_ERROR, "\007%s: %p\nFreed %ld, Allocated is: %ld!\n",
 		 what, p, (long) n, Total_Allocated);
-	SLang_doerror (buf);
      }
 #ifdef SLDEBUG_DOUT
    fprintf (dout, "FREE: %s:\t%p %ld\n", what, p, (long) n);
