@@ -1,11 +1,25 @@
 /* This module implements an interface to posix system calls */
 /* file stdio intrinsics for S-Lang */
-/* Copyright (c) 1992, 1999, 2001, 2002, 2003 John E. Davis
- * This file is part of the S-Lang library.
- *
- * You may distribute under the terms of either the GNU General Public
- * License or the Perl Artistic License.
- */
+/*
+Copyright (C) 2004, 2005, 2006 John E. Davis
+
+This file is part of the S-Lang Library.
+
+The S-Lang Library is free software; you can redistribute it and/or
+modify it under the terms of the GNU General Public License as
+published by the Free Software Foundation; either version 2 of the
+License, or (at your option) any later version.
+
+The S-Lang Library is distributed in the hope that it will be useful,
+but WITHOUT ANY WARRANTY; without even the implied warranty of
+MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+General Public License for more details.
+
+You should have received a copy of the GNU General Public License
+along with this library; if not, write to the Free Software
+Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
+USA.  
+*/
 
 #include "slinclud.h"
 
@@ -48,7 +62,7 @@
 #include "slang.h"
 #include "_slang.h"
 
-struct _SLFile_FD_Type
+struct _pSLFile_FD_Type
 {
    char *name;
    unsigned int num_refs;	       /* reference counting */
@@ -61,6 +75,27 @@ struct _SLFile_FD_Type
    int (*write)(int, char *, unsigned int *);
 };
 
+
+static int is_interrupt (int e)
+{
+#ifdef EINTR
+   if (e == EINTR)
+     {
+	if (0 == SLang_handle_interrupt ())
+	  return 1;
+     }
+#endif
+#ifdef EAGAIN
+   if (e == EAGAIN)
+     {
+	if (0 == SLang_handle_interrupt ())
+	  return 1;
+     }
+#endif
+   return 0;
+}
+	
+
 static int close_method (int fd)
 {
    return close (fd);
@@ -99,7 +134,7 @@ static int check_fd (int fd)
    if (fd == -1)
      {
 #ifdef EBADF
-	_SLerrno_errno = EBADF;
+	_pSLerrno_errno = EBADF;
 #endif
 	return -1;
      }
@@ -112,11 +147,16 @@ static int posix_close (SLFile_FD_Type *f)
    if (-1 == check_fd (f->fd))
      return -1;
 
-   if ((f->close != NULL)
-       && (-1 == f->close (f->fd)))
+   if (f->close != NULL)
      {
-	_SLerrno_errno = errno;
-	return -1;
+	while (-1 == f->close (f->fd))
+	  {
+	     if (is_interrupt (errno))
+	       continue;
+
+	     _pSLerrno_errno = errno;
+	     return -1;
+	  }
      }
 
    if (f->stdio_mmt != NULL)
@@ -142,9 +182,12 @@ static void posix_write (SLFile_FD_Type *f, SLang_BString_Type *bstr)
 	return;
      }
 
-   if (-1 == f->write (f->fd, p, &len))
+   while (-1 == f->write (f->fd, p, &len))
      {
-	_SLerrno_errno = errno;
+	if (is_interrupt (errno))
+	  continue;
+
+	_pSLerrno_errno = errno;
 	SLang_push_integer (-1);
 	return;
      }
@@ -166,9 +209,12 @@ static void posix_read (SLFile_FD_Type *f, SLang_Ref_Type *ref, unsigned int *nb
        || (NULL == (b = SLmalloc (len + 1))))
      goto return_error;
    
-   if (-1 == f->read (f->fd, b, &len))
+   while (-1 == f->read (f->fd, b, &len))
      {
-	_SLerrno_errno = errno;
+	if (is_interrupt (errno))
+	  continue;
+
+	_pSLerrno_errno = errno;
 	goto return_error;
      }
 
@@ -182,14 +228,15 @@ static void posix_read (SLFile_FD_Type *f, SLang_Ref_Type *ref, unsigned int *nb
 
    bstr = SLbstring_create_malloced ((unsigned char *) b, len, 0);
    if (bstr != NULL)
-     {
-	if ((-1 != SLang_assign_to_ref (ref, SLANG_BSTRING_TYPE, (VOID_STAR)&bstr))
-	    && (-1 != SLang_push_uinteger (len)))
-	  return;
-
+     {	
+	if (-1 == SLang_assign_to_ref (ref, SLANG_BSTRING_TYPE, (VOID_STAR)&bstr))
+	  {
+	     SLbstring_free (bstr);
+	     return;
+	  }
 	SLbstring_free (bstr);
-	b = NULL;
-	/* drop */
+	(void) SLang_push_uinteger (len);
+	return;
      }
    
    return_error:
@@ -235,11 +282,10 @@ SLFile_FD_Type *SLfile_dup_fd (SLFile_FD_Type *f0)
 
    while (-1 == (fd = dup (fd0)))
      {
-#ifdef EINTR
-	if (errno == EINTR)
+	if (is_interrupt (errno))
 	  continue;
-#endif
-	_SLerrno_errno = errno;
+
+	_pSLerrno_errno = errno;
 	return NULL;
      }
    
@@ -278,8 +324,11 @@ void SLfile_free_fd (SLFile_FD_Type *f)
    if (f->fd != -1)
      {
 	if (f->close != NULL)
-	  (void) f->close (f->fd);
-
+	  {
+	     while ((-1 == f->close (f->fd))
+		    && is_interrupt (errno))
+	       ;
+	  }
 	f->fd = -1;
      }
 
@@ -342,9 +391,12 @@ static void posix_open (void)
      }
    SLang_free_slstring (file);
 
-   if (-1 == (f->fd = open (f->name, flags, mode)))
+   while (-1 == (f->fd = open (f->name, flags, mode)))
      {
-	_SLerrno_errno = errno;
+	if (is_interrupt (errno))
+	  continue;
+
+	_pSLerrno_errno = errno;
 	SLfile_free_fd (f);
 	SLang_push_null ();
 	return;
@@ -387,7 +439,7 @@ static void posix_fdopen (SLFile_FD_Type *f, char *mode)
 {
    if (f->stdio_mmt == NULL)
      {
-	if (-1 == _SLstdio_fdopen (f->name, f->fd, mode))
+	if (-1 == _pSLstdio_fdopen (f->name, f->fd, mode))
 	  return;
 
 	if (NULL == (f->stdio_mmt = SLang_pop_mmt (SLANG_FILE_PTR_TYPE)))
@@ -397,13 +449,19 @@ static void posix_fdopen (SLFile_FD_Type *f, char *mode)
    (void) SLang_push_mmt (f->stdio_mmt);
 }
 
-static long posix_lseek (SLFile_FD_Type *f, long ofs, int whence)
+static _pSLc_off_t_Type posix_lseek (SLFile_FD_Type *f, _pSLc_off_t_Type *ofs, int *whence)
 {
-   long status;
-   
-   if (-1 == (status = lseek (f->fd, ofs, whence)))
-     _SLerrno_errno = errno;
+   _pSLc_off_t_Type status;
    
+   while (-1 == (status = lseek (f->fd, *ofs, *whence)))
+     {
+	if (is_interrupt (errno))
+	  continue;
+
+	_pSLerrno_errno = errno;
+	return -1;
+     }
+
    return status;
 }
 
@@ -457,7 +515,7 @@ static SLang_Intrin_Fun_Type Fd_Name_Table [] =
    MAKE_INTRINSIC_0("isatty", posix_isatty, I),
    MAKE_INTRINSIC_0("open", posix_open, V),
    MAKE_INTRINSIC_3("read", posix_read, V, F, R, U),
-   MAKE_INTRINSIC_3("lseek", posix_lseek, L, F, L, I),
+   MAKE_INTRINSIC_3("lseek", posix_lseek, SLANG_C_OFF_T_TYPE, F, SLANG_C_OFF_T_TYPE, I),
    MAKE_INTRINSIC_2("fdopen", posix_fdopen, V, F, S),
    MAKE_INTRINSIC_2("write", posix_write, V, F, B),
    MAKE_INTRINSIC_1("dup_fd", posix_dup, V, F),
@@ -534,13 +592,13 @@ int SLfile_pop_fd (SLFile_FD_Type **f)
    return SLclass_pop_ptr_obj (SLANG_FILE_FD_TYPE, (VOID_STAR *) f);
 }
 
-static void destroy_fd_type (unsigned char type, VOID_STAR ptr)
+static void destroy_fd_type (SLtype type, VOID_STAR ptr)
 {
    (void) type;
    SLfile_free_fd (*(SLFile_FD_Type **) ptr);
 }
 
-static int fd_push (unsigned char type, VOID_STAR v)
+static int fd_push (SLtype type, VOID_STAR v)
 {
    (void) type;
    return SLfile_push_fd (*(SLFile_FD_Type **)v);
@@ -560,7 +618,7 @@ int SLang_init_posix_io (void)
 
    if ((-1 == SLadd_intrin_fun_table(Fd_Name_Table, "__POSIXIO__"))
        || (-1 == SLadd_iconstant_table (PosixIO_Consts, NULL))
-       || (-1 == _SLerrno_init ()))
+       || (-1 == _pSLerrno_init ()))
      return -1;
 
    return 0;
