@@ -1,27 +1,91 @@
 /* SLang_read_line interface --- uses SLang tty stuff */
-/* Copyright (c) 1992, 1999, 2001, 2002, 2003 John E. Davis
- * This file is part of the S-Lang library.
- *
- * You may distribute under the terms of either the GNU General Public
- * License or the Perl Artistic License.
- */
+/*
+Copyright (C) 2004, 2005, 2006 John E. Davis
+
+This file is part of the S-Lang Library.
+
+The S-Lang Library is free software; you can redistribute it and/or
+modify it under the terms of the GNU General Public License as
+published by the Free Software Foundation; either version 2 of the
+License, or (at your option) any later version.
+
+The S-Lang Library is distributed in the hope that it will be useful,
+but WITHOUT ANY WARRANTY; without even the implied warranty of
+MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+General Public License for more details.
+
+You should have received a copy of the GNU General Public License
+along with this library; if not, write to the Free Software
+Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
+USA.  
+*/
 
 #include "slinclud.h"
 
 #include "slang.h"
 #include "_slang.h"
 
-#ifdef REAL_UNIX_SYSTEM
-int SLang_RL_EOF_Char = 4;
-#else
-int SLang_RL_EOF_Char = 26;
-#endif
+typedef struct RL_History_Type
+{
+   struct RL_History_Type *prev, *next;
+   char *buf;
+   unsigned int len;
+   unsigned int point;
+} RL_History_Type;
+
+/* Maximum size of display */
+#define SLRL_DISPLAY_BUFFER_SIZE 1024
+
+struct _pSLrline_Type
+{
+   RL_History_Type *root, *tail, *last;
+   unsigned char *buf;		       /* edit buffer */
+   unsigned int buf_len;			       /* sizeof buffer */
+   unsigned int point;			       /* current editing point */
+   unsigned int tab;			       /* tab width */
+   unsigned int len;			       /* current line size */
+
+   /* display variables */
+   unsigned int edit_width;		       /* length of display field */
+   int curs_pos;			       /* current column */
+   int start_column;		       /* column offset of display */
+   unsigned int hscroll;		       /* amount to use for horiz scroll */
+   char *prompt;
+
+   FVOID_STAR last_fun;		       /* last function executed by rl */
+
+   /* These two contain an image of what is on the display */
+   unsigned char upd_buf1[SLRL_DISPLAY_BUFFER_SIZE];
+   unsigned char upd_buf2[SLRL_DISPLAY_BUFFER_SIZE];
+   unsigned char *old_upd, *new_upd;   /* pointers to previous two buffers */
+   int new_upd_len, old_upd_len;       /* length of output buffers */
+
+   SLKeyMap_List_Type *keymap;
+   int eof_char;
+
+   /* tty variables */
+   unsigned int flags;		       /*  */
+   int state;
+#define RLI_LINE_INVALID	0
+#define RLI_LINE_SET		1
+#define RLI_LINE_IN_PROGRESS	2
+#define RLI_LINE_READ		3
+
+   unsigned int (*getkey)(void);       /* getkey function -- required */
+   void (*tt_goto_column)(int);
+   void (*tt_insert)(char);
+   void (*update_hook)(SLrline_Type *rli,
+		       char *prompt, char *buf, unsigned int len, unsigned int point,
+		       VOID_STAR client_data);
+   VOID_STAR update_client_data;
+   /* This function is only called when blinking matches */
+   int (*input_pending)(int);
+};
 
 int SLang_Rline_Quit;
-static SLang_RLine_Info_Type *This_RLI;
 
 static unsigned char Char_Widths[256];
-static void position_cursor (int);
+static void position_cursor (SLrline_Type *, int);
 
 static void rl_beep (void)
 {
@@ -29,135 +93,167 @@ static void rl_beep (void)
    fflush (stdout);
 }
 
-/* editing functions */
-static int rl_bol (void)
+static int check_space (SLrline_Type *This_RLI, unsigned int dn)
 {
-   if (This_RLI->point == 0) return 0;
-   This_RLI->point = 0;
-   return 1;
-}
+   unsigned char *new_buf;
+   unsigned int new_len;
 
-static int rl_eol (void)
-{
-   if (This_RLI->point == This_RLI->len) return 0;
-   This_RLI->point = This_RLI->len;
-   return 1;
-}
+   new_len = 1 + This_RLI->len + dn;
 
-static int rl_right (void)
-{
-   if (This_RLI->point == This_RLI->len) return 0;
-   This_RLI->point++;
-   return 1;
-}
+   if (new_len <= This_RLI->buf_len)
+     return 0;
 
-static int rl_left (void)
-{
-   if (This_RLI->point == 0) return 0;
-   This_RLI->point--;
-   return 1;
-}
-
-static int rl_self_insert (void)
-{
-   unsigned char *pmin, *p;
-
-   if (This_RLI->len == This_RLI->buf_len)
-     {
-	rl_beep ();
-	return 0;
-     }
-
-   pmin = This_RLI->buf + This_RLI->point;
-   p = This_RLI->buf + This_RLI->len;
-   while (p > pmin)
-     {
-	*p = *(p - 1);
-	p--;
-     }
-   *pmin = SLang_Last_Key_Char;
-
-   This_RLI->len++;
-   This_RLI->point++;
-   if ((This_RLI->curs_pos + 2 >= This_RLI->edit_width)
-       || (This_RLI->tt_insert == NULL)
-       || (Char_Widths[SLang_Last_Key_Char] != 1)) return 1;
-
-   (*This_RLI->tt_insert)((char) SLang_Last_Key_Char);
-   /* update screen buf */
-   p = This_RLI->old_upd + (This_RLI->len - 1);
-   pmin = This_RLI->old_upd + (This_RLI->point - 1);
-   while (p > pmin)
-     {
-	*p = *(p - 1);
-	p--;
-     }
-   *pmin = SLang_Last_Key_Char;
+   if (NULL == (new_buf = (unsigned char *) SLrealloc ((char *)This_RLI->buf, new_len)))
+     return -1;
+	
+   This_RLI->buf_len = new_len;
+   This_RLI->buf = new_buf;
    return 0;
 }
 
-int SLang_rline_insert (char *s)
+
+/* editing functions */
+static int rl_bol (SLrline_Type *This_RLI)
 {
-   unsigned char *pmin, *p;
-   int n;
+   This_RLI->point = 0;
+   return 0;
+}
 
-   n = strlen (s);
-   if (n > This_RLI->buf_len - This_RLI->len)
-     n = This_RLI->buf_len - This_RLI->len;
+static int rl_eol (SLrline_Type *This_RLI)
+{
+   This_RLI->point = This_RLI->len;
+   return 0;
+}
 
-   if (n == 0) return 0;
+static int rl_right (SLrline_Type *This_RLI)
+{
+   SLuchar_Type *s, *smax;
+   int ignore_combining = 1;
+
+   s = This_RLI->buf + This_RLI->point;
+   smax = This_RLI->buf + This_RLI->len;
+
+   if (s < smax)
+     {
+	if (This_RLI->flags & SL_RLINE_UTF8_MODE)
+	  s = SLutf8_skip_chars (s, smax, 1, NULL, ignore_combining);
+	else
+	  s++;
+
+	This_RLI->point = s - This_RLI->buf;
+     }
+
+   return 0;
+}
+
+static int rl_left (SLrline_Type *This_RLI)
+{
+   SLuchar_Type *s, *smin;
+   int ignore_combining = 1;
+
+   smin = This_RLI->buf;
+   s = smin + This_RLI->point;
+
+   if (s > smin)
+     {
+	if (This_RLI->flags & SL_RLINE_UTF8_MODE)
+	  s = SLutf8_bskip_chars (smin, s, 1, NULL, ignore_combining);
+	else
+	  s--;
+
+	This_RLI->point = s - This_RLI->buf;
+     }
+
+   return 0;
+}
+
+int SLrline_ins (SLrline_Type *This_RLI, char *s, unsigned int n)
+{
+   unsigned char *pmin;
+
+   if (-1 == check_space (This_RLI, n + 128))
+     return -1;
 
    pmin = This_RLI->buf + This_RLI->point;
-   p = This_RLI->buf + (This_RLI->len - 1);
-
-   while (p >= pmin)
+   if (This_RLI->len)
      {
-	*(p + n) = *p;
-	p--;
+	unsigned char *p = This_RLI->buf + This_RLI->len;
+	while (p >= pmin)
+	  {
+	     *(p + n) = *p;
+	     p--;
+	  }
      }
-   SLMEMCPY ((char *) pmin, s, n);
+   memcpy ((char *) pmin, s, n);
 
    This_RLI->len += n;
    This_RLI->point += n;
    return n;
 }
 
-static int rl_deln (int n)
+
+static int rl_self_insert (SLrline_Type *This_RLI)
 {
-   unsigned char *pmax, *p;
+   char buf[8];
+
+   buf[0] = SLang_Last_Key_Char;
+   buf[1] = 0;
+
+   return SLrline_ins (This_RLI, buf, 1);
+}
+
+int SLrline_del (SLrline_Type *This_RLI, unsigned int n)
+{
+   SLuchar_Type *pmax, *p, *pn;
+   int ignore_combining = 1;
 
    p = This_RLI->buf + This_RLI->point;
    pmax = This_RLI->buf + This_RLI->len;
-
-   if (p + n > pmax) n = (int) (pmax - p);
-   while (p < pmax)
+   
+   if (This_RLI->flags & SL_RLINE_UTF8_MODE)
      {
-	*p = *(p + n);
-	p++;
+	pn = SLutf8_skip_chars (p, pmax, n, NULL, ignore_combining);
+	n = pn - p;
+     }
+   else
+     {
+	if (p + n > pmax) n = (pmax - p);
+	pn = p + n;
      }
    This_RLI->len -= n;
-   return n;
+
+   while (pn < pmax)
+     {
+	*p++ = *pn++;
+     }
+   return 0;
 }
 
-static int rl_del (void)
+static int rl_del (SLrline_Type *This_RLI)
 {
-   return rl_deln(1);
+   return SLrline_del (This_RLI, 1);
 }
 
-static int rl_quote_insert (void)
+static int rl_quote_insert (SLrline_Type *This_RLI)
 {
-   int err = SLang_Error;
-   SLang_Error = 0;
+   /* FIXME.  We should not be messing with SLang_Error here */
+   int err = _pSLang_Error;
+   _pSLang_Error = 0;
    SLang_Last_Key_Char = (*This_RLI->getkey)();
-   rl_self_insert ();
-   if (SLang_Error == SL_USER_BREAK) SLang_Error = 0;
-   else SLang_Error = err;
-   return 1;
+   rl_self_insert (This_RLI);
+   if (_pSLang_Error == SL_USER_BREAK) 
+     {
+	SLKeyBoard_Quit = 0;
+	_pSLang_Error = 0;
+     }
+   else _pSLang_Error = err;
+   return 0;
 }
 
-static int rl_trim (void)
+static int rl_trim (SLrline_Type *This_RLI)
 {
    unsigned char *p, *pmax, *p1;
+
    p = This_RLI->buf + This_RLI->point;
    pmax = This_RLI->buf + This_RLI->len;
 
@@ -178,16 +274,20 @@ static int rl_trim (void)
    p++;
 
    This_RLI->point = (int) (p - p1);
-   return rl_deln ((int) (pmax - p));
+   return SLrline_del (This_RLI, (int) (pmax - p));
 }
 
-static int rl_bdel (void)
+static int rl_bdel (SLrline_Type *This_RLI)
 {
-   if (rl_left()) return rl_del();
+   if (This_RLI->point)
+     {
+	rl_left (This_RLI);
+	rl_del(This_RLI);
+     }
    return 0;
 }
 
-static int rl_deleol (void)
+static int rl_deleol (SLrline_Type *This_RLI)
 {
    if (This_RLI->point == This_RLI->len) return 0;
    *(This_RLI->buf + This_RLI->point) = 0;
@@ -195,19 +295,21 @@ static int rl_deleol (void)
    return 1;
 }
 
-static int rl_delete_line (void)
+static int rl_delete_line (SLrline_Type *This_RLI)
 {
-   This_RLI->point = 0;
-   *(This_RLI->buf + This_RLI->point) = 0;
-   This_RLI->len = 0;
-   return 1;
+   rl_bol (This_RLI);
+   rl_deleol (This_RLI);
+   return 0;
 }
 
-static int rl_enter (void)
+static int rl_enter (SLrline_Type *This_RLI)
 {
+   if (-1 == check_space (This_RLI, 1))
+     return -1;
+
    *(This_RLI->buf + This_RLI->len) = 0;
    SLang_Rline_Quit = 1;
-   return 1;
+   return 0;
 }
 
 static SLKeyMap_List_Type *RL_Keymap;
@@ -217,8 +319,7 @@ static SLKeyMap_List_Type *RL_Keymap;
  * There is a hook so the user can provide a more sophisticated update if
  * necessary.
  */
-
-static void position_cursor (int col)
+static void position_cursor (SLrline_Type *This_RLI, int col)
 {
    unsigned char *p, *pmax;
    int dc;
@@ -262,7 +363,7 @@ static void position_cursor (int col)
    fflush (stdout);
 }
 
-static void erase_eol (SLang_RLine_Info_Type *rli)
+static void erase_eol (SLrline_Type *rli)
 {
    unsigned char *p, *pmax;
 
@@ -274,45 +375,38 @@ static void erase_eol (SLang_RLine_Info_Type *rli)
    rli->curs_pos = rli->old_upd_len;
 }
 
-static unsigned char *spit_out(SLang_RLine_Info_Type *rli, unsigned char *p)
+static unsigned char *spit_out(SLrline_Type *rli, unsigned char *p)
 {
    unsigned char *pmax;
-   position_cursor ((int) (p - rli->new_upd));
+   position_cursor (rli, (int) (p - rli->new_upd));
    pmax = rli->new_upd + rli->new_upd_len;
    while (p < pmax) putc((char) *p++, stdout);
    rli->curs_pos = rli->new_upd_len;
    return pmax;
 }
 
-static void really_update (SLang_RLine_Info_Type *rli, int new_curs_position)
+static void really_update (SLrline_Type *rli, int new_curs_position)
 {
    unsigned char *b = rli->old_upd, *p = rli->new_upd, chb, chp;
    unsigned char *pmax;
 
-   if (rli->update_hook != NULL)
+   pmax = p + rli->edit_width;
+   while (p < pmax)
      {
-	(*rli->update_hook)(p, rli->edit_width, new_curs_position);
-     }
-   else
-     {
-	pmax = p + rli->edit_width;
-	while (p < pmax)
+	chb = *b++; chp = *p++;
+	if (chb == chp) continue;
+	
+	if (rli->old_upd_len <= rli->new_upd_len)
 	  {
-	     chb = *b++; chp = *p++;
-	     if (chb == chp) continue;
-
-	     if (rli->old_upd_len <= rli->new_upd_len)
-	       {
-		  /* easy one */
-		  (void) spit_out (rli, p - 1);
-		  break;
-	       }
-	     spit_out(rli, p - 1);
-	     erase_eol (rli);
+	     /* easy one */
+	     (void) spit_out (rli, p - 1);
 	     break;
 	  }
-	position_cursor (new_curs_position);
+	spit_out(rli, p - 1);
+	erase_eol (rli);
+	break;
      }
+   position_cursor (rli, new_curs_position);
 
    /* update finished, so swap */
 
@@ -322,7 +416,7 @@ static void really_update (SLang_RLine_Info_Type *rli, int new_curs_position)
    rli->new_upd = p;
 }
 
-static void RLupdate (SLang_RLine_Info_Type *rli)
+static void RLupdate (SLrline_Type *rli)
 {
    int len, dlen, start_len, prompt_len = 0, tw = 0, count;
    int want_cursor_pos;
@@ -330,9 +424,18 @@ static void RLupdate (SLang_RLine_Info_Type *rli)
    int no_echo;
 
    no_echo = rli->flags & SL_RLINE_NO_ECHO;
+   *(rli->buf + rli->len) = 0;
+   
+   if (rli->update_hook != NULL)
+     {
+	if (no_echo)
+	  (*rli->update_hook) (rli, rli->prompt, "", 0, 0, rli->update_client_data);
+	else
+	  (*rli->update_hook) (rli, rli->prompt, (char *)rli->buf, rli->len, rli->point, rli->update_client_data);
+	return;
+     }
 
    b_point = (unsigned char *) (rli->buf + rli->point);
-   *(rli->buf + rli->len) = 0;
 
    /* expand characters for output buffer --- handle prompt first.
     * Do two passes --- first to find out where to begin upon horiz
@@ -365,11 +468,11 @@ static void RLupdate (SLang_RLine_Info_Type *rli)
 	if (count == 1) prompt_len = len;
      }
 
-   if (len < rli->edit_width - rli->dhscroll) start_len = 0;
-   else if ((rli->start_column > len)
-	    || (rli->start_column + rli->edit_width <= len))
+   if (len + rli->hscroll < rli->edit_width) start_len = 0;
+   else if ((rli->start_column > (int)len)
+	    || (rli->start_column + (int)rli->edit_width <= len))
      {
-	start_len = len - (rli->edit_width - rli->dhscroll);
+	start_len = len - (rli->edit_width - rli->hscroll);
 	if (start_len < 0) start_len = 0;
      }
    else start_len = rli->start_column;
@@ -408,7 +511,7 @@ static void RLupdate (SLang_RLine_Info_Type *rli)
 	if ((count == 0) && (no_echo))
 	  break;
 
-	while ((len < rli->edit_width) && ((chb = *b++) != 0))
+	while ((len < (int)rli->edit_width) && ((chb = *b++) != 0))
 	  {
 	     dlen = Char_Widths[chb];
 	     if (dlen == 1) *p++ = chb;
@@ -418,7 +521,7 @@ static void RLupdate (SLang_RLine_Info_Type *rli)
 		    {
 		       dlen = tw * ((len + start_len - prompt_len) / tw + 1) - (len + start_len - prompt_len);
 		       len += dlen;	       /* ok since dlen comes out 0  */
-		       if (len > rli->edit_width) dlen = len - rli->edit_width;
+		       if (len > (int)rli->edit_width) dlen = len - rli->edit_width;
 		       while (dlen--) *p++ = ' ';
 		       dlen = 0;
 		    }
@@ -447,31 +550,31 @@ static void RLupdate (SLang_RLine_Info_Type *rli)
    really_update (rli, want_cursor_pos);
 }
 
-void SLrline_redraw (SLang_RLine_Info_Type *rli)
+void SLrline_redraw (SLrline_Type *rli)
 {
-   unsigned char *p = rli->new_upd;
-   unsigned char *pmax = p + rli->edit_width;
+   unsigned char *p;
+   unsigned char *pmax;
+   
+   if (rli == NULL)
+     return;
+
+   p = rli->new_upd;
+   pmax = p + rli->edit_width;
    while (p < pmax) *p++ = ' ';
    rli->new_upd_len = rli->edit_width;
    really_update (rli, 0);
    RLupdate (rli);
 }
 
-static int rl_eof_insert (void)
+static int rl_eof_insert (SLrline_Type *This_RLI)
 {
-   if (This_RLI->len == 0)
-     {
-	SLang_Last_Key_Char = SLang_RL_EOF_Char;
-	/* rl_self_insert (); */
-	return rl_enter ();
-     }
-   return 0;
+   return rl_enter (This_RLI);
 }
 
 /* This is very naive.  It knows very little about nesting and nothing
  * about quoting.
  */
-static void blink_match (SLang_RLine_Info_Type *rli)
+static void blink_match (SLrline_Type *rli)
 {
    unsigned char bra, ket;
    unsigned int delta_column;
@@ -540,77 +643,122 @@ static void blink_match (SLang_RLine_Info_Type *rli)
      }
 }
 
-int SLang_read_line (SLang_RLine_Info_Type *rli)
+char *SLrline_read_line (SLrline_Type *rli, char *prompt, unsigned int *lenp)
 {
    unsigned char *p, *pmax;
    SLang_Key_Type *key;
+   int last_input_char;
+   unsigned int dummy_len_buf;
+
+   if (lenp == NULL)
+     lenp = &dummy_len_buf;
+
+   *lenp = 0;
+
+   if (rli == NULL)
+     return NULL;
+
+   if (rli->state == RLI_LINE_IN_PROGRESS)
+     {
+	*lenp = 0;
+	return NULL;
+     }
+
+   if (prompt == NULL)
+     prompt = "";
+
+   if ((rli->prompt == NULL)
+       || strcmp (rli->prompt, prompt))
+     {
+	if (NULL == (prompt = SLmake_string (prompt)))
+	  return NULL;
+	
+	SLfree (rli->prompt);
+	rli->prompt = prompt;
+     }
 
    SLang_Rline_Quit = 0;
-   This_RLI = rli;
    p = rli->old_upd; pmax = p + rli->edit_width;
    while (p < pmax) *p++ = ' ';
 
-   /* Sanity checking */
-   rli->len = strlen ((char *) rli->buf);
-   if (rli->len >= rli->buf_len)
+   if (rli->state != RLI_LINE_SET)
      {
 	rli->len = 0;
+	rli->point = 0;
 	*rli->buf = 0;
      }
-   if (rli->point > rli->len) rli->point = rli->len;
-   if (rli->point < 0) rli->point = 0;
+   rli->state = RLI_LINE_IN_PROGRESS;
 
    rli->curs_pos = rli->start_column = 0;
    rli->new_upd_len = rli->old_upd_len = 0;
 
-   This_RLI->last_fun = NULL;
+   rli->last_fun = NULL;
    if (rli->update_hook == NULL)
      putc ('\r', stdout);
 
    RLupdate (rli);
 
+   last_input_char = 0;
    while (1)
      {
 	key = SLang_do_key (RL_Keymap, (int (*)(void)) rli->getkey);
 
 	if ((key == NULL) || (key->f.f == NULL))
-	  rl_beep ();
+	  {
+	     rl_beep ();
+	     continue;
+	  }
+
+	if ((*key->str != 2) || (key->str[1] != rli->eof_char))
+	  last_input_char = 0;
 	else
 	  {
-	     if ((SLang_Last_Key_Char == SLang_RL_EOF_Char)
-		 && (*key->str == 2)
-		 && (This_RLI->len == 0))
-	       rl_eof_insert ();
-	     else if (key->type == SLKEY_F_INTRINSIC)
+	     if ((rli->len == 0) && (last_input_char != rli->eof_char))
 	       {
-		  if ((key->f.f)())
-		    RLupdate (rli);
-
-		  if ((rli->flags & SL_RLINE_BLINK_MATCH)
-		      && (rli->input_pending != NULL))
-		    blink_match (rli);
+		  rli->buf[rli->len] = 0;
+		  rli->state = RLI_LINE_READ;
+		  *lenp = 0;
+		  return NULL;	       /* EOF */
 	       }
+	     
+	     last_input_char = rli->eof_char;
+	  }
 
-	     if (SLang_Rline_Quit)
+	if (key->type == SLKEY_F_INTRINSIC)
+	  {
+	     int (*func)(SLrline_Type *);
+	     func = (int (*)(SLrline_Type *)) key->f.f;
+	     
+	     (void) (*func)(rli);
+
+	     RLupdate (rli);
+	     
+	     if ((rli->flags & SL_RLINE_BLINK_MATCH)
+		 && (rli->input_pending != NULL))
+	       blink_match (rli);
+	  }
+	
+	if ((SLang_Rline_Quit) || (_pSLang_Error == SL_USER_BREAK))
+	  {
+	     if (_pSLang_Error == SL_USER_BREAK)
 	       {
-		  This_RLI->buf[This_RLI->len] = 0;
-		  if (SLang_Error == SL_USER_BREAK)
-		    {
-		       SLang_Error = 0;
-		       return -1;
-		    }
-		  return This_RLI->len;
+		  rli->len = 0;
 	       }
+	     rli->buf[rli->len] = 0;
+	     rli->state = RLI_LINE_READ;
+	     *lenp = rli->len;
+
+	     return SLmake_nstring ((char *)rli->buf, rli->len);
 	  }
 	if (key != NULL)
-	  This_RLI->last_fun = key->f.f;
+	  rli->last_fun = key->f.f;
      }
 }
 
-static int rl_abort (void)
+static int rl_abort (SLrline_Type *This_RLI)
 {
-   rl_delete_line ();
-   return rl_enter ();
+   rl_delete_line (This_RLI);
+   return rl_enter (This_RLI);
 }
 
 /* TTY interface --- ANSI */
@@ -621,16 +769,25 @@ static void ansi_goto_column (int n)
    if (n) fprintf(stdout, "\033[%dC", n);
 }
 
-static void rl_select_line (SLang_Read_Line_Type *p)
+static int rl_select_line (SLrline_Type *rli, RL_History_Type *p)
 {
-   This_RLI->last = p;
-   strcpy ((char *) This_RLI->buf, (char *) p->buf);
-   This_RLI->point = This_RLI->len = strlen((char *) p->buf);
+   unsigned int len;
+
+   len = p->len;
+   if (-1 == check_space (rli, len))
+     return -1;
+
+   rli->last = p;
+   strcpy ((char *) rli->buf, p->buf);
+   rli->point = p->point;
+   rli->len = len;
+   return 0;
 }
-static int rl_next_line (void);
-static int rl_prev_line (void)
+
+static int rl_next_line (SLrline_Type *This_RLI);
+static int rl_prev_line (SLrline_Type *This_RLI)
 {
-   SLang_Read_Line_Type *prev;
+   RL_History_Type *prev;
 
    if (((This_RLI->last_fun != (FVOID_STAR) rl_prev_line)
 	&& (This_RLI->last_fun != (FVOID_STAR) rl_next_line))
@@ -646,18 +803,18 @@ static int rl_prev_line (void)
 	return 0;
      }
 
-   rl_select_line (prev);
-   return 1;
+   return rl_select_line (This_RLI, prev);
 }
-static int rl_redraw (void)
+
+static int rl_redraw (SLrline_Type *This_RLI)
 {
    SLrline_redraw (This_RLI);
    return 1;
 }
 
-static int rl_next_line (void)
+static int rl_next_line (SLrline_Type *This_RLI)
 {
-   SLang_Read_Line_Type *next;
+   RL_History_Type *next;
 
    if (((This_RLI->last_fun != (FVOID_STAR) rl_prev_line)
 	&& (This_RLI->last_fun != (FVOID_STAR) rl_next_line))
@@ -675,38 +832,89 @@ static int rl_next_line (void)
 	*This_RLI->buf = 0;
 	This_RLI->last = NULL;
      }
-   else rl_select_line (next);
+   else rl_select_line (This_RLI, next);
    return 1;
 }
 
+#define AKEY(name,func) {name,(int (*)(void))func}
+
 static SLKeymap_Function_Type SLReadLine_Functions[] =
 {
-   {"up", rl_prev_line},
-   {"down", rl_next_line},
-   {"bol", rl_bol},
-   {"eol", rl_eol},
-   {"right", rl_right},
-   {"left", rl_left},
-   {"self_insert", rl_self_insert},
-   {"bdel", rl_bdel},
-   {"del", rl_del},
-   {"deleol", rl_deleol},
-   {"enter", rl_enter},
-   {"trim", rl_trim},
-   {"quoted_insert", rl_quote_insert},
-   {(char *) NULL, NULL}
+   AKEY("up", rl_prev_line),
+   AKEY("down", rl_next_line),
+   AKEY("bol", rl_bol),
+   AKEY("eol", rl_eol),
+   AKEY("right", rl_right),
+   AKEY("left", rl_left),
+   AKEY("self_insert", rl_self_insert),
+   AKEY("bdel", rl_bdel),
+   AKEY("del", rl_del),
+   AKEY("deleol", rl_deleol),
+   AKEY("enter", rl_enter),
+   AKEY("trim", rl_trim),
+   AKEY("quoted_insert", rl_quote_insert),
+   AKEY(NULL, NULL),
 };
 
-int SLang_init_readline (SLang_RLine_Info_Type *rli)
+void SLrline_close (SLrline_Type *rli)
+{
+   if (rli == NULL)
+     return;
+   
+   SLfree (rli->prompt);
+   SLfree ((char *)rli->buf);
+   SLfree ((char *)rli);
+}
+
+SLrline_Type *SLrline_open (unsigned int width, unsigned int flags)
 {
    int ch;
    char simple[2];
+   SLrline_Type *rli;
+
+   if (NULL == (rli = (SLrline_Type *)SLcalloc (1, sizeof (SLrline_Type))))
+     return NULL;
+   
+   if (width == 0)
+     width = 80;
+
+   if (width < 256) rli->buf_len = 256;
+   else rli->buf_len = width;
+
+   if (NULL == (rli->buf = (unsigned char *)SLmalloc (rli->buf_len)))
+     {
+	SLrline_close (rli);
+	return NULL;
+     }
+   *rli->buf = 0;
+#ifdef REAL_UNIX_SYSTEM
+   rli->eof_char = 4;
+#else
+   rli->eof_char = 26;
+#endif
+   
+   rli->point = 0;
+   rli->flags = flags;
+   rli->edit_width = width;
+   rli->hscroll = width/4;
+   rli->tab = 8;
+   rli->getkey = SLang_getkey;
+   rli->input_pending = SLang_input_pending;
+   rli->state = RLI_LINE_INVALID;
+
+   if (rli->flags & SL_RLINE_USE_ANSI)
+     {
+	if (rli->tt_goto_column == NULL) rli->tt_goto_column = ansi_goto_column;
+     }
 
    if (RL_Keymap == NULL)
      {
 	simple[1] = 0;
 	if (NULL == (RL_Keymap = SLang_create_keymap ("ReadLine", NULL)))
-	  return -1;
+	  {
+	     SLrline_close (rli);
+	     return NULL;
+	  }
 
 	RL_Keymap->functions = SLReadLine_Functions;
 
@@ -730,7 +938,7 @@ int SLang_init_readline (SLang_RLine_Info_Type *rli)
 
 	simple[0] = SLang_Abort_Char;
 	SLkm_define_key (simple, (FVOID_STAR) rl_abort, RL_Keymap);
-	simple[0] = SLang_RL_EOF_Char;
+	simple[0] = (char) rli->eof_char;
 	SLkm_define_key (simple, (FVOID_STAR) rl_eof_insert, RL_Keymap);
 
 #ifndef IBMPC_SYSTEM
@@ -779,58 +987,201 @@ int SLang_init_readline (SLang_RLine_Info_Type *rli)
 	SLkm_define_key  ("^R", (FVOID_STAR) rl_redraw, RL_Keymap);
 	SLkm_define_key  ("`", (FVOID_STAR) rl_quote_insert, RL_Keymap);
 	SLkm_define_key  ("\033\\", (FVOID_STAR) rl_trim, RL_Keymap);
-	if (SLang_Error) return -1;
+	if (_pSLang_Error)
+	  {
+	     SLrline_close (rli);
+	     return NULL;
+	  }
      }
 
-   if (rli->prompt == NULL) rli->prompt = "";
    if (rli->keymap == NULL) rli->keymap = RL_Keymap;
    rli->old_upd = rli->upd_buf1;
    rli->new_upd = rli->upd_buf2;
-   *rli->buf = 0;
-   rli->point = 0;
 
-   if (rli->flags & SL_RLINE_USE_ANSI)
+   if (Char_Widths[0] == 0)
      {
-	if (rli->tt_goto_column == NULL) rli->tt_goto_column = ansi_goto_column;
+	/* FIXME: This does not support UTF-8 */
+	for (ch = 0; ch < 32; ch++) Char_Widths[ch] = 2;
+	for (ch = 32; ch < 256; ch++) Char_Widths[ch] = 1;
+	Char_Widths[127] = 2;
+#ifndef IBMPC_SYSTEM
+	for (ch = 128; ch < 160; ch++) Char_Widths[ch] = 3;
+#endif
      }
 
-   if (Char_Widths[0] == 2) return 0;
+   return rli;
+}
 
-   for (ch = 0; ch < 32; ch++) Char_Widths[ch] = 2;
-   for (ch = 32; ch < 256; ch++) Char_Widths[ch] = 1;
-   Char_Widths[127] = 2;
-#ifndef IBMPC_SYSTEM
-   for (ch = 128; ch < 160; ch++) Char_Widths[ch] = 3;
-#endif
+int SLrline_add_to_history (SLrline_Type *rli, char *hist)
+{
+   RL_History_Type *h;
+
+   if ((rli == NULL) || (hist == NULL))
+     return -1;
+
+   if (NULL == (h = (RL_History_Type *) SLcalloc (1, sizeof (RL_History_Type)))
+       || (NULL == (h->buf = SLmake_string (hist))))
+     {
+	SLfree ((char *)h);
+	return -1;
+     }
+
+   if (rli->tail != NULL)
+     rli->tail->next = h;
+   
+   h->prev = rli->tail;
+   rli->tail = h;
+   h->next = NULL;
+   
+   h->point = h->len = strlen (hist);
+   return 0;
+}
+   
+  
+int SLrline_save_line (SLrline_Type *rli)
+{
+   if (rli == NULL)
+     return -1;
+   
+   return SLrline_add_to_history (rli, (char *) rli->buf);
+}
+
+SLkeymap_Type *SLrline_get_keymap (SLrline_Type *rli)
+{
+   if (rli == NULL)
+     return NULL;
+   return rli->keymap;
+}
+
+int SLrline_set_update_hook (SLrline_Type *rli, 
+			     void (*fun)(SLrline_Type *, char *, char *, unsigned int, unsigned int, VOID_STAR),
+			     VOID_STAR client_data)
+{
+   if (rli == NULL)
+     return -1;
+   
+   rli->update_hook = fun;
+   rli->update_client_data = client_data;
+   return 0;
+}
+
+char *SLrline_get_line (SLrline_Type *rli)
+{
+   if (rli == NULL)
+     return NULL;
+   return SLmake_nstring ((char *)rli->buf, rli->len);
+}
+
+int SLrline_get_point (SLrline_Type *rli, unsigned int *pointp)
+{
+   if (rli == NULL)
+     return -1;
+   *pointp = rli->point;
+   return 0;
+}
+
+int SLrline_set_point (SLrline_Type *rli, unsigned int point)
+{
+   if (rli == NULL)
+     return -1;
+   
+   if (rli->state == RLI_LINE_INVALID)
+     return -1;
+
+   if (rli->len > point)
+     point = rli->len;
+   
+   rli->point = point;
+   return 0;
+}
+
+int SLrline_get_tab (SLrline_Type *rli, unsigned int *p)
+{
+   if (rli == NULL)
+     return -1;
+   *p = rli->tab;
+   return 0;
+}
+
+int SLrline_set_tab (SLrline_Type *rli, unsigned int tab)
+{
+   if (rli == NULL)
+     return -1;
+   
+   rli->tab = tab;
+   return 0;
+}
+
+int SLrline_get_hscroll (SLrline_Type *rli, unsigned int *p)
+{
+   if (rli == NULL)
+     return -1;
+   *p = rli->hscroll;
+   return 0;
+}
+
+int SLrline_set_hscroll (SLrline_Type *rli, unsigned int p)
+{
+   if (rli == NULL)
+     return -1;
+   rli->hscroll = p;
+   return 0;
+}
+
+int SLrline_set_line (SLrline_Type *rli, char *buf)
+{
+   unsigned int len;
+
+   if (rli == NULL)
+     return -1;
+
+   if (buf == NULL)
+     buf = "";
+   
+   len = strlen (buf);
+
+   buf = SLmake_string (buf);
+   if (buf == NULL)
+     return -1;
+   
+   SLfree ((char *)rli->buf);
+   rli->buf = (unsigned char *)buf;
+   rli->buf_len = len;
+
+   rli->point = len;
+   rli->len = len;
+   
+   rli->state = RLI_LINE_SET;
+   return 0;
+}
+
+int SLrline_set_echo (SLrline_Type *rli, int state)
+{
+   if (rli == NULL)
+     return -1;
+
+   if (state == 0)
+     rli->flags |= SL_RLINE_NO_ECHO;
+   else
+     rli->flags &= ~SL_RLINE_NO_ECHO;
 
    return 0;
 }
 
-SLang_Read_Line_Type *SLang_rline_save_line (SLang_RLine_Info_Type *rli)
+int SLrline_get_echo (SLrline_Type *rli, int *statep)
 {
-   SLang_Read_Line_Type *rl;
-   unsigned char *buf;
-
-   if ((rli == NULL) || (rli->buf == NULL))
-     return NULL;
-
-   if (NULL == (rl = (SLang_Read_Line_Type *) SLmalloc (sizeof (SLang_Read_Line_Type)))
-       || (NULL == (buf = (unsigned char *) SLmake_string ((char *)rli->buf))))
-     {
-	SLfree ((char *)rl);
-	return NULL;
-     }
-   rl->buf = buf;
-   rl->buf_len = strlen ((char *)buf);
-   rl->num = rl->misc = 0;
-   rl->next = rl->prev = NULL;
-
-   if (rli->tail != NULL)
-     {
-	rli->tail->next = rl;
-	rl->prev = rli->tail;
-     }
-   rli->tail = rl;
-
-   return rl;
+   if (rli == NULL)
+     return -1;
+   
+   *statep = (0 == (rli->flags & SL_RLINE_NO_ECHO));
+   return 0;
+}
+
+int SLrline_set_display_width (SLrline_Type *rli, unsigned int w)
+{
+   if (rli == NULL)
+     return -1;
+   
+   rli->edit_width = w;
+   return 0;
 }
