@@ -1,10 +1,24 @@
 /* file stdio intrinsics for S-Lang */
-/* Copyright (c) 1992, 1999, 2001, 2002, 2003 John E. Davis
- * This file is part of the S-Lang library.
- *
- * You may distribute under the terms of either the GNU General Public
- * License or the Perl Artistic License.
- */
+/*
+Copyright (C) 2004, 2005, 2006 John E. Davis
+
+This file is part of the S-Lang Library.
+
+The S-Lang Library is free software; you can redistribute it and/or
+modify it under the terms of the GNU General Public License as
+published by the Free Software Foundation; either version 2 of the
+License, or (at your option) any later version.
+
+The S-Lang Library is distributed in the hope that it will be useful,
+but WITHOUT ANY WARRANTY; without even the implied warranty of
+MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+General Public License for more details.
+
+You should have received a copy of the GNU General Public License
+along with this library; if not, write to the Free Software
+Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
+USA.  
+*/
 
 #include "slinclud.h"
 
@@ -47,6 +61,14 @@
 #include "slang.h"
 #include "_slang.h"
 
+#ifdef HAVE_FSEEKO
+# define FSEEK(a,b,c) fseeko(a,b,c)
+# define FTELL(a) ftello(a)
+#else
+# define FSEEK(a,b,c) fseek(a,b,c)
+# define FTELL(a) ftell(a)
+#endif
+
 typedef struct
 {
    FILE *fp;			       /* kind of obvious */
@@ -138,7 +160,7 @@ static int open_file_type (char *file, int fd, char *mode,
 
    if (fp == NULL)
      {
-	_SLerrno_errno = errno;
+	_pSLerrno_errno = errno;
 	goto return_error;
      }
 
@@ -181,11 +203,11 @@ static void stdio_fopen (char *file, char *mode)
    (void) open_file_type (file, -1, mode, fopen_fun, fclose_fun, 0);
 }
 
-int _SLstdio_fdopen (char *file, int fd, char *mode)
+int _pSLstdio_fdopen (char *file, int fd, char *mode)
 {
    if (fd == -1)
      {
-	_SLerrno_errno = EBADF;
+	_pSLerrno_errno = EBADF;
 	(void) SLang_push_null ();
 	return -1;
      }
@@ -227,6 +249,9 @@ static SLang_MMT_Type *pop_fp (unsigned int flags, FILE **fp_ptr)
        && (NULL != (*fp_ptr = t->fp)))
      return mmt;
 
+#ifdef EBADF
+   _pSLerrno_errno = EBADF;
+#endif
    SLang_free_mmt (mmt);
    return NULL;
 }
@@ -252,16 +277,27 @@ char *SLang_get_name_from_fileptr (SLang_MMT_Type *mmt)
 int SLang_pop_fileptr (SLang_MMT_Type **mmt, FILE **fp)
 {
    if (NULL == (*mmt = pop_fp (0xFFFF, fp)))
-     {
-#ifdef EBADF
-	_SLerrno_errno = EBADF;
-#endif
-	return -1;
-     }
+     return -1;
 
    return 0;
 }
 
+int SLang_get_fileptr (SLang_MMT_Type *mmt, FILE **fpp)
+{
+   SL_File_Table_Type *t;
+
+   *fpp = NULL;
+   if (mmt == NULL)
+     return -1;
+
+   t = (SL_File_Table_Type *) SLang_object_from_mmt (mmt);
+   if (t == NULL)
+     return -1;
+
+   *fpp = t->fp;
+   return 0;
+}
+
 static int close_file_type (SL_File_Table_Type *t)
 {
    int ret = 0;
@@ -278,13 +314,13 @@ static int close_file_type (SL_File_Table_Type *t)
 	if (0 == (t->flags & SL_PIPE))
 	  {
 	     if (EOF == (ret = fclose (fp)))
-	       _SLerrno_errno = errno;
+	       _pSLerrno_errno = errno;
 	  }
 #ifdef HAVE_POPEN
 	else
 	  {
 	     if (-1 == (ret = pclose (fp)))
-	       _SLerrno_errno = errno;
+	       _pSLerrno_errno = errno;
 	  }
 #endif
      }
@@ -395,7 +431,7 @@ static int stdio_fgets (SLang_Ref_Type *ref, SL_File_Table_Type *t)
      return -1;
 
    status = SLang_assign_to_ref (ref, SLANG_STRING_TYPE, (VOID_STAR)&s);
-   SLang_free_slstring (s);
+   _pSLang_free_slstring (s);
 
    if (status == -1)
      return -1;
@@ -408,7 +444,7 @@ static void stdio_fgetslines_internal (FILE *fp, unsigned int n)
    unsigned int num_lines, max_num_lines;
    char **list;
    SLang_Array_Type *at;
-   int inum_lines;
+   SLindex_Type inum_lines;
 
    if (n > 1024)
      max_num_lines = 1024;
@@ -449,7 +485,7 @@ static void stdio_fgetslines_internal (FILE *fp, unsigned int n)
 	     new_list = (char **) SLrealloc ((char *)list, sizeof (char *) * max_num_lines);
 	     if (new_list == NULL)
 	       {
-		  SLang_free_slstring (line);
+		  _pSLang_free_slstring (line);
 		  goto return_error;
 	       }
 	     list = new_list;
@@ -470,7 +506,7 @@ static void stdio_fgetslines_internal (FILE *fp, unsigned int n)
 	list = new_list;
      }
 
-   inum_lines = (int) num_lines;
+   inum_lines = (SLindex_Type) num_lines;
    if (NULL == (at = SLang_create_array (SLANG_STRING_TYPE, 0, (VOID_STAR) list, &inum_lines, 1)))
      goto return_error;
 
@@ -512,6 +548,41 @@ static void stdio_fgetslines (void)
    SLang_free_mmt (mmt);
 }
 
+static int stdio_fputslines (void)
+{
+   char **lines, **lines_max;
+   int num;
+   FILE *fp;
+   SLang_MMT_Type *mmt;
+   SLang_Array_Type *at;
+   
+   if (NULL == (mmt = pop_fp (SL_WRITE, &fp)))
+     return -1;
+
+   if (-1 == SLang_pop_array_of_type (&at, SLANG_STRING_TYPE))
+     {
+	SLang_free_mmt (mmt);
+	return -1;
+     }
+   lines = (char **)at->data;
+   lines_max = lines + at->num_elements;
+   
+   while (lines < lines_max)
+     {
+	if ((*lines != NULL)
+	    && (EOF == fputs (*lines, fp)))
+	  {
+	     _pSLerrno_errno = errno;
+	     break;
+	  }
+	lines++;
+     }
+   SLang_free_mmt (mmt);
+   num = (int) (lines - (char **)at->data);
+   SLang_free_array (at);
+   return num;
+}
+
 
 static int stdio_fputs (char *s, SL_File_Table_Type *t)
 {
@@ -521,7 +592,7 @@ static int stdio_fputs (char *s, SL_File_Table_Type *t)
      return -1;
 
    if (EOF == fputs(s, fp)) return -1;
-   return (int) strlen (s);
+   return (int) _pSLstring_bytelen (s);
 }
 
 static int stdio_fflush (SL_File_Table_Type *t)
@@ -533,13 +604,73 @@ static int stdio_fflush (SL_File_Table_Type *t)
 
    if (EOF == fflush (fp))
      {
-	_SLerrno_errno = errno;
+	_pSLerrno_errno = errno;
 	return -1;
      }
 
    return 0;
 }
 
+/* Helper function */
+static int check_ferror_and_realloc (FILE *fp, int ret, char **strp, unsigned int num_wanted,
+				     unsigned int num_read, unsigned int sizeof_type)
+{
+   if ((num_read == 0)
+       && (num_read != num_wanted))
+     ret = -1;
+
+   if ((ret == -1) && ferror (fp))
+     _pSLerrno_errno = errno;
+
+   if ((ret == 0)
+       && (num_read != num_wanted))
+     {
+	char *new_s;
+	unsigned int nbytes = num_read * sizeof_type;
+	new_s = SLrealloc (*strp, nbytes + 1);
+	if (new_s == NULL)
+	  ret = -1;
+	else
+	  *strp = new_s;
+     }
+   return ret;
+}
+
+static void stdio_fread_bytes (SLang_Ref_Type *ref, unsigned int *num_wantedp, SL_File_Table_Type *t)
+{
+   FILE *fp;
+   unsigned int num_read = 0, num_wanted = *num_wantedp;
+   int ret = -1;
+   char *buf = NULL;
+   SLang_BString_Type *bs;
+
+   if (NULL == (fp = check_fp (t, SL_READ)))
+     goto the_return;
+   
+   if (NULL == (buf = SLmalloc (num_wanted + 1)))
+     goto the_return;
+   
+   num_read = fread (buf, sizeof(char), num_wanted, fp);
+   
+   ret = check_ferror_and_realloc (fp, 0, &buf, num_wanted, num_read, sizeof (char));
+   if (ret == -1)
+     goto the_return;
+   
+
+   bs = SLbstring_create_malloced ((unsigned char *)buf, num_read, 1);
+   ret = SLang_assign_to_ref (ref, SLANG_BSTRING_TYPE, (VOID_STAR)&bs);
+   SLbstring_free (bs);
+   buf = NULL;
+
+   the_return:
+
+   if (buf != NULL) SLfree (buf);
+   if (ret == -1)
+     (void) SLang_push_integer (ret);
+   else
+     (void) SLang_push_uinteger (num_read);
+}
+
 /* Usage: n = fread (&str, data-type, nelems, fp); */
 static void stdio_fread (SLang_Ref_Type *ref, int *data_typep, unsigned int *num_elemns, SL_File_Table_Type *t)
 {
@@ -552,6 +683,16 @@ static void stdio_fread (SLang_Ref_Type *ref, int *data_typep, unsigned int *num
    unsigned int sizeof_type;
    int data_type;
 
+   /* FIXME: priority = low : I should add some mechanism to support
+    * other types.
+    */
+   data_type = *data_typep;
+   if ((data_type == SLANG_BSTRING_TYPE) || (data_type == SLANG_STRING_TYPE))
+     {
+	stdio_fread_bytes (ref, num_elemns, t);
+	return;
+     }
+
    ret = -1;
    s = NULL;
    /* cl = NULL; */
@@ -559,12 +700,7 @@ static void stdio_fread (SLang_Ref_Type *ref, int *data_typep, unsigned int *num
    if (NULL == (fp = check_fp (t, SL_READ)))
      goto the_return;
 
-   /* FIXME: priority = low : I should add some mechanism to support
-    * other types.
-    */
-   data_type = *data_typep;
-
-   cl = _SLclass_get_class ((unsigned char) data_type);
+   cl = _pSLclass_get_class ((SLtype) data_type);
 
    if (cl->cl_fread == NULL)
      {
@@ -585,52 +721,24 @@ static void stdio_fread (SLang_Ref_Type *ref, int *data_typep, unsigned int *num
 
    ret = cl->cl_fread (data_type, fp, (VOID_STAR)s, num_to_read, &num_read);
 
-   if ((num_read == 0)
-       && (num_read != num_to_read))
-     ret = -1;
+   ret = check_ferror_and_realloc (fp, ret, &s, num_to_read, num_read, sizeof_type);
+   if (ret == -1)
+     goto the_return;
 
-   if ((ret == -1) && ferror (fp))
-     _SLerrno_errno = errno;
-
-   if ((ret == 0)
-       && (num_read != num_to_read))
+   if (num_read == 1)
      {
-	char *new_s;
-
-	nbytes = num_read * sizeof_type;
-	new_s = SLrealloc (s, nbytes + 1);
-	if (new_s == NULL)
-	  ret = -1;
-	else
-	  s = new_s;
+	ret = SLang_assign_to_ref (ref, data_type, (VOID_STAR)s);
+	SLfree (s);
      }
-
-   if (ret == 0)
+   else
      {
-	if (num_read == 1)
-	  {
-	     ret = SLang_assign_to_ref (ref, data_type, (VOID_STAR)s);
-	     SLfree (s);
-	  }
-	else if ((data_type == SLANG_CHAR_TYPE)
-		 || (data_type == SLANG_UCHAR_TYPE))
-	  {
-	     SLang_BString_Type *bs;
-
-	     bs = SLbstring_create_malloced ((unsigned char *)s, num_read, 1);
-	     ret = SLang_assign_to_ref (ref, SLANG_BSTRING_TYPE, (VOID_STAR)&bs);
-	     SLbstring_free (bs);
-	  }
-	else
-	  {
-	     SLang_Array_Type *at;
-	     int inum_read = (int) num_read;
-	     at = SLang_create_array (data_type, 0, (VOID_STAR)s, &inum_read, 1);
-	     ret = SLang_assign_to_ref (ref, SLANG_ARRAY_TYPE, (VOID_STAR)&at);
-	     SLang_free_array (at);
-	  }
-	s = NULL;
+	SLang_Array_Type *at;
+	SLindex_Type inum_read = (SLindex_Type) num_read;
+	at = SLang_create_array (data_type, 0, (VOID_STAR)s, &inum_read, 1);
+	ret = SLang_assign_to_ref (ref, SLANG_ARRAY_TYPE, (VOID_STAR)&at);
+	SLang_free_array (at);
      }
+   s = NULL;
 
    the_return:
 
@@ -668,7 +776,7 @@ static void stdio_fwrite (SL_File_Table_Type *t)
 	if (NULL == (s = SLbstring_get_pointer (b, &num_to_write)))
 	  goto the_return;
 
-	cl = _SLclass_get_class (SLANG_UCHAR_TYPE);
+	cl = _pSLclass_get_class (SLANG_UCHAR_TYPE);
 	break;
 
       default:
@@ -693,7 +801,7 @@ static void stdio_fwrite (SL_File_Table_Type *t)
    ret = cl->cl_fwrite (cl->cl_data_type, fp, s, num_to_write, &num_write);
 
    if ((ret == -1) && ferror (fp))
-     _SLerrno_errno = errno;
+     _pSLerrno_errno = errno;
 
    /* drop */
    the_return:
@@ -708,37 +816,39 @@ static void stdio_fwrite (SL_File_Table_Type *t)
      SLang_push_uinteger (num_write);
 }
 
-static int stdio_fseek (SL_File_Table_Type *t, int *ofs, int *whence)
+static int stdio_fseek (SL_File_Table_Type *t, _pSLc_off_t_Type *ofs, int *whence)
 {
    FILE *fp;
 
    if (NULL == (fp = check_fp (t, 0xFFFF)))
      return -1;
 
-   if (-1  == fseek (fp, (long) *ofs, *whence))
+   if (-1 == FSEEK (fp, *ofs, *whence))
      {
-	_SLerrno_errno = errno;
+	_pSLerrno_errno = errno;
 	return -1;
      }
-
    return 0;
 }
 
-static int stdio_ftell (SL_File_Table_Type *t)
+static void stdio_ftell (SL_File_Table_Type *t)
 {
    FILE *fp;
-   long ofs;
+   _pSLc_off_t_Type ofs;
 
    if (NULL == (fp = check_fp (t, 0xFFFF)))
-     return -1;
-
-   if (-1L == (ofs = ftell (fp)))
      {
-	_SLerrno_errno = errno;
-	return -1;
+	(void) SLang_push_int (-1);
+	return;
      }
 
-   return (int) ofs;
+   if (-1L == (ofs = FTELL (fp)))
+     {
+	_pSLerrno_errno = errno;
+	(void) SLang_push_int (-1);
+	return;
+     }
+   (void) SLANG_PUSH_OFF_T (ofs);
 }
 
 static int stdio_feof (SL_File_Table_Type *t)
@@ -777,7 +887,7 @@ static int stdio_fprintf (void)
    SLang_MMT_Type *mmt;
    int status;
 
-   if (-1 == _SLstrops_do_sprintf_n (SLang_Num_Function_Args - 2))
+   if (-1 == _pSLstrops_do_sprintf_n (SLang_Num_Function_Args - 2))
      return -1;
    
    if (-1 == SLang_pop_slstring (&s))
@@ -785,17 +895,17 @@ static int stdio_fprintf (void)
    
    if (NULL == (mmt = pop_fp (SL_WRITE, &fp)))
      {
-	SLang_free_slstring (s);
+	_pSLang_free_slstring (s);
 	return -1;
      }
    
    if (EOF == fputs(s, fp))
      status = -1;
    else
-     status = (int) strlen (s);
+     status = (int) _pSLstring_bytelen (s);
 
    SLang_free_mmt (mmt);
-   SLang_free_slstring (s);
+   _pSLang_free_slstring (s);
    return status;
 }
 
@@ -804,7 +914,7 @@ static int stdio_printf (void)
    char *s;
    int status;
 
-   if (-1 == _SLstrops_do_sprintf_n (SLang_Num_Function_Args - 1))
+   if (-1 == _pSLstrops_do_sprintf_n (SLang_Num_Function_Args - 1))
      return -1;
 
    if (-1 == SLang_pop_slstring (&s))
@@ -813,9 +923,9 @@ static int stdio_printf (void)
    if (EOF == fputs(s, stdout))
      status = -1;
    else
-     status = (int) strlen (s);
+     status = (int) _pSLstring_bytelen (s);
 
-   SLang_free_slstring (s);
+   _pSLang_free_slstring (s);
    return status;
 }
 
@@ -831,6 +941,7 @@ static int stdio_printf (void)
 static SLang_Intrin_Fun_Type Stdio_Name_Table[] =
 {
    MAKE_INTRINSIC_0("fgetslines", stdio_fgetslines, V),
+   MAKE_INTRINSIC_0("fputslines", stdio_fputslines, I),
    MAKE_INTRINSIC_SS("fopen", stdio_fopen, V),
    MAKE_INTRINSIC_1("feof", stdio_feof, I, F),
    MAKE_INTRINSIC_1("ferror", stdio_ferror, I, F),
@@ -840,10 +951,11 @@ static SLang_Intrin_Fun_Type Stdio_Name_Table[] =
    MAKE_INTRINSIC_2("fputs", stdio_fputs, I, S, F),
    MAKE_INTRINSIC_0("fprintf", stdio_fprintf, I),
    MAKE_INTRINSIC_0("printf", stdio_printf, I),
-   MAKE_INTRINSIC_3("fseek", stdio_fseek, I, F, I, I),
-   MAKE_INTRINSIC_1("ftell", stdio_ftell, I, F),
+   MAKE_INTRINSIC_3("fseek", stdio_fseek, I, F, SLANG_C_OFF_T_TYPE, I),
+   MAKE_INTRINSIC_1("ftell", stdio_ftell, V, F),
    MAKE_INTRINSIC_1("clearerr", stdio_clearerr, V, F),
    MAKE_INTRINSIC_4("fread", stdio_fread, V, R, D, U, F),
+   MAKE_INTRINSIC_3("fread_bytes", stdio_fread_bytes, V, R, U, F),
    MAKE_INTRINSIC_1("fwrite", stdio_fwrite, V, F),
 #ifdef HAVE_POPEN
    MAKE_INTRINSIC_SS("popen", stdio_popen, V),
@@ -878,14 +990,14 @@ static SLang_IConstant_Type Stdio_Consts [] =
    SLANG_END_ICONST_TABLE
 };
 
-static void destroy_file_type (unsigned char type, VOID_STAR ptr)
+static void destroy_file_type (SLtype type, VOID_STAR ptr)
 {
    (void) type;
    (void) close_file_type ((SL_File_Table_Type *) ptr);
 }
 
 
-struct _SLang_Foreach_Context_Type
+struct _pSLang_Foreach_Context_Type
 {
    SLang_MMT_Type *mmt;
    FILE *fp;
@@ -897,7 +1009,7 @@ struct _SLang_Foreach_Context_Type
 
 
 static SLang_Foreach_Context_Type *
-cl_foreach_open (unsigned char type, unsigned int num)
+cl_foreach_open (SLtype type, unsigned int num)
 {
    SLang_Foreach_Context_Type *c;
    SLang_MMT_Type *mmt;
@@ -933,11 +1045,11 @@ cl_foreach_open (unsigned char type, unsigned int num)
 	     SLang_verror (SL_NOT_IMPLEMENTED,
 			   "using '%s' not supported by File_Type",
 			   s);
-	     SLang_free_slstring (s);
+	     _pSLang_free_slstring (s);
 	     SLang_free_mmt (mmt);
 	     return NULL;
 	  }
-	SLang_free_slstring (s);
+	_pSLang_free_slstring (s);
 	break;
 
       default:
@@ -962,7 +1074,7 @@ cl_foreach_open (unsigned char type, unsigned int num)
    return c;
 }
 
-static void cl_foreach_close (unsigned char type, SLang_Foreach_Context_Type *c)
+static void cl_foreach_close (SLtype type, SLang_Foreach_Context_Type *c)
 {
    (void) type;
    if (c == NULL) return;
@@ -970,7 +1082,7 @@ static void cl_foreach_close (unsigned char type, SLang_Foreach_Context_Type *c)
    SLfree ((char *) c);
 }
 
-static int cl_foreach (unsigned char type, SLang_Foreach_Context_Type *c)
+static int cl_foreach (SLtype type, SLang_Foreach_Context_Type *c)
 {
    int status;
    int ch;
@@ -996,7 +1108,7 @@ static int cl_foreach (unsigned char type, SLang_Foreach_Context_Type *c)
 	status = read_one_line (c->fp, &s, &len, (c->type==CTX_USE_LINE_WS));
 	if (status <= 0)
 	  return status;
-	if (0 == _SLang_push_slstring (s))
+	if (0 == _pSLang_push_slstring (s))
 	  return 1;
 	return -1;
      }
@@ -1023,18 +1135,16 @@ int SLang_init_stdio (void)
 
    if (NULL == (cl = SLclass_allocate_class ("File_Type")))
      return -1;
-   cl->cl_destroy = destroy_file_type;
-   cl->cl_foreach_open = cl_foreach_open;
-   cl->cl_foreach_close = cl_foreach_close;
-   cl->cl_foreach = cl_foreach;
-
+   
+   (void) SLclass_set_destroy_function (cl, destroy_file_type);
+   (void) SLclass_set_foreach_functions (cl, cl_foreach_open, cl_foreach, cl_foreach_close);
 
    if (-1 == SLclass_register_class (cl, SLANG_FILE_PTR_TYPE, sizeof (SL_File_Table_Type), SLANG_CLASS_TYPE_MMT))
      return -1;
 
    if ((-1 == SLadd_intrin_fun_table(Stdio_Name_Table, "__STDIO__"))
        || (-1 == SLadd_iconstant_table (Stdio_Consts, NULL))
-       || (-1 == _SLerrno_init ()))
+       || (-1 == _pSLerrno_init ()))
      return -1;
 
    names[0] = "stdin";
