@@ -1,9 +1,24 @@
-/* Copyright (c) 1998, 1999, 2001, 2002, 2003 John E. Davis
- * This file is part of the S-Lang library.
- *
- * You may distribute under the terms of either the GNU General Public
- * License or the Perl Artistic License.
- */
+/*
+Copyright (C) 2004, 2005, 2006 John E. Davis
+
+This file is part of the S-Lang Library.
+
+The S-Lang Library is free software; you can redistribute it and/or
+modify it under the terms of the GNU General Public License as
+published by the Free Software Foundation; either version 2 of the
+License, or (at your option) any later version.
+
+The S-Lang Library is distributed in the hope that it will be useful,
+but WITHOUT ANY WARRANTY; without even the implied warranty of
+MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+General Public License for more details.
+
+You should have received a copy of the GNU General Public License
+along with this library; if not, write to the Free Software
+Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
+USA.  
+*/
+
 #include "slinclud.h"
 
 #include "slang.h"
@@ -12,17 +27,27 @@
 #define MAX_TOKEN_LEN 254
 #define MAX_FILE_LINE_LEN 256
 
+int _pSLang_Compile_Line_Num_Info;
+#if SLANG_HAS_BOSEOS
+int _pSLang_Compile_BOSEOS;
+#endif
+#if SLANG_HAS_DEBUG_CODE
+static int Default_Compile_Line_Num_Info;
+#if 0
+static int Default_Compile_BOSEOS;
+#endif
+#endif
+
 static char Empty_Line[1] = {0};
 
-static int Default_Compile_Line_Num_Info;
 static char *Input_Line = Empty_Line;
 static char *Input_Line_Pointer;
 
-static SLPreprocess_Type *This_SLpp;
+static SLprep_Type *This_SLpp;
 
 static SLang_Load_Type *LLT;
 
-static char *map_token_to_string (_SLang_Token_Type *tok)
+static char *map_token_to_string (_pSLang_Token_Type *tok)
 {
    char *s;
    static char numbuf [32];
@@ -54,6 +79,18 @@ static char *map_token_to_string (_SLang_Token_Type *tok)
 	sprintf (s, "%lu", (unsigned long)tok->v.long_val);
 	break;
 
+#if HAVE_LONG_LONG
+      case LLONG_TOKEN:
+	s = numbuf;
+	sprintf (s, "%lld", tok->v.llong_val);
+	break;
+	
+      case ULLONG_TOKEN:
+	s = numbuf;
+	sprintf (s, "%llu", tok->v.ullong_val);
+	break;
+#endif
+	
       case OBRACKET_TOKEN: s = "["; break;
       case CBRACKET_TOKEN: s = "]"; break;
       case OPAREN_TOKEN: s = "("; break;
@@ -89,40 +126,26 @@ static char *map_token_to_string (_SLang_Token_Type *tok)
    return s;
 }
 
-static char *make_line_file_error (char *buf, unsigned int buflen,
-				   _SLang_Token_Type *tok, char *dsc, int line, char *file)
+void _pSLparse_error (int errcode, char *str, _pSLang_Token_Type *tok, int flag)
 {
-#if _SLANG_HAS_DEBUG_CODE
-   if (tok != NULL) line = tok->line_number;
-#endif
-   if (file == NULL) file = "??";
-
-   (void) _SLsnprintf (buf, buflen, "%s: found '%s', line %d, file: %s",
-		       dsc, map_token_to_string (tok), line, file);
-
-   return buf;
-}
-
-void _SLparse_error(char *str, _SLang_Token_Type *tok, int flag)
-{
-   char buf [1024];
+   int line = LLT->line_num;
+   char *file = (char *) LLT->name;
 
    if (str == NULL)
      str = "Parse Error";
 
-   make_line_file_error (buf, sizeof (buf), tok, str, LLT->line_num, (char *) LLT->name);
+#if SLANG_HAS_DEBUG_CODE
+   if (tok != NULL) line = tok->line_number;
+#endif
+   if (file == NULL) file = "??";
 
-   if ((flag == 0) && SLang_Error)
-     return;
+   if (flag || (_pSLang_Error == 0))
+     SLang_verror (errcode, "%s:%d: %s: found '%s'",
+		   file, line, str, map_token_to_string (tok));
 
-   SLang_verror (SL_SYNTAX_ERROR, "%s", buf);
+   (void) _pSLerr_set_line_info (file, line, NULL);
 }
 
-static void do_line_file_error (int line, char *file)
-{
-   SLang_verror (SL_SYNTAX_ERROR,
-		 "called from line %d, file: %s", line, file);
-}
 
 #define ALPHA_CHAR 	1
 #define DIGIT_CHAR	2
@@ -151,7 +174,7 @@ static void do_line_file_error (int line, char *file)
  * The third character null terminates the operator string.  This is for
  * the token structure.
  */
-static SLCONST char Operators [29][4] =
+static SLCONST char Operators [31][4] =
 {
 #define OFS_EXCL	0
      {'!',	'=',	0, NE_TOKEN},
@@ -178,22 +201,24 @@ static SLCONST char Operators [29][4] =
      {'/',	'=',	0, DIVEQS_TOKEN},
 #define OFS_LT		16
      {'<',	0,	0, LT_TOKEN},
+     {'<',	'<',	0, SHL_TOKEN},
      {'<',	'=',	0, LE_TOKEN},
-#define OFS_EQS		18
+#define OFS_EQS		19
      {'=',	0,	0, ASSIGN_TOKEN},
      {'=',	'=',	0, EQ_TOKEN},
-#define OFS_GT		20
+#define OFS_GT		21
      {'>',	0,	0, GT_TOKEN},
      {'>',	'=',	0, GE_TOKEN},
-#define OFS_AT		22
+     {'>',	'>',	0, SHR_TOKEN},
+#define OFS_AT		24
      {'@',	0,	0, DEREF_TOKEN},
-#define OFS_POW		23
+#define OFS_POW		25
      {'^',	0,	0, POW_TOKEN},
-#define OFS_BOR		24
+#define OFS_BOR		26
      {'|',	0,	0, BOR_TOKEN},
      {'|',	'|',	0, EOF_TOKEN},
      {'|',	'=',	0, BOREQS_TOKEN},
-#define OFS_BNOT	27
+#define OFS_BNOT	29
      {'~',	0,	0, BNOT_TOKEN},
      {	0,	0,	0, EOF_TOKEN}
 };
@@ -331,7 +356,7 @@ static SLCONST unsigned char Char_Type_Table[256][2] =
  { ALPHA_CHAR, 0 },	/*  */	    { ALPHA_CHAR, 0 },	/*  */
 };
 
-int _SLcheck_identifier_syntax (char *name)
+int _pSLcheck_identifier_syntax (char *name)
 {
    unsigned char *p;
    
@@ -378,7 +403,7 @@ static void unget_prep_char (unsigned char ch)
 
 #include "keywhash.c"
 
-static int get_ident_token (_SLang_Token_Type *tok, unsigned char *s, unsigned int len)
+static int get_ident_token (_pSLang_Token_Type *tok, unsigned char *s, unsigned int len)
 {
    unsigned char ch;
    unsigned char type;
@@ -393,6 +418,11 @@ static int get_ident_token (_SLang_Token_Type *tok, unsigned char *s, unsigned i
 	     unget_prep_char (ch);
 	     break;
 	  }
+	if (len == (MAX_TOKEN_LEN - 1))
+	  {
+	     _pSLparse_error (SL_BUILTIN_LIMIT_EXCEEDED, "Identifier length exceeded maximum supported value", NULL, 0);
+	     return tok->type = EOF_TOKEN;
+	  }
 	s [len++] = ch;
      }
 
@@ -406,12 +436,12 @@ static int get_ident_token (_SLang_Token_Type *tok, unsigned char *s, unsigned i
 	return (tok->type = table->type);
      }
 
-   tok->v.s_val = _SLstring_make_hashed_string ((char *)s, len, &tok->hash);
+   tok->v.s_val = _pSLstring_make_hashed_string ((char *)s, len, &tok->hash);
    tok->free_sval_flag = 1;
    return (tok->type = IDENT_TOKEN);
 }
 
-static int get_number_token (_SLang_Token_Type *tok, unsigned char *s, unsigned int len)
+static int get_number_token (_pSLang_Token_Type *tok, unsigned char *s, unsigned int len)
 {
    unsigned char ch;
    unsigned char type;
@@ -485,23 +515,23 @@ static int get_number_token (_SLang_Token_Type *tok, unsigned char *s, unsigned
      {
       default:
 	tok->v.s_val = (char *) s;
-	_SLparse_error ("Not a number", tok, 0);
+	_pSLparse_error (SL_TYPE_MISMATCH, "Not a number", tok, 0);
 	return (tok->type = EOF_TOKEN);
 
 #if SLANG_HAS_FLOAT
       case SLANG_FLOAT_TYPE:
-	tok->v.s_val = _SLstring_make_hashed_string ((char *)s, len, &tok->hash);
+	tok->v.s_val = _pSLstring_make_hashed_string ((char *)s, len, &tok->hash);
 	tok->free_sval_flag = 1;
 	return (tok->type = FLOAT_TOKEN);
 
       case SLANG_DOUBLE_TYPE:
-	tok->v.s_val = _SLstring_make_hashed_string ((char *)s, len, &tok->hash);
+	tok->v.s_val = _pSLstring_make_hashed_string ((char *)s, len, &tok->hash);
 	tok->free_sval_flag = 1;
 	return (tok->type = DOUBLE_TOKEN);
 #endif
 #if SLANG_HAS_COMPLEX
       case SLANG_COMPLEX_TYPE:
-	tok->v.s_val = _SLstring_make_hashed_string ((char *)s, len, &tok->hash);
+	tok->v.s_val = _pSLstring_make_hashed_string ((char *)s, len, &tok->hash);
 	tok->free_sval_flag = 1;
 	return (tok->type = COMPLEX_TOKEN);
 #endif
@@ -529,14 +559,22 @@ static int get_number_token (_SLang_Token_Type *tok, unsigned char *s, unsigned
       case SLANG_ULONG_TYPE:
 	tok->v.long_val = SLatoul (s);
 	return tok->type = ULONG_TOKEN;
+#ifdef HAVE_LONG_LONG
+      case SLANG_LLONG_TYPE:
+	tok->v.llong_val = SLatoll (s);
+	return tok->type = LLONG_TOKEN;
+      case SLANG_ULLONG_TYPE:
+	tok->v.ullong_val = SLatoull (s);
+	return tok->type = ULLONG_TOKEN;
+#endif
      }
 
    too_long_return_error:
-   _SLparse_error ("Number too long for buffer", NULL, 0);
-   return (tok->type == EOF_TOKEN);
+   _pSLparse_error (SL_BUILTIN_LIMIT_EXCEEDED, "Number too long for buffer", NULL, 0);
+   return (tok->type = EOF_TOKEN);
 }
 
-static int get_op_token (_SLang_Token_Type *tok, char ch)
+static int get_op_token (_pSLang_Token_Type *tok, char ch)
 {
    unsigned int offset;
    char second_char;
@@ -576,7 +614,7 @@ static int get_op_token (_SLang_Token_Type *tok, char ch)
 
    if (type == EOF_TOKEN)
      {
-	_SLparse_error ("Operator not supported", NULL, 0);
+	_pSLparse_error (SL_NOT_IMPLEMENTED, "Operator not supported", NULL, 0);
 	return type;
      }
 
@@ -588,46 +626,114 @@ static int get_op_token (_SLang_Token_Type *tok, char ch)
    return type;
 }
 
+
+/* s and t may point to the same buffer --- even for unicode.  This
+ * is because a wchar is denoted by (greater than) 4 characters \x{...}, which
+ * will expand to at most 6 bytes when UTF-8 encoded.  That is:
+ * \x{F} expands to 1 byte
+ * \x{FF} expands to 2 bytes
+ * \x{FFF} expands to 3 bytes
+ * \x{FFFF} expands to 3 bytes
+ * \x{FFFFF} expands to 4 bytes
+ * \x{FFFFFF} expands to 5 bytes
+ * \x{7FFFFFF} expands to 6 bytes
+ * 
+ * Also, consider octal, decimal, and hex forms:
+ * 
+ *    \200   (0x80)
+ *    \d128
+ *    \x80
+ * 
+ * In all these cases, the escaped form uses 4 bytes.  Hence, these forms also
+ * may be converted to UTF-8.
+ */
 /* If this returns non-zero, then it is a binary string */
 static int expand_escaped_string (register char *s,
 				  register char *t, register char *tmax,
-				  unsigned int *lenp)
+				  unsigned int *lenp, int is_binary)
 {
    char *s0;
-   int is_binary = 0;
    char ch;
+#if 0
+   int utf8_encode;
 
+   utf8_encode = (is_binary == 0) && _pSLinterp_UTF8_Mode;
+#endif
    s0 = s;
    while (t < tmax)
      {
+	int isunicode;
+	SLwchar_Type wch;
+	char *s1;
 	ch = *t++;
-	if (ch == '\\')
+	
+	if (ch != '\\')
 	  {
-	     t = _SLexpand_escaped_char (t, &ch);
 	     if (ch == 0) is_binary = 1;
+	     *s++ = ch;
+	     continue;
 	  }
-	*s++ = ch;
+	  
+	if (NULL == (t = _pSLexpand_escaped_char (t, &wch, &isunicode)))
+	  {
+	     is_binary = -1;
+	     break;
+	  }
+	if ((isunicode == 0)
+#if 0
+	    && ((wch < 127)
+		|| (utf8_encode == 0))
+#endif
+	    )
+	  {
+	     if (wch == 0)
+	       is_binary = 1;
+	     
+	     *s++ = (char) wch;
+	     continue;
+	  }
+#if 0	
+	if (isunicode && (utf8_encode == 0))
+	  {
+	     SLang_verror (SL_NOT_IMPLEMENTED, "Unicode is not supported by this application");
+	     is_binary = -1;
+	     break;
+	  }
+#endif
+	/* Escaped representation is always greater than encoded form.
+	 * So, 6 below is ok (although ugly).
+	 */
+	s1 = (char *) SLutf8_encode (wch, (SLuchar_Type *)s, 6);
+	if (s1 == NULL)
+	  {
+	     SLang_verror (SL_INVALID_UTF8, "Unable to UTF-8 encode 0x%lX\n", (unsigned long)wch);
+	     is_binary = -1;
+	     break;
+	  }
+	s = s1;
      }
    *s = 0;
 
    *lenp = (unsigned char) (s - s0);
    return is_binary;
 }
-
-static int get_string_token (_SLang_Token_Type *tok, unsigned char quote_char,
+   
+static int get_string_token (_pSLang_Token_Type *tok, unsigned char quote_char,
 			     unsigned char *s)
 {
+   SLwchar_Type wch;
    unsigned char ch;
    unsigned int len = 0;
    int has_quote = 0;
-   int is_binary;
-
+#if 0
+   int is_unicode = 0;
+#endif
    while (1)
      {
 	ch = prep_get_char ();
 	if (ch == 0)
 	  {
-	     _SLparse_error("Expecting quote-character", NULL, 0);
+	     _pSLparse_error(SL_SYNTAX_ERROR, "Expecting quote-character", NULL, 0);
 	     return (tok->type = EOF_TOKEN);
 	  }
 	if (ch == quote_char) break;
@@ -636,8 +742,8 @@ static int get_string_token (_SLang_Token_Type *tok, unsigned char quote_char,
 
 	if (len == (MAX_TOKEN_LEN - 1))
 	  {
-	     _SLparse_error ("String too long for buffer", NULL, 0);
-	     return (tok->type == EOF_TOKEN);
+	     _pSLparse_error (SL_BUILTIN_LIMIT_EXCEEDED, "String too long for buffer", NULL, 0);
+	     return (tok->type = EOF_TOKEN);
 	  }
 
 	if (ch == '\\')
@@ -646,16 +752,67 @@ static int get_string_token (_SLang_Token_Type *tok, unsigned char quote_char,
 	     ch = prep_get_char ();
 	     s[len++] = ch;
 	  }
+#if 0
+	else if (ch & 0x80)	       /* could be unicode */
+	  is_unicode = 1;
+#endif
      }
 
    s[len] = 0;
 
-   if (has_quote)
-     is_binary = expand_escaped_string ((char *) s, (char *)s, (char *)s + len, &len);
-   else is_binary = 0;
-
    if ('"' == quote_char)
      {
+#define STRING_SUFFIX_B		1
+#define STRING_SUFFIX_Q		2
+#define STRING_SUFFIX_R		4
+#define STRING_SUFFIX_S		8	
+	int suffix = 0;
+	int is_binary = 0;
+
+	while (1)
+	  {
+	     ch = prep_get_char ();
+	     if (ch == 'B')
+	       {
+		  is_binary = 1;
+		  suffix |= STRING_SUFFIX_B;
+		  continue;
+	       }
+	     if (ch == 'R')
+	       {
+		  suffix |= STRING_SUFFIX_R;
+		  has_quote = 0;
+		  continue;
+	       }
+	     if (ch == 'Q')
+	       {
+		  suffix |= STRING_SUFFIX_Q;
+		  continue;
+	       }
+	     if (ch == '$')
+	       {
+		  suffix |= STRING_SUFFIX_S;
+		  continue;
+	       }
+	     unget_prep_char (ch);
+	     break;
+	  }
+	
+	if ((suffix & STRING_SUFFIX_R) && (suffix & STRING_SUFFIX_Q))
+	  {
+	     _pSLparse_error (SL_SYNTAX_ERROR, "Conflicting suffix for string literal", NULL, 0);
+	     return (tok->type = EOF_TOKEN);
+	  }
+
+	if (has_quote)
+	  is_binary = expand_escaped_string ((char *) s, (char *)s, (char *)s + len, &len, is_binary);
+
+	if (is_binary && (suffix & STRING_SUFFIX_S))
+	  {
+	     _pSLparse_error (SL_SYNTAX_ERROR, "A binary string cannot have the $ suffix", NULL, 0);
+	     return tok->type = EOF_TOKEN;
+	  }
+
 	tok->free_sval_flag = 1;
 	if (is_binary)
 	  {
@@ -664,26 +821,48 @@ static int get_string_token (_SLang_Token_Type *tok, unsigned char quote_char,
 	  }
 	else
 	  {
-	     tok->v.s_val = _SLstring_make_hashed_string ((char *)s,
+	     tok->v.s_val = _pSLstring_make_hashed_string ((char *)s,
 							  len,
 							  &tok->hash);
 	     tok->free_sval_flag = 1;
+	     if (suffix & STRING_SUFFIX_S)
+	       return tok->type = STRING_DOLLAR_TOKEN;
 	     return (tok->type = STRING_TOKEN);
 	  }
      }
 
    /* else single character */
-   if ((len == 0) || (s[1] != 0))
-     {
-	_SLparse_error("Single char expected", NULL, 0);
-	return (tok->type = EOF_TOKEN);
-     }
 
-   tok->v.long_val = s[0];
+   if (has_quote)
+     {
+	if ((s[0] != '\\')
+	    || (NULL == (s = (unsigned char *)_pSLexpand_escaped_char ((char *)s+1, &wch, NULL)))
+	    || (*s != 0))
+	  {
+	     _pSLparse_error (SL_SYNTAX_ERROR, "Unable to parse character", NULL, 0);
+	     return (tok->type = EOF_TOKEN);
+	  }
+     }
+   else if (len == 1)
+     wch = s[0];
+   else /* Assume unicode */
+     {
+	unsigned char *ss = SLutf8_decode (s, s+len, &wch, NULL);
+	if ((ss == NULL) || (*ss != 0))
+	  {
+	     _pSLparse_error(SL_SYNTAX_ERROR, "Single char expected", NULL, 0);
+	     return (tok->type = EOF_TOKEN);
+	  }
+     }
+   tok->v.long_val = wch;
+
+   if (wch > 256)
+     return tok->type = ULONG_TOKEN;
+
    return (tok->type = UCHAR_TOKEN);
 }
 
-static int extract_token (_SLang_Token_Type *tok, unsigned char ch, unsigned char t)
+static int extract_token (_pSLang_Token_Type *tok, unsigned char ch, unsigned char t)
 {
    unsigned char s [MAX_TOKEN_LEN];
    unsigned int slen;
@@ -712,7 +891,7 @@ static int extract_token (_SLang_Token_Type *tok, unsigned char ch, unsigned cha
 	     unget_prep_char (ch);
 	     return get_op_token (tok, '!');
 	  }
-	_SLparse_error("Misplaced !", NULL, 0);
+	_pSLparse_error(SL_SYNTAX_ERROR, "Misplaced !", NULL, 0);
 	return -1;
 
       case DOT_CHAR:
@@ -733,12 +912,12 @@ static int extract_token (_SLang_Token_Type *tok, unsigned char ch, unsigned cha
 	return get_string_token (tok, ch, s);
 
       default:
-	_SLparse_error("Invalid character", NULL, 0);
+	_pSLparse_error(SL_SYNTAX_ERROR, "Invalid character", NULL, 0);
 	return (tok->type = EOF_TOKEN);
      }
 }
 
-int _SLget_rpn_token (_SLang_Token_Type *tok)
+int _pSLget_rpn_token (_pSLang_Token_Type *tok)
 {
    unsigned char ch;
 
@@ -760,7 +939,7 @@ int _SLget_rpn_token (_SLang_Token_Type *tok)
    return EOF_TOKEN;
 }
 
-int _SLget_token (_SLang_Token_Type *tok)
+int _pSLget_token (_pSLang_Token_Type *tok)
 {
    unsigned char ch;
    unsigned char t;
@@ -768,10 +947,10 @@ int _SLget_token (_SLang_Token_Type *tok)
    tok->num_refs = 1;
    tok->free_sval_flag = 0;
    tok->v.s_val = "??";
-#if _SLANG_HAS_DEBUG_CODE
+#if SLANG_HAS_DEBUG_CODE
    tok->line_number = LLT->line_num;
 #endif
-   if (SLang_Error || (Input_Line == NULL))
+   if (_pSLang_Error || (Input_Line == NULL))
      return (tok->type = EOF_TOKEN);
 
    while (1)
@@ -786,11 +965,11 @@ int _SLget_token (_SLang_Token_Type *tok)
 	do
 	  {
 	     LLT->line_num++;
-#if _SLANG_HAS_DEBUG_CODE
+#if SLANG_HAS_DEBUG_CODE
 	     tok->line_number++;
 #endif
 	     Input_Line = LLT->read(LLT);
-	     if ((NULL == Input_Line) || SLang_Error)
+	     if ((NULL == Input_Line) || _pSLang_Error)
 	       {
 		  Input_Line_Pointer = Input_Line = NULL;
 		  return (tok->type = EOF_TOKEN);
@@ -807,13 +986,16 @@ int _SLget_token (_SLang_Token_Type *tok)
      }
 }
 
-static int prep_exists_function (char *line, char comment)
+static int prep_exists_function (SLprep_Type *pt, char *line)
 {
    char buf[MAX_FILE_LINE_LEN], *b, *bmax;
    unsigned char ch;
+   unsigned char comment;
 
+   (void) pt;
    bmax = buf + (sizeof (buf) - 1);
-
+   
+   comment = (unsigned char)'%';
    while (1)
      {
 	/* skip whitespace */
@@ -822,7 +1004,7 @@ static int prep_exists_function (char *line, char comment)
 	  line++;
 
 	if ((ch <= '\n')
-	    || (ch == (unsigned char) comment)) break;
+	    || (ch == comment)) break;
 
 	b = buf;
 	while ((ch = (unsigned char) *line) > ' ')
@@ -839,23 +1021,28 @@ static int prep_exists_function (char *line, char comment)
    return 0;
 }
 
-static int prep_eval_expr (char *expr)
+static int prep_eval_expr (SLprep_Type *pt, char *expr)
 {
    int ret;
    char *end;
+   void (*compile)(_pSLang_Token_Type *);
 
+   (void) pt;
    end = strchr (expr, '\n');
    if (end == NULL)
      end = expr + strlen (expr);
    expr = SLmake_nstring (expr, (unsigned int) (end - expr));
    if (expr == NULL)
      return -1;
-	
+
+   compile = _pSLcompile_ptr;
+   _pSLcompile_ptr = _pSLcompile;
    if ((0 != SLang_load_string (expr))
        || (-1 == SLang_pop_integer (&ret)))
      ret = -1;
    else
      ret = (ret != 0);
+   _pSLcompile_ptr = compile;
 
    SLfree (expr);
    return ret;
@@ -864,67 +1051,99 @@ static int prep_eval_expr (char *expr)
 
 int SLang_load_object (SLang_Load_Type *x)
 {
-   SLPreprocess_Type this_pp;
-   SLPreprocess_Type *save_this_pp;
+   SLprep_Type *this_pp;
+   SLprep_Type *save_this_pp;
    SLang_Load_Type *save_llt;
    char *save_input_line, *save_input_line_ptr;
-#if _SLANG_HAS_DEBUG_CODE
+#if SLANG_HAS_DEBUG_CODE
    int save_compile_line_num_info;
+#endif
+#if SLANG_HAS_BOSEOS
+   int save_compile_boseos;
 #endif
    int save_auto_declare_variables;
 
-   if (SLprep_exists_hook == NULL)
-     SLprep_exists_hook = prep_exists_function;
-
-   if (_SLprep_eval_hook == NULL)
-     _SLprep_eval_hook = prep_eval_expr;
-
-   if (-1 == SLprep_open_prep (&this_pp)) return -1;
-
-   if (-1 == _SLcompile_push_context (x))
+   if (NULL == (this_pp = SLprep_new ()))
      return -1;
+   (void) SLprep_set_exists_hook (this_pp, prep_exists_function);
+   (void) SLprep_set_eval_hook (this_pp, prep_eval_expr);
 
-#if _SLANG_HAS_DEBUG_CODE
-   save_compile_line_num_info = _SLang_Compile_Line_Num_Info;
+   if (-1 == _pSLcompile_push_context (x))
+     {
+	SLprep_delete (this_pp);
+	return -1;
+     }
+
+#if SLANG_HAS_DEBUG_CODE
+   save_compile_line_num_info = _pSLang_Compile_Line_Num_Info;
+#endif
+#if SLANG_HAS_BOSEOS
+   save_compile_boseos = _pSLang_Compile_BOSEOS;
 #endif
    save_this_pp = This_SLpp;
    save_input_line = Input_Line;
    save_input_line_ptr = Input_Line_Pointer;
    save_llt = LLT;
-   save_auto_declare_variables = _SLang_Auto_Declare_Globals;
+   save_auto_declare_variables = _pSLang_Auto_Declare_Globals;
 
-   This_SLpp = &this_pp;
+   This_SLpp = this_pp;
    Input_Line_Pointer = Input_Line = Empty_Line;
    LLT = x;
 
-   x->line_num = 0;
+   /* x->line_num = 0; */  /* already set to 0 when allocated. */
    x->parse_level = 0;
-   _SLang_Auto_Declare_Globals = x->auto_declare_globals;
+   _pSLang_Auto_Declare_Globals = x->auto_declare_globals;
 
-#if _SLANG_HAS_DEBUG_CODE
-   _SLang_Compile_Line_Num_Info = Default_Compile_Line_Num_Info;
+#if SLANG_HAS_DEBUG_CODE
+   _pSLang_Compile_Line_Num_Info = Default_Compile_Line_Num_Info;
 #endif
+#if SLANG_HAS_BOSEOS
+#if 0
+   /* Instead of setting this variable to 0, let it keep its current value.
+    * Suppose that the following evalfiles take place:
+    *  
+    *  A -> B1  --> C1  --> D1
+    *    -> B2  --> C1  --> D2
+    * 
+    * and that B1 sets _boseos_info to 1.  Then C1 and D1 will get this value
+    * but B2 will not, since it will get rest to 0 when the routine has finished
+    * loading B1.
+    */
+     {
+	char *env = getenv ("SLANG_BOSEOS");
+	if (env != NULL)
+	  _pSLang_Compile_BOSEOS = atoi (env);
+	else
+	  _pSLang_Compile_BOSEOS = 0;
+     }
+#endif
+#endif
+   _pSLparse_start (x);
+   if (_pSLang_Error)
+     {
+       if (_pSLang_Error != SL_Usage_Error)
+	  (void) _pSLerr_set_line_info (x->name, x->line_num, NULL);
+	/* Doing this resets the state of the line_info object */
+	(void) _pSLerr_set_line_info (x->name, x->line_num, "");
+     }
 
-   _SLparse_start (x);
-   if (SLang_Error)
-     do_line_file_error (x->line_num, x->name);
+   _pSLang_Auto_Declare_Globals = save_auto_declare_variables;
 
-   _SLang_Auto_Declare_Globals = save_auto_declare_variables;
-
-   if (SLang_Error) SLang_restart (0);
-
-   (void) _SLcompile_pop_context ();
+   (void) _pSLcompile_pop_context ();
 
    Input_Line = save_input_line;
    Input_Line_Pointer = save_input_line_ptr;
    LLT = save_llt;
+   SLprep_delete (this_pp);
    This_SLpp = save_this_pp;
 
-#if _SLANG_HAS_DEBUG_CODE
-   _SLang_Compile_Line_Num_Info = save_compile_line_num_info;
+#if SLANG_HAS_DEBUG_CODE
+   _pSLang_Compile_Line_Num_Info = save_compile_line_num_info;
 #endif
-
-   if (SLang_Error) return -1;
+#if SLANG_HAS_BOSEOS
+   _pSLang_Compile_BOSEOS = save_compile_boseos;
+#endif
+   if (_pSLang_Error) return -1;
    return 0;
 }
 
@@ -953,7 +1172,7 @@ SLang_Load_Type *SLns_allocate_load_type (char *name, char *namespace_name)
 	     return NULL;
 	  }
      }
-   
+
    return x;
 }
 
@@ -1033,8 +1252,9 @@ int SLns_load_string (char *string, char *ns_name)
    x->read = read_from_string;
 
    data.ptr = data.string = string;
-   if (-1 == (ret = SLang_load_object (x)))
-     SLang_verror (SLang_Error, "called from eval: %s", string);
+   if ((-1 == (ret = SLang_load_object (x)))
+       && (SLang_Traceback))
+     _pSLerr_traceback_msg ("Traceback: called from eval: %s\n", string);
 
    SLang_free_slstring (string);
    SLdeallocate_load_type (x);
@@ -1100,7 +1320,7 @@ int SLns_load_file (char *f, char *ns_name)
    if (f == NULL) 
      name = SLang_create_slstring ("<stdin>");
    else
-     name = _SLpath_find_file (f);
+     name = _pSLpath_find_file (f);
 	
    if (name == NULL)
      return -1;
@@ -1141,7 +1361,7 @@ int SLns_load_file (char *f, char *ns_name)
    SLang_free_slstring (name);
    SLdeallocate_load_type (x);
 
-   if (SLang_Error)
+   if (_pSLang_Error)
      return -1;
 
    return 0;
@@ -1156,7 +1376,7 @@ static char *check_byte_compiled_token (char *buf)
        || ((len_hi = (unsigned char)*Input_Line_Pointer++) < 32)
        || ((len = (len_lo - 32) | ((len_hi - 32) << 7)) >= MAX_TOKEN_LEN))
      {
-	SLang_doerror ("Byte compiled file appears corrupt");
+	SLang_verror (SL_INVALID_DATA_ERROR, "Byte compiled file appears corrupt");
 	return NULL;
      }
 
@@ -1167,17 +1387,17 @@ static char *check_byte_compiled_token (char *buf)
    return buf;
 }
 
-void _SLcompile_byte_compiled (void)
+void _pSLcompile_byte_compiled (void)
 {
    unsigned char type;
-   _SLang_Token_Type tok;
+   _pSLang_Token_Type tok;
    char buf[MAX_TOKEN_LEN];
    char *ebuf;
    unsigned int len;
 
-   memset ((char *) &tok, 0, sizeof (_SLang_Token_Type));
+   memset ((char *) &tok, 0, sizeof (_pSLang_Token_Type));
 
-   while (SLang_Error == 0)
+   while (_pSLang_Error == 0)
      {
 	top_of_switch:
 	type = (unsigned char) *Input_Line_Pointer++;
@@ -1206,7 +1426,11 @@ void _SLcompile_byte_compiled (void)
 	       return;
 	     tok.v.long_val = atol (buf);
 	     break;
-
+#ifdef X___HAVE_LONG_LONG
+	   case LLONG_TOKEN:
+	   case ULLONG_TOKEN:
+	     XXX
+#endif
 	   case COMPLEX_TOKEN:
 	   case FLOAT_TOKEN:
 	   case DOUBLE_TOKEN:
@@ -1219,16 +1443,18 @@ void _SLcompile_byte_compiled (void)
 	     if (NULL == (ebuf = check_byte_compiled_token (buf)))
 	       return;
 	     tok.v.s_val = buf;
-	     if (expand_escaped_string (buf, buf, ebuf, &len))
-	       {
-		  tok.hash = len;
-		  type = _BSTRING_TOKEN;
-	       }
-	     else
-	       {
-		  tok.hash = _SLstring_hash ((unsigned char *)buf, (unsigned char *)buf + len);
-		  type = STRING_TOKEN;
-	       }
+	     (void) expand_escaped_string (buf, buf, ebuf, &len, 0);
+	     tok.hash = _pSLstring_hash ((unsigned char *)buf, (unsigned char *)buf + len);
+	     type = STRING_TOKEN;
+	     break;
+
+	   case ESC_BSTRING_TOKEN:
+	     if (NULL == (ebuf = check_byte_compiled_token (buf)))
+	       return;
+	     tok.v.s_val = buf;
+	     (void) expand_escaped_string (buf, buf, ebuf, &len, 1);
+	     tok.hash = len;
+	     type = _BSTRING_TOKEN;
 	     break;
 
 	   case TMP_TOKEN:
@@ -1240,7 +1466,7 @@ void _SLcompile_byte_compiled (void)
 	   case STRING_TOKEN:
 	   case IDENT_TOKEN:
 	   case _REF_TOKEN:
-	   case _DEREF_ASSIGN_TOKEN:
+	   /* case _DEREF_ASSIGN_TOKEN: */
 	   case _SCALAR_ASSIGN_TOKEN:
 	   case _SCALAR_PLUSEQS_TOKEN:
 	   case _SCALAR_MINUSEQS_TOKEN:
@@ -1266,7 +1492,7 @@ void _SLcompile_byte_compiled (void)
 	     if (NULL == (ebuf = check_byte_compiled_token (buf)))
 	       return;
 	     tok.v.s_val = buf;
-	     tok.hash = _SLstring_hash ((unsigned char *)buf, (unsigned char *)ebuf);
+	     tok.hash = _pSLstring_hash ((unsigned char *)buf, (unsigned char *)ebuf);
 	     break;
 
 	   default:
@@ -1274,7 +1500,7 @@ void _SLcompile_byte_compiled (void)
 	  }
 	tok.type = type;
 
-	(*_SLcompile_ptr) (&tok);
+	(*_pSLcompile_ptr) (&tok);
      }
 }
 
@@ -1335,7 +1561,7 @@ static int escape_string (unsigned char *s, unsigned char *smax,
 	     break;
 	  }
      }
-   _SLparse_error ("String too long to byte-compile", NULL, 0);
+   _pSLparse_error (SL_BUILTIN_LIMIT_EXCEEDED, "String too long to byte-compile", NULL, 0);
    return -1;
 }
 
@@ -1344,13 +1570,11 @@ static unsigned int Byte_Compile_Line_Len;
 
 static int bytecomp_write_data (char *buf, unsigned int len)
 {
-   char *err = "Write Error";
-
    if ((Byte_Compile_Line_Len + len + 1) >= MAX_FILE_LINE_LEN)
      {
 	if (EOF == fputs ("\n", Byte_Compile_Fp))
 	  {
-	     SLang_doerror (err);
+	     SLang_set_error (SL_IO_WRITE_ERROR);
 	     return -1;
 	  }
 	Byte_Compile_Line_Len = 0;
@@ -1358,14 +1582,14 @@ static int bytecomp_write_data (char *buf, unsigned int len)
 
    if (EOF == fputs (buf, Byte_Compile_Fp))
      {
-	SLang_doerror (err);
+	SLang_set_error (SL_IO_WRITE_ERROR);
 	return -1;
      }
    Byte_Compile_Line_Len += len;
    return 0;
 }
 
-static void byte_compile_token (_SLang_Token_Type *tok)
+static void byte_compile_token (_pSLang_Token_Type *tok)
 {
    unsigned char buf [MAX_TOKEN_LEN + 4], *buf_max;
    unsigned int len;
@@ -1373,7 +1597,7 @@ static void byte_compile_token (_SLang_Token_Type *tok)
    int is_escaped;
    unsigned char *s;
 
-   if (SLang_Error) return;
+   if (_pSLang_Error) return;
 
    buf [0] = (unsigned char) tok->type;
    buf [1] = 0;
@@ -1383,6 +1607,7 @@ static void byte_compile_token (_SLang_Token_Type *tok)
 
    switch (tok->type)
      {
+      case BOS_TOKEN:
       case LINE_NUM_TOKEN:
       case CHAR_TOKEN:
       case SHORT_TOKEN:
@@ -1407,7 +1632,7 @@ static void byte_compile_token (_SLang_Token_Type *tok)
 				 &is_escaped))
 	    return;
 
-	buf[0] = ESC_STRING_TOKEN;
+	buf[0] = ESC_BSTRING_TOKEN;
 	break;
 
       case BSTRING_TOKEN:
@@ -1418,7 +1643,7 @@ static void byte_compile_token (_SLang_Token_Type *tok)
 				 (unsigned char *)b3, buf_max,
 				 &is_escaped))
 	    return;
-	buf[0] = ESC_STRING_TOKEN;
+	buf[0] = ESC_BSTRING_TOKEN;
 	break;
 
       case STRING_TOKEN:
@@ -1433,8 +1658,8 @@ static void byte_compile_token (_SLang_Token_Type *tok)
 	  buf[0] = ESC_STRING_TOKEN;
 	break;
 
+      /* case _DEREF_ASSIGN_TOKEN: */
 	/* a _SCALAR_* token is attached to an identifier. */
-      case _DEREF_ASSIGN_TOKEN:
       case _SCALAR_ASSIGN_TOKEN:
       case _SCALAR_PLUSEQS_TOKEN:
       case _SCALAR_MINUSEQS_TOKEN:
@@ -1507,17 +1732,17 @@ int SLang_byte_compile_file (char *name, int method)
    Byte_Compile_Line_Len = 0;
    if (-1 != bytecomp_write_data (".#", 2))
      {
-	_SLcompile_ptr = byte_compile_token;
+	_pSLcompile_ptr = byte_compile_token;
 	(void) SLang_load_file (name);
-	_SLcompile_ptr = _SLcompile;
+	_pSLcompile_ptr = _pSLcompile;
 
 	(void) bytecomp_write_data ("\n", 1);
      }
 
    if (EOF == fclose (Byte_Compile_Fp))
-     SLang_doerror ("Write Error");
+     SLang_set_error (SL_IO_WRITE_ERROR);
 
-   if (SLang_Error)
+   if (_pSLang_Error)
      {
 	SLang_verror (0, "Error processing %s", name);
 	return -1;
@@ -1527,7 +1752,18 @@ int SLang_byte_compile_file (char *name, int method)
 
 int SLang_generate_debug_info (int x)
 {
+#if SLANG_HAS_DEBUG_CODE
    int y = Default_Compile_Line_Num_Info;
    Default_Compile_Line_Num_Info = x;
+#if 0
+   if (x == 0)
+     Default_Compile_BOSEOS = 0;
+   else
+     Default_Compile_BOSEOS = 3;
+#endif
    return y;
+#else
+   (void) x;
+   return 0;
+#endif
 }
