@@ -1,10 +1,24 @@
 /* Basic type operations for S-Lang */
-/* Copyright (c) 1992, 1996, 2001, 2002 John E. Davis
- * This file is part of the S-Lang library.
- *
- * You may distribute under the terms of either the GNU General Public
- * License or the Perl Artistic License.
- */
+/*
+Copyright (C) 2004, 2005, 2006 John E. Davis
+
+This file is part of the S-Lang Library.
+
+The S-Lang Library is free software; you can redistribute it and/or
+modify it under the terms of the GNU General Public License as
+published by the Free Software Foundation; either version 2 of the
+License, or (at your option) any later version.
+
+The S-Lang Library is distributed in the hope that it will be useful,
+but WITHOUT ANY WARRANTY; without even the implied warranty of
+MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+General Public License for more details.
+
+You should have received a copy of the GNU General Public License
+along with this library; if not, write to the Free Software
+Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
+USA.  
+*/
 
 #include "slinclud.h"
 
@@ -44,19 +58,7 @@ int SLang_pop_slstring (char **s) /*{{{*/
 
 /*}}}*/
 
-/* if *data != 0, string should be freed upon use. */
-int SLang_pop_string(char **s, int *data) /*{{{*/
-{
-   if (SLpop_string (s))
-     return -1;
-
-   *data = 1;
-   return 0;
-}
-
-/*}}}*/
-
-int _SLang_push_slstring (char *s)
+int _pSLang_push_slstring (char *s)
 {
    if (0 == SLclass_push_ptr_obj (SLANG_STRING_TYPE, (VOID_STAR)s))
      return 0;
@@ -65,12 +67,12 @@ int _SLang_push_slstring (char *s)
    return -1;
 }
 
-int _SLpush_alloced_slstring (char *s, unsigned int len)
+int _pSLpush_alloced_slstring (char *s, unsigned int len)
 {
-   if (NULL == (s = _SLcreate_via_alloced_slstring (s, len)))
+   if (NULL == (s = _pSLcreate_via_alloced_slstring (s, len)))
      return -1;
    
-   return _SLang_push_slstring (s);
+   return _pSLang_push_slstring (s);
 }
 
 int SLang_push_string (char *t) /*{{{*/
@@ -81,17 +83,17 @@ int SLang_push_string (char *t) /*{{{*/
    if (NULL == (t = SLang_create_slstring (t)))
      return -1;
 
-   return _SLang_push_slstring (t);
+   return _pSLang_push_slstring (t);
 }
 
 /*}}}*/
 
-int _SLang_dup_and_push_slstring (char *s)
+int _pSLang_dup_and_push_slstring (char *s)
 {
-   if (NULL == (s = _SLstring_dup_slstring (s)))
+   if (NULL == (s = _pSLstring_dup_slstring (s)))
      return SLang_push_null ();
 
-   return _SLang_push_slstring (s);
+   return _pSLang_push_slstring (s);
 }
 
 
@@ -138,8 +140,8 @@ static int int_int_power (int a, int b)
 #endif
 
 static int
-string_string_bin_op_result (int op, unsigned char a, unsigned char b,
-			     unsigned char *c)
+string_string_bin_op_result (int op, SLtype a, SLtype b,
+			     SLtype *c)
 {
    (void) a;
    (void) b;
@@ -166,8 +168,8 @@ string_string_bin_op_result (int op, unsigned char a, unsigned char b,
 
 static int
 string_string_bin_op (int op,
-		      unsigned char a_type, VOID_STAR ap, unsigned int na,
-		      unsigned char b_type, VOID_STAR bp, unsigned int nb,
+		      SLtype a_type, VOID_STAR ap, unsigned int na,
+		      SLtype b_type, VOID_STAR bp, unsigned int nb,
 		      VOID_STAR cp)
 {
    char *ic;
@@ -294,19 +296,19 @@ string_string_bin_op (int op,
    return -1;
 }
 
-static void string_destroy (unsigned char unused, VOID_STAR s)
+static void string_destroy (SLtype unused, VOID_STAR s)
 {
    (void) unused;
    SLang_free_slstring (*(char **) s);
 }
 
-static int string_push (unsigned char unused, VOID_STAR sptr)
+static int string_push (SLtype unused, VOID_STAR sptr)
 {
    (void) unused;
    return SLang_push_string (*(char **) sptr);
 }
 
-static int string_cmp (unsigned char unused, VOID_STAR ap, VOID_STAR bp, int *c)
+static int string_cmp (SLtype unused, VOID_STAR ap, VOID_STAR bp, int *c)
 {
    char *a, *b;
    (void) unused;
@@ -324,8 +326,8 @@ static int string_cmp (unsigned char unused, VOID_STAR ap, VOID_STAR bp, int *c)
    return 0;
 }
 
-static int string_to_int (unsigned char a_type, VOID_STAR ap, unsigned int na,
-			  unsigned char b_type, VOID_STAR bp)
+static int string_to_int (SLtype a_type, VOID_STAR ap, unsigned int na,
+			  SLtype b_type, VOID_STAR bp)
 {
    char **s;
    unsigned int i;
@@ -354,14 +356,14 @@ static int string_acopy (SLtype unused, VOID_STAR src_sptr, VOID_STAR dest_sptr)
    return 0;
 }
 
-struct _SLang_Foreach_Context_Type
+struct _pSLang_Foreach_Context_Type
 {
    char *string;
    unsigned int n;
 };
 
 static SLang_Foreach_Context_Type *
-string_foreach_open (unsigned char type, unsigned int num)
+string_foreach_open (SLtype type, unsigned int num)
 {
    char *s;
    SLang_Foreach_Context_Type *c;
@@ -390,7 +392,7 @@ string_foreach_open (unsigned char type, unsigned int num)
    return c;
 }
 
-static void string_foreach_close (unsigned char type, SLang_Foreach_Context_Type *c)
+static void string_foreach_close (SLtype type, SLang_Foreach_Context_Type *c)
 {
    (void) type;
    if (c == NULL) return;
@@ -398,7 +400,7 @@ static void string_foreach_close (unsigned char type, SLang_Foreach_Context_Type
    SLfree ((char *) c);
 }
 
-static int string_foreach (unsigned char type, SLang_Foreach_Context_Type *c)
+static int string_foreach (SLtype type, SLang_Foreach_Context_Type *c)
 {
    char ch;
 
@@ -415,42 +417,92 @@ static int string_foreach (unsigned char type, SLang_Foreach_Context_Type *c)
    return 1;
 }
 
-int _SLstring_list_push (_SLString_List_Type *p)
+SLang_Array_Type *_pSLstrings_to_array (char **strs, unsigned int n)
+{
+   char **data;
+   SLindex_Type i, inum;
+   SLang_Array_Type *at;
+
+   inum = (SLindex_Type) n;
+
+   if (NULL == (at = SLang_create_array (SLANG_STRING_TYPE, 0, NULL, &inum, 1)))
+     return NULL;
+
+   data = (char **)at->data;
+   for (i = 0; i < inum; i++)
+     {
+	if (strs[i] == NULL)
+	  {
+	     data[i] = NULL;
+	     continue;
+	  }
+	
+	if (NULL == (data[i] = SLang_create_slstring (strs[i])))
+	  {
+	     SLang_free_array (at);
+	     return NULL;
+	  }
+     }
+   return at;
+}
+
+static SLang_Array_Type *string_list_to_array (_pSLString_List_Type *p, int delete_list)
 {
    unsigned int num;
-   int inum;
+   SLindex_Type inum;
    SLang_Array_Type *at;
    char **buf;
-
-   if ((buf = p->buf) == NULL)
-     return SLang_push_null ();
    
+   buf = p->buf;
    num = p->num;
-   inum = (int) num;
+
+   if (delete_list == 0)
+     return _pSLstrings_to_array (buf, num);
+
+   inum = (SLindex_Type) num;
+   if (num == 0) num++;		       /* so realloc succeeds */
    
-   if (num == 0) num++;
-   if (num != p->max_num)
+   /* Since the list is to be deleted, we can steal the buffer */
+   if ((num != p->max_num)
+       && (NULL == (buf = (char **)SLrealloc ((char *) buf, sizeof (char *) * num))))
      {
-	if (NULL == (buf = (char **)SLrealloc ((char *) buf, sizeof (char *) * num)))
-	  {
-	     _SLstring_list_delete (p);
-	     return -1;
-	  }
-	p->max_num = num;
-	p->buf = buf;
+	_pSLstring_list_delete (p);
+	return NULL;
      }
-     
+   p->max_num = num;
+   p->buf = buf;
+   
    if (NULL == (at = SLang_create_array (SLANG_STRING_TYPE, 0, (VOID_STAR) buf, &inum, 1)))
      {
-	_SLstring_list_delete (p);
-	return -1;
+	_pSLstring_list_delete (p);
+	return NULL;
      }
    p->buf = NULL;
-   _SLstring_list_delete (p);
+   _pSLstring_list_delete (p);
+   return at;
+}
+
+int _pSLstring_list_push (_pSLString_List_Type *p, int delete_list)
+{
+   SLang_Array_Type *at;
+
+   if ((p == NULL)
+       || (p->buf == NULL))
+     {
+	int ret = SLang_push_null ();
+	if (delete_list)
+	  _pSLstring_list_delete (p);
+	return ret;
+     }
+   
+   if (NULL == (at = string_list_to_array (p, delete_list)))
+     return -1;
+
    return SLang_push_array (at, 1);
 }
 
-int _SLstring_list_init (_SLString_List_Type *p, unsigned int max_num, unsigned int delta_num)
+
+int _pSLstring_list_init (_pSLString_List_Type *p, unsigned int max_num, unsigned int delta_num)
 {
    if (NULL == (p->buf = (char **) SLmalloc (max_num * sizeof (char *))))
      return -1;
@@ -458,16 +510,14 @@ int _SLstring_list_init (_SLString_List_Type *p, unsigned int max_num, unsigned
    p->max_num = max_num;
    p->num = 0;
    p->delta_num = delta_num;
+   p->is_malloced = 0;
    return 0;
 }
 
-int _SLstring_list_append (_SLString_List_Type *p, char *s)
+int _pSLstring_list_append (_pSLString_List_Type *p, char *s)
 {
    if (s == NULL)
-     {
-	_SLstring_list_delete (p);
-	return -1;
-     }
+     return -1;
 
    if (p->max_num == p->num)
      {
@@ -475,11 +525,7 @@ int _SLstring_list_append (_SLString_List_Type *p, char *s)
 	unsigned int max_num = p->num + p->delta_num;
 	b = (char **)SLrealloc ((char *)p->buf, max_num * sizeof (char *));
 	if (b == NULL)
-	  {
-	     _SLstring_list_delete (p);
-	     SLang_free_slstring (s);
-	     return -1;
-	  }
+	  return -1;
 	p->buf = b;
 	p->max_num = max_num;
      }
@@ -489,8 +535,24 @@ int _SLstring_list_append (_SLString_List_Type *p, char *s)
    return 0;
 }
 
-void _SLstring_list_delete (_SLString_List_Type *p)
+int _pSLstring_list_append_copy (_pSLString_List_Type *p, char *s)
 {
+   s = SLang_create_slstring (s);
+   if (s == NULL)
+     return -1;
+   if (-1 == _pSLstring_list_append (p, s))
+     {
+	SLang_free_slstring (s);
+	return -1;
+     }
+   return 0;
+}
+
+void _pSLstring_list_delete (_pSLString_List_Type *p)
+{
+   if (p == NULL)
+     return;
+
    if (p->buf != NULL)
      {
 	unsigned int i, imax;
@@ -501,27 +563,76 @@ void _SLstring_list_delete (_SLString_List_Type *p)
 	SLfree ((char *)buf);
 	p->buf = NULL;
      }
+   if (p->is_malloced)
+     SLfree ((char *) p);
 }
 
+_pSLString_List_Type *_pSLstring_list_new (unsigned int max_num, unsigned int delta_num)
+{
+   _pSLString_List_Type *p;
+   
+   p = (_pSLString_List_Type *)SLmalloc (sizeof (_pSLString_List_Type));
+   if (p == NULL)
+     return NULL;
+   
+   if (-1 == _pSLstring_list_init (p, max_num, delta_num))
+     {
+	SLfree ((char *) p);
+	return NULL;
+     }
+   p->is_malloced = 1;
+   return p;
+}
+   
+
 /* Ref type */
+
 int SLang_pop_ref (SLang_Ref_Type **ref)
 {
    return SLclass_pop_ptr_obj (SLANG_REF_TYPE, (VOID_STAR *)ref);
 }
-
-/* Note: This is ok if ptr is NULL.  Some routines rely on this behavior */
-int _SLang_push_ref (int is_global, VOID_STAR ptr)
+ 
+static SLang_Ref_Type *create_ref_to_nametype (SLang_Name_Type *nt, int is_global)
 {
    SLang_Ref_Type *r;
 
-   if (ptr == NULL)
-     return SLang_push_null ();
-
    r = (SLang_Ref_Type *) SLmalloc (sizeof (SLang_Ref_Type));
-   if (r == NULL) return -1;
+   if (r == NULL) return NULL;
 
    r->is_global = is_global;
-   r->v.nt = (SLang_Name_Type *) ptr;
+   r->v.nt = nt;
+   return r;
+}
+
+int SLang_assign_nametype_to_ref (SLang_Ref_Type *ref, SLang_Name_Type *nt)
+{
+   SLang_Ref_Type *r;
+
+   if ((nt == NULL) || (ref == NULL))
+     return -1;
+   
+   if (NULL == (r = create_ref_to_nametype (nt, 1)))
+     return -1;
+
+   if (-1 == SLang_assign_to_ref (ref, SLANG_REF_TYPE, (VOID_STAR) &r))
+     {
+	SLang_free_ref (r);
+	return -1;
+     }
+   SLang_free_ref (r);
+   return 0;
+}
+
+/* Note: This is ok if nt is NULL.  Some routines rely on this behavior */
+static int push_nametype_as_ref (SLang_Name_Type *nt, int is_global)
+{
+   SLang_Ref_Type *r;
+
+   if (nt == NULL)
+     return SLang_push_null ();
+
+   r = create_ref_to_nametype (nt, is_global);
+   if (r == NULL) return -1;
 
    if (-1 == SLclass_push_ptr_obj (SLANG_REF_TYPE, (VOID_STAR) r))
      {
@@ -531,7 +642,12 @@ int _SLang_push_ref (int is_global, VOID_STAR ptr)
    return 0;
 }
 
-static void ref_destroy (unsigned char type, VOID_STAR ptr)
+int _pSLang_push_ref (int is_global, VOID_STAR ptr)
+{
+   return push_nametype_as_ref ((SLang_Name_Type *)ptr, is_global);
+}
+
+static void ref_destroy (SLtype type, VOID_STAR ptr)
 {
    (void) type;
    SLfree ((char *) *(SLang_Ref_Type **)ptr);
@@ -542,7 +658,7 @@ void SLang_free_ref (SLang_Ref_Type *ref)
    SLfree ((char *) ref);
 }
 
-static int ref_push (unsigned char type, VOID_STAR ptr)
+static int ref_push (SLtype type, VOID_STAR ptr)
 {
    SLang_Ref_Type *ref;
 
@@ -553,7 +669,7 @@ static int ref_push (unsigned char type, VOID_STAR ptr)
    if (ref == NULL)
      return SLang_push_null ();
 
-   return _SLang_push_ref (ref->is_global, (VOID_STAR) ref->v.nt);
+   return _pSLang_push_ref (ref->is_global, (VOID_STAR) ref->v.nt);
 }
 
 int SLang_assign_to_ref (SLang_Ref_Type *ref, SLtype type, VOID_STAR v)
@@ -561,7 +677,7 @@ int SLang_assign_to_ref (SLang_Ref_Type *ref, SLtype type, VOID_STAR v)
    SLang_Object_Type *stkptr;
    SLang_Class_Type *cl;
    
-   cl = _SLclass_get_class (type);
+   cl = _pSLclass_get_class (type);
 
    /* Use apush since this function is passing ``array'' bytes rather than the
     * address of the data.  I need to somehow make this more consistent.  To
@@ -569,6 +685,7 @@ int SLang_assign_to_ref (SLang_Ref_Type *ref, SLtype type, VOID_STAR v)
     * 
     *    double z[2];
     *    char *s = "silly";
+    *    char bytes[10];  BAD--- Don't do this
     *    int i;
     * 
     *    SLang_assign_to_ref (ref, SLANG_INT_TYPE,    &i);
@@ -582,17 +699,17 @@ int SLang_assign_to_ref (SLang_Ref_Type *ref, SLtype type, VOID_STAR v)
    if (-1 == (*cl->cl_apush) (type, v))
      return -1;
 
-   stkptr = _SLang_get_run_stack_pointer ();
-   if (0 == _SLang_deref_assign (ref))
+   stkptr = _pSLang_get_run_stack_pointer ();
+   if (0 == _pSLang_deref_assign (ref))
      return 0;
 
-   if (stkptr != _SLang_get_run_stack_pointer ())
+   if (stkptr != _pSLang_get_run_stack_pointer ())
      SLdo_pop ();
 
    return -1;
 }
 
-static char *ref_string (unsigned char type, VOID_STAR ptr)
+static char *ref_string (SLtype type, VOID_STAR ptr)
 {
    SLang_Ref_Type *ref;
 
@@ -616,13 +733,13 @@ static char *ref_string (unsigned char type, VOID_STAR ptr)
    return SLmake_string ("Local Variable Reference");
 }
 
-static int ref_dereference (unsigned char unused, VOID_STAR ptr)
+static int ref_dereference (SLtype unused, VOID_STAR ptr)
 {
    (void) unused;
-   return _SLang_dereference_ref (*(SLang_Ref_Type **) ptr);
+   return _pSLang_dereference_ref (*(SLang_Ref_Type **) ptr);
 }
 
-static int ref_cmp (unsigned char type, VOID_STAR a, VOID_STAR b, int *c)
+static int ref_cmp (SLtype type, VOID_STAR a, VOID_STAR b, int *c)
 {
    SLang_Ref_Type *ra, *rb;
 
@@ -649,6 +766,10 @@ static int ref_cmp (unsigned char type, VOID_STAR a, VOID_STAR b, int *c)
    return 0;
 }
 
+int SLang_push_function (SLang_Name_Type *nt)
+{
+   return push_nametype_as_ref (nt, 1);
+}
    
 SLang_Name_Type *SLang_pop_function (void)
 {
@@ -680,12 +801,18 @@ SLang_Name_Type *SLang_pop_function (void)
    return f;
 }
 
-/* This is a placeholder for version 2 */
+/* This is a placeholder for version 3 --- NULL assumed to be ok */
 void SLang_free_function (SLang_Name_Type *f)
 {
    (void) f;
 }
 
+SLang_Name_Type *SLang_copy_function (SLang_Name_Type *f)
+{
+   /* Add ref-count here */
+   return f;
+}
+
 /* NULL type */
 int SLang_push_null (void)
 {
@@ -695,16 +822,16 @@ int SLang_push_null (void)
 int SLang_pop_null (void)
 {
    SLang_Object_Type obj;
-   return _SLang_pop_object_of_type (SLANG_NULL_TYPE, &obj, 0);
+   return _pSLang_pop_object_of_type (SLANG_NULL_TYPE, &obj, 0);
 }
 
-static int null_push (unsigned char unused, VOID_STAR ptr_unused)
+static int null_push (SLtype unused, VOID_STAR ptr_unused)
 {
    (void) unused; (void) ptr_unused;
    return SLang_push_null ();
 }
 
-static int null_pop (unsigned char type, VOID_STAR ptr)
+static int null_pop (SLtype type, VOID_STAR ptr)
 {
    (void) type;
 
@@ -720,27 +847,27 @@ static int null_pop (unsigned char type, VOID_STAR ptr)
  * when it is NULL, we want foreach(X) to do nothing.
  */
 static SLang_Foreach_Context_Type *
-null_foreach_open (unsigned char type, unsigned int num)
+null_foreach_open (SLtype type, unsigned int num)
 {
    (void) type;
    SLdo_pop_n (num + 1);
    return (SLang_Foreach_Context_Type *)1;
 }
 
-static void null_foreach_close (unsigned char type, SLang_Foreach_Context_Type *c)
+static void null_foreach_close (SLtype type, SLang_Foreach_Context_Type *c)
 {
    (void) type;
    (void) c;
 }
 
-static int null_foreach (unsigned char type, SLang_Foreach_Context_Type *c)
+static int null_foreach (SLtype type, SLang_Foreach_Context_Type *c)
 {
    (void) type;
    (void) c;
    return 0;
 }
 
-static int null_to_bool (unsigned char type, int *t)
+static int null_to_bool (SLtype type, int *t)
 {
    (void) type;
    *t = 0;
@@ -748,8 +875,8 @@ static int null_to_bool (unsigned char type, int *t)
 }
 
 /* AnyType */
-int _SLanytype_typecast (unsigned char a_type, VOID_STAR ap, unsigned int na,
-			 unsigned char b_type, VOID_STAR bp)
+int _pSLanytype_typecast (SLtype a_type, VOID_STAR ap, unsigned int na,
+			 SLtype b_type, VOID_STAR bp)
 {
    SLang_Class_Type *cl;
    SLang_Any_Type **any;
@@ -760,7 +887,7 @@ int _SLanytype_typecast (unsigned char a_type, VOID_STAR ap, unsigned int na,
 
    any = (SLang_Any_Type **) bp;
    
-   cl = _SLclass_get_class (a_type);
+   cl = _pSLclass_get_class (a_type);
    sizeof_type = cl->cl_sizeof_type;
 
    for (i = 0; i < na; i++)
@@ -786,13 +913,15 @@ int SLang_pop_anytype (SLang_Any_Type **any)
 {
    SLang_Object_Type *obj;
 
-   *any = NULL;
-
    if (NULL == (obj = (SLang_Object_Type *) SLmalloc (sizeof (SLang_Object_Type))))
-     return -1;
+     {
+	*any = NULL;
+	return -1;
+     }
 
    if (-1 == SLang_pop (obj))
      {
+	*any = NULL;
 	SLfree ((char *) obj);
 	return -1;
      }
@@ -805,11 +934,11 @@ int SLang_pop_anytype (SLang_Any_Type **any)
  */
 int SLang_push_anytype (SLang_Any_Type *any)
 {
-   return _SLpush_slang_obj ((SLang_Object_Type *)any);
+   return _pSLpush_slang_obj ((SLang_Object_Type *)any);
 }
 
 /* After this call, the stack will contain an Any_Type object */
-static int anytype_push (unsigned char type, VOID_STAR ptr)
+static int anytype_push (SLtype type, VOID_STAR ptr)
 {
    SLang_Any_Type *obj;
 
@@ -819,7 +948,7 @@ static int anytype_push (unsigned char type, VOID_STAR ptr)
     * pushing an Any_Type onto the stack will create another copy of the
     * object represented by it.
     */
-   if (-1 == _SLpush_slang_obj (*(SLang_Object_Type **)ptr))
+   if (-1 == _pSLpush_slang_obj (*(SLang_Object_Type **)ptr))
      return -1;
 
    if (-1 == SLang_pop_anytype (&obj))
@@ -837,7 +966,7 @@ static int anytype_push (unsigned char type, VOID_STAR ptr)
    return 0;
 }
 
-static void anytype_destroy (unsigned char type, VOID_STAR ptr)
+static void anytype_destroy (SLtype type, VOID_STAR ptr)
 {
    SLang_Object_Type *obj;
 
@@ -853,34 +982,53 @@ void SLang_free_anytype (SLang_Any_Type *any)
      anytype_destroy (SLANG_ANY_TYPE, (VOID_STAR) &any);
 }
 
-static int anytype_dereference (unsigned char unused, VOID_STAR ptr)
+static int anytype_dereference (SLtype unused, VOID_STAR ptr)
 {
    (void) unused;
-   return _SLpush_slang_obj (*(SLang_Object_Type **) ptr);
+   return _pSLpush_slang_obj (*(SLang_Object_Type **) ptr);
 }
 
+#if 0
+/* This function performs a deref since we may want the symmetry
+ *  a = Any_Type[1];  a[x] = "foo"; bar = a[x]; ==> bar == "foo"
+ * That is, we do not want bar to be an Any_Type.
+ * 
+ * Unfortunately, this does not work because of the use of the transfer
+ * buffer by both slarray.c and sltypecast.c.  I can work around that 
+ * but I am not sure that I like typeof(Any_Type[0]) != Any_Type.
+ */
+static int anytype_apush (SLtype type, VOID_STAR ptr)
+{
+   (void) type;
+   return _pSLpush_slang_obj (*(SLang_Object_Type **)ptr);
+}
+#endif
+
 /* SLANG_INTP_TYPE */
-static int intp_push (unsigned char unused, VOID_STAR ptr)
+static int intp_push (SLtype unused, VOID_STAR ptr)
 {
    (void) unused;
    return SLclass_push_int_obj (SLANG_INT_TYPE, **(int **)ptr);
 }
 
-static int intp_pop (unsigned char unused, VOID_STAR ptr)
+static int intp_pop (SLtype unused, VOID_STAR ptr)
 {
    (void) unused;
    return SLang_pop_integer (*(int **) ptr);
 }
 
-static int undefined_push (unsigned char t, VOID_STAR p)
+static int undefined_method (SLtype t, VOID_STAR p)
 {
    (void) t; (void) p;
-   if (SLang_Error == 0)
-     SLang_Error = SL_VARIABLE_UNINITIALIZED;
+   SLang_set_error (SL_VARIABLE_UNINITIALIZED);
    return -1;
 }
+static void void_undefined_method (SLtype t, VOID_STAR p)
+{
+   (void) t; (void) p;
+}
 
-int _SLregister_types (void)
+int _pSLregister_types (void)
 {
    SLang_Class_Type *cl;
 
@@ -893,16 +1041,17 @@ int _SLregister_types (void)
        || (sizeof(double) != SIZEOF_DOUBLE))
      SLang_exit_error ("S-Lang Library not built properly.  Fix SIZEOF_* in config.h and recompile");
 #endif
-   if (-1 == _SLclass_init ())
+   if (-1 == _pSLclass_init ())
      return -1;
 
    /* Undefined Type */
    if (NULL == (cl = SLclass_allocate_class ("Undefined_Type")))
      return -1;
-   (void) SLclass_set_push_function (cl, undefined_push);
-   (void) SLclass_set_pop_function (cl, undefined_push);
-   if (-1 == SLclass_register_class (cl, SLANG_UNDEFINED_TYPE, sizeof (int),
-				     SLANG_CLASS_TYPE_SCALAR))
+   (void) SLclass_set_push_function (cl, undefined_method);
+   (void) SLclass_set_pop_function (cl, undefined_method);
+   (void) SLclass_set_destroy_function (cl, void_undefined_method);
+   if (-1 == SLclass_register_class (cl, SLANG_UNDEFINED_TYPE, sizeof (char *),
+				     SLANG_CLASS_TYPE_PTR))
      return -1;
    /* Make Void_Type a synonym for Undefined_Type.  Note that this does 
     * not mean that Void_Type represents SLANG_VOID_TYPE.  Void_Type is 
@@ -911,7 +1060,7 @@ int _SLregister_types (void)
    if (-1 == SLclass_create_synonym ("Void_Type", SLANG_UNDEFINED_TYPE))
      return -1;
 
-   if (-1 == _SLarith_register_types ())
+   if (-1 == _pSLarith_register_types ())
      return -1;
 
    /* SLANG_INTP_TYPE */
@@ -970,12 +1119,15 @@ int _SLregister_types (void)
      return -1;
    (void) SLclass_set_push_function (cl, anytype_push);
    (void) SLclass_set_destroy_function (cl, anytype_destroy);
+#if 0
+   (void) SLclass_set_apush_function (cl, anytype_apush);
+#endif
    cl->cl_dereference = anytype_dereference;
    if (-1 == SLclass_register_class (cl, SLANG_ANY_TYPE, sizeof (VOID_STAR),
 				     SLANG_CLASS_TYPE_PTR))
      return -1;
 
-   if (-1 == _SLang_init_bstring ())
+   if (-1 == _pSLang_init_bstring ())
      return -1;
 
    if ((-1 == SLclass_add_typecast (SLANG_STRING_TYPE, SLANG_INT_TYPE, string_to_int, 0))
