@@ -1,9 +1,24 @@
-/* Copyright (c) 1992, 1999, 2001, 2002, 2003 John E. Davis
- * This file is part of the S-Lang library.
- *
- * You may distribute under the terms of either the GNU General Public
- * License or the Perl Artistic License.
- */
+/*
+Copyright (C) 2004, 2005, 2006 John E. Davis
+
+This file is part of the S-Lang Library.
+
+The S-Lang Library is free software; you can redistribute it and/or
+modify it under the terms of the GNU General Public License as
+published by the Free Software Foundation; either version 2 of the
+License, or (at your option) any later version.
+
+The S-Lang Library is distributed in the hope that it will be useful,
+but WITHOUT ANY WARRANTY; without even the implied warranty of
+MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+General Public License for more details.
+
+You should have received a copy of the GNU General Public License
+along with this library; if not, write to the Free Software
+Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
+USA.  
+*/
+
 #include "slinclud.h"
 
 #include <ssdef.h>
@@ -119,7 +134,8 @@ static int getkey_ast(int not_used)
      {
 	if (c == SLang_Abort_Char)
 	  {
-	     if (SLang_Ignore_User_Abort == 0) SLang_Error = SL_USER_BREAK;
+	     if (SLang_Ignore_User_Abort == 0) 
+	       SLang_set_error (SL_USER_BREAK);
 	     SLKeyBoard_Quit = 1;
 	  }
 
@@ -317,7 +333,7 @@ void SLang_reset_tty (void)
 
 }
 
-unsigned int _SLsys_getkey()
+unsigned int _pSLsys_getkey()
 {
    unsigned int c;
 
@@ -330,13 +346,13 @@ unsigned int _SLsys_getkey()
    Waiting_For_Ast = 0;
 
    if (SLang_Input_Buffer_Len) return(SLang_getkey());
-   while (!_SLsys_input_pending(450));
+   while (!_pSLsys_input_pending(450));
    c = SLang_getkey();
    return(c);
 }
 
 /* waits *secs tenth of seconds for input */
-int _SLsys_input_pending(int tsecs)
+int _pSLsys_input_pending(int tsecs)
 {
    unsigned long daytim[2];
 
