@@ -1,4 +1,4 @@
-/* Copyright (c) 1998, 2001 John E. Davis
+/* Copyright (c) 1998, 2001, 2004 John E. Davis
  *
  * You may distribute under the terms of either the GNU General Public
  * License or the Perl Artistic License.
@@ -42,7 +42,7 @@ static void write_table (unsigned int min_hash, unsigned int max_hash)
    unsigned int i;
 
    fprintf (stdout, "\n\
-typedef struct\n\
+typedef SLCONST struct\n\
 {\n\
    char *name;\n\
    unsigned int type;\n\
@@ -157,6 +157,7 @@ static void write_map (int *map)
 #endif
      }
 
+   fprintf (stdout, "#ifndef SLCONST\n#define SLCONST const\n#endif\n");
    fprintf (stdout, "#define MIN_HASH_VALUE\t%u\n", min_hash);
    fprintf (stdout, "#define MAX_HASH_VALUE\t%u\n", max_hash);
    fprintf (stdout, "#define MIN_KEYWORD_LEN %u\n", Min_String_Len);
@@ -175,7 +176,7 @@ static void write_map (int *map)
    fprintf (stderr, "Hash Table is of type %s with max hash %u\n",
 	    type, max_hash);
 
-   fprintf (stdout, "\nstatic %s %s [256] =\n", 
+   fprintf (stdout, "\nstatic SLCONST %s %s [256] =\n", 
 	    type, C_Char_Map_Name);
 
    fprintf (stdout, "{\n  ");
